//////////////////////////////////////////////////////////////////////////////////
// Created on Wednesday March 27 17:10 2024
//
// Description:
// Used to obtain the power consumption and resource utilization
// of the device with only the ADC and the DAC running. New ADC
// samples are sent directly out of the DAC.
//
// Notes:
//
// @author: Robert Kent
//////////////////////////////////////////////////////////////////////////////////

module dual_adc_to_dac(

	//////////// CLOCK //////////
	input 		          		MAX10_CLK1_50,

	//////////// KEY //////////
	input 		          		FPGA_RESET_n,
	input 		     [4:0]		KEY,

	//////////// SW //////////
//	input 		     [9:0]		SW,

	//////////// LEDR //////////
	output		     [9:0]		LEDR,

	//////////// HEX //////////
	output		     [6:0]		HEX0,
	output		     [6:0]		HEX1,

	//////////// DAC //////////
	inout 		          		DIN,
	output		          		SCLK,
	output		          		SYNC
);

//******************************************LOCAL PARAMETERS******************************************//

// DAC
localparam output_word_length = 16;
// The DAC value that gives zero current as an output, after being passed
// through the custom built voltage to current converter
localparam DAC_ZERO_CURRENT = 33190;

//******************************************REG/WIRE DECLARATIONS******************************************//

// CLOCKS
wire clk_80MHz/* synthesis keep */;
wire clk_30MHz/* synthesis keep */;
wire locked/* synthesis keep */;
wire DAC_FLAG/* synthesis keep */;

// COUNTERS/STATE MACHINES
reg [6:0] clk_80_counter = 0/* synthesis noprune */;
localparam clk_80_counter_max = 79;
// Which clock cycle to start the DAC write sequence
localparam start_write = 63;
// Which clock cycle to interrupt the DAC write sequence
localparam interrupt_write = 52;

// ADC1
reg  		   cmd_valid/* synthesis noprune */;
wire [4:0]  cmd_channel /* synthesis keep */;
reg		   cmd_sop/* synthesis noprune */;
wire		   cmd_eop/* synthesis keep */;
wire		   cmd_rdy/* synthesis keep */;

wire        rspn_valid/* synthesis keep */;
wire [4:0]  rspn_ch/* synthesis keep */;
wire [11:0] rspn_data/* synthesis keep */;

reg  [16:0] lock_dly_cnt/* synthesis noprune */;
reg  [11:0] adc_1_data/* synthesis noprune */;
reg [11:0] adc_1_out = 0/* synthesis noprune */;

// ADC2
reg		   cmd_2_sop/* synthesis noprune */;
wire		   cmd_2_eop/* synthesis keep */;
wire		   cmd_2_rdy/* synthesis keep */;

wire        rspn_2_valid/* synthesis keep */;
wire [4:0]  rspn_2_ch/* synthesis keep */;
wire [11:0] rspn_2_data/* synthesis keep */;

reg  [11:0] adc_2_data/* synthesis noprune */;
reg [11:0] adc_2_out = 0/* synthesis noprune */;

// DAC
reg RESET = 0 /* synthesis noprune */;
reg [output_word_length - 1:0] DAC_DATA = DAC_ZERO_CURRENT /* synthesis noprune */;
wire [7:0] CNT /* synthesis noprune */;

//******************************************MODULE INSTANTIATION******************************************//

// DUAL ADC
	// MAX ADC CLOCK: 80MHz
	// MAX IP CLOCK: ~95MHz
dual_adc my_dual_adc (
	.clock_clk                (clk_80MHz),      //          clock.clk (IP CLOCK)
	.reset_sink_reset_n       (KEY[0]),       //     reset_sink.reset_n
	.adc_pll_clock_clk        (clk_80MHz),      //  adc_pll_clock.clk (ADC CLOCK)
	.adc_pll_locked_export    (locked),  // adc_pll_locked.export
	.command_valid            (cmd_valid),    //        command.valid
	.command_channel          (cmd_channel),  //               .channel
	.command_startofpacket    (cmd_sop),      //               .startofpacket
	.command_endofpacket      (cmd_eop),      //               .endofpacket
	.command_ready            (cmd_rdy),      //               .ready
	.command_2_valid          (cmd_valid),  //      command_2.valid
	.command_2_channel        (cmd_channel),//               .channel
	.command_2_startofpacket  (cmd_2_sop),    //               .startofpacket
	.command_2_endofpacket    (cmd_2_eop),    //               .endofpacket
	.command_2_ready          (cmd_2_rdy),    //               .ready
	.response_valid           (rspn_valid),   //       response.valid
	.response_channel         (rspn_ch),      //               .channel
	.response_data            (rspn_data),    //               .data
	.response_startofpacket   (),             //               .startofpacket
	.response_endofpacket     (),             //               .endofpacket
	.response_2_valid         (rspn_2_valid), //     response_2.valid
	.response_2_channel       (rspn_2_ch),    //               .channel
	.response_2_data          (rspn_2_data),  //               .data
	.response_2_startofpacket (),             //               .startofpacket
	.response_2_endofpacket   ()              //               .endofpacket
);
	
	
// PLL
pll pll1(
	.inclk0(MAX10_CLK1_50), // 50 MHz input clock
	.c0(clk_80MHz), // 80 MHz output clock with 0 phase shift and 50% duty cycle
	.c1(clk_30MHz), // 30 MHz output clock with -50.62 phase shift and 50% duty cycle
	.locked(locked)
	);
	
// DAC
dac_fast mydac(
    .CLK( clk_30MHz ), // 30 MHz output clock with -50.62 phase shift and 50% duty cycle
    .RESET( RESET ), // active low, raises SYNC to interrupt write sequence
    .DAC_DATA( DAC_DATA ), // 16-bit data from user to convert to analog signal
    .SYNC( SYNC ), // frame synchronization signal for the input data
    .SCLK( SCLK ), // Serial clock input
    .DIN( DIN ), // Serial data input
    .CNT2( CNT ), // counter for SCLK
	 .DAC_FLAG( DAC_FLAG ) // high when DIN is the first or Last data bit
    );
	
	
//******************************************STRUCTURAL CODING******************************************//

// ADC LOCK
always@(posedge clk_80MHz)	
	if ( locked && lock_dly_cnt != 80000 )
		lock_dly_cnt <= lock_dly_cnt + 1;

wire lock_rdy = (lock_dly_cnt == 80000)? 1 : 0;	

assign cmd_channel = 0;

// ADC COMMAND VALID
always@(posedge clk_80MHz)
begin
	cmd_valid <= lock_rdy;
end

//******************************************MAIN CODE******************************************//

// Turn off 7 segment displays
assign HEX0[6:0] = 7'b1111111;
assign HEX1[6:0] = 7'b1111111;

// to decrease sample rate from 1MHZ ADC
localparam SAMPLE_DIVIDER = 5;//10 for 100KHz 5 for 200KHz
reg [4-1:0] sample_counter = 1/* synthesis noprune */;

always @ (posedge clk_80MHz)
begin
	if (rspn_valid)
	begin
		// Save incoming ADC data when rspn_valid is high
		adc_1_data = rspn_data;
		adc_2_data = rspn_2_data;
		
		// if the sample counter is less than sample divider, increment sample counter
		if (sample_counter < SAMPLE_DIVIDER)
		begin
			sample_counter = sample_counter + 1;
		end
		// else reset the sample counter and the clock counter.
		// the clock counter does not automatically reset after
		// reaching the maximum value, it needs to wait until the
		// sample counter has reached the maximum value
		else
		begin
			sample_counter = 1;
			clk_80_counter = 0;
		end
		
	end
	
	// if the clock counter is less than clock counter max, increment it
	else if (clk_80_counter < clk_80_counter_max)
		clk_80_counter = clk_80_counter + 1;
	
	case(clk_80_counter)
	4: // if the clock counter is 4, pad the 12-bit ADC data with 0000
		// and send the output out of the 16-bit DAC
		begin
			DAC_DATA = {adc_1_data,4'b0000};
		end
	
	interrupt_write: // interrupt the DAC write sequence
		begin
			RESET = 0;
		end
		
	start_write: // initiate the DAC write sequence
	begin
		RESET = 1;
	end	
	
	endcase
end
	
endmodule

