
//////////////////////////////////////////////////////////////////////////////////
// Created on Wednesday March 27 17:10 2024
//
// Description:
// Used to control the double scroll circuit using the control algorithm with
// nonlinear terms, and m=1 step ahead predictions. This code is controlled by 
// writing to block memory using the In-System Memory Content Editor. It has
// full functionality, including writing training and testing perturbations,
// writing arbitrary desired signals, setting the control gain, switching between
// full nonlinear (NG-RC) and linear operating modes, setting the added delay
// in applying the control perturbation, and switching the control algorithm
// on and off by writing to memory. It also computes and saves the perturbed
// predictions, which is not required by the control algorithm. This project
// is not used to estimate the resource utilization and power consumption, see
// the other included files.
// 
//
// Notes:
//
// @author: Robert Kent
//////////////////////////////////////////////////////////////////////////////////

module dual_adc_to_dac(

	//////////// CLOCK //////////
	input 		          		MAX10_CLK1_50,

	//////////// KEY //////////
	input 		          		FPGA_RESET_n,
	input 		     [4:0]		KEY,

	//////////// SW //////////
	input 		     [9:0]		SW,

	//////////// LEDR //////////
	output		     [9:0]		LEDR,

	//////////// HEX //////////
	output		     [6:0]		HEX0,
	output		     [6:0]		HEX1,

	//////////// DAC //////////
	inout 		          		DIN,
	output		          		SCLK,
	output		          		SYNC
);

//******************************************LOCAL PARAMETERS******************************************//

//Weights (trained on 04/02/2024 07:15 PM)
// Fixed-Point Parameters
// Linear Features
localparam input_word_length = 12;
localparam difference_word_length = 13;
localparam output_word_length = 16;
localparam linear_word_length = 17;
localparam feature_word_length = 18;
reg [5 - 1:0] input_padding = 0;

// Quadratic Features
localparam quadratic_full_word_length = 26;
localparam quadratic_round_word_length = 18;
localparam quadratic_start = 8;

// Cubic Features
localparam cubic_full_word_length = 29;
localparam cubic_round_word_length = 18;
localparam cubic_start = 11;

// Desired
localparam desired_word_length = 16;
localparam desired_padded_word_length = 32;
reg [16 - 1:0] desired_padding = 0;
reg [1 - 1:0] desired_error_padding = 0;

// Error
localparam error_word_length = 18;

// Weights
localparam weight_word_length = 18;
localparam Kerror_word_length = 36;
localparam weight_fractional_bits = 15;

// Prediction
localparam unperturbed_prediction_word_length = 38;
localparam prediction_word_length = 38;
localparam pred_start = 16;
localparam prediction_trunc_word_length = 16;

// Control Perturbation
localparam u_full_word_length = 55;
localparam u_round_word_length = 26;
localparam u_start = 29;

// Actuator Limits
localparam V_FP_MAX = 32767;
localparam V_FP_MIN = -32768;

// DAC
localparam DAC_MIN = 0;
localparam DAC_HALF = 32768;
localparam DAC_ZERO_CURRENT = 33190;
localparam DAC_MAX = 65535;

// RAM
localparam train_test_depth = 4000;
localparam train_test_depth_length = 12;
localparam desired_depth = 4200;
localparam desired_depth_length = 13;
localparam write_depth = 16000;
localparam write_depth_length = 14;

// CONTROL WAIT
localparam control_wait = 1500;
localparam train_test_wait = 1;
localparam total_wait_ctr_length = 12;
wire [total_wait_ctr_length - 1:0] total_wait/* synthesis keep */;

//******************************************MACROS******************************************//

// Truncate 'val' from 'start' to 'end' with rounding
`define round(val,start,end) $signed(val[start - 1] ? (val[start +: end] + 1'b1) : (val[start +: end]))
//`define round(val,start,end) $signed(val[start +: end])
`define trunc(val,start,end) $signed(val[start +: end])

//******************************************REG/WIRE DECLARATIONS******************************************//

// CLOCKS
wire clk_80MHz/* synthesis keep */;
wire clk_30MHz/* synthesis keep */;
wire locked/* synthesis keep */;
wire DAC_FLAG/* synthesis keep */;


// COUNTERS/STATE MACHINES
reg [6:0] clk_80_counter = 0/* synthesis noprune */;
localparam clk_80_counter_max = 79;
localparam start_write = 63;//64;//will have to be adjusted if DAC clock changes
localparam interrupt_write = 52;//53;//will have to be adjusted if DAC clock changes

// CONTROL WAIT
reg [total_wait_ctr_length - 1:0] total_wait_ctr = 0/* synthesis noprune */;
reg control_on = 0/* synthesis noprune */;

// SWITCH
wire control_switch/* synthesis keep */;


// ADC1
reg  		   cmd_valid/* synthesis noprune */;
wire [4:0]  cmd_channel /* synthesis keep */;
reg		   cmd_sop/* synthesis noprune */;
wire		   cmd_eop/* synthesis keep */;
wire		   cmd_rdy/* synthesis keep */;

wire        rspn_valid/* synthesis keep */;
wire [4:0]  rspn_ch/* synthesis keep */;
wire [11:0] rspn_data/* synthesis keep */;

reg  [16:0] lock_dly_cnt/* synthesis noprune */;
reg  [11:0] adc_1_data/* synthesis noprune */;
reg [11:0] adc_1_out = 0/* synthesis noprune */;

// ADC2
reg		   cmd_2_sop/* synthesis noprune */;
wire		   cmd_2_eop/* synthesis keep */;
wire		   cmd_2_rdy/* synthesis keep */;

wire        rspn_2_valid/* synthesis keep */;
wire [4:0]  rspn_2_ch/* synthesis keep */;
wire [11:0] rspn_2_data/* synthesis keep */;

reg  [11:0] adc_2_data/* synthesis noprune */;
reg [11:0] adc_2_out = 0/* synthesis noprune */;

// DAC
reg RESET = 0 /* synthesis noprune */; // setting reset high (although it might need to start low then go high)
reg [3:0] reset_counter = 0 /* synthesis noprune */;
reg [output_word_length - 1:0] DAC_DATA = DAC_ZERO_CURRENT /* synthesis noprune */; // start at halfway point, which is no perturbation
wire [7:0] CNT /* synthesis noprune */;

// BRAM READ
// V1 Desired
reg [desired_depth_length - 1:0] desired_address = 0/* synthesis noprune */;
wire signed [desired_word_length - 1:0] datain_V1/* synthesis keep */;
wire write_enable_V1/* synthesis keep */;
reg signed [desired_word_length - 1:0] desired_V1_out = 0/* synthesis noprune */;
wire signed [desired_word_length - 1:0] desired_V1_data/* synthesis keep */;
reg signed [desired_word_length - 1:0] desired_V1_data_delayed/* synthesis noprune */;
// U train test
reg [train_test_depth_length - 1:0] u_train_test_address = 0/* synthesis noprune */;
wire [output_word_length - 1:0] datain_u_train_test/* synthesis keep */;
wire write_enable_u_train_test/* synthesis keep */;
wire [output_word_length - 1:0] u_train_test_data/* synthesis keep */;

// BRAM WRITE
reg [write_depth_length - 1:0] write_address = 0/* synthesis noprune */;
reg write_enable_out=0/* synthesis noprune */;
// ADC 1
wire [input_word_length - 1:0] datain_adc_1/* synthesis keep */;
wire [input_word_length - 1:0] dataout_adc_1/* synthesis keep */;
// ADC 2
wire [input_word_length - 1:0] datain_adc_2/* synthesis keep */;
wire [input_word_length - 1:0] dataout_adc_2/* synthesis keep */;
// Desired V1
wire [desired_word_length - 1:0] datain_V1_out/* synthesis keep */;
wire [desired_word_length - 1:0] dataout_V1_out/* synthesis keep */;
// Control perturbation
wire [output_word_length - 1:0] datain_control_perturbation/* synthesis keep */;
wire [output_word_length - 1:0] dataout_control_perturbation/* synthesis keep */;
// Prediction
wire [prediction_trunc_word_length - 1:0] datain_prediction/* synthesis keep */;
wire [prediction_trunc_word_length - 1:0] dataout_prediction/* synthesis keep */;

// K00
wire K00_address = 0;
wire signed [weight_word_length - 1:0] datain_K00/* synthesis keep */;
wire signed [weight_word_length - 1:0] K00/* synthesis keep */;
wire write_enable_K00/* synthesis keep */;
// Mode switch (mode=0 is control, mode=1 is train/test)
wire mode_address = 0/* synthesis keep */;
wire datain_mode/* synthesis keep */;
wire write_enable_mode/* synthesis keep */;
wire mode/* synthesis keep */;
// Control on switch (switch=0 is control off, switch=1 is control on)
wire switch_address = 0/* synthesis keep */;
wire datain_switch/* synthesis keep */;
wire write_enable_switch/* synthesis keep */;
wire switch/* synthesis keep */;
// Delay
wire delay_address = 0/* synthesis keep */;
wire [5-1:0] datain_delay/* synthesis keep */;
wire write_enable_delay/* synthesis keep */;
wire [5-1:0] delay/* synthesis keep */;
wire [5-1:0] delay_flag/* synthesis keep */; // technically should be one bit larger than 5...
// Feedback linearization
wire fl_address = 0/* synthesis keep */;
wire datain_fl/* synthesis keep */;
wire write_enable_fl/* synthesis keep */;
wire fl/* synthesis keep */;
localparam fl_padding_width = weight_fractional_bits;//WIP
wire [fl_padding_width-1:0] fl_padding = 0;/* synthesis keep */

// LINEAR FEATURES
reg signed [output_word_length - 1:0] u_i_SM1_16 = 0/* synthesis noprune */;
reg signed [output_word_length - 1:0] u_im1_SM1_16 = 0/* synthesis noprune */;
reg signed [input_word_length - 1:0] V1_i_SM1_12 = 0/* synthesis noprune */;
reg signed [input_word_length - 1:0] V2_i_SM1_12 = 0/* synthesis noprune */;
reg signed [input_word_length - 1:0] V1_im1_SM1_12 = 0/* synthesis noprune */;
reg signed [input_word_length - 1:0] V2_im1_SM1_12 = 0/* synthesis noprune */;
reg signed [difference_word_length - 1:0] V1mV2_i_S0_12 = 0/* synthesis noprune */;
reg signed [difference_word_length - 1:0] V1mV2_im1_S0_12 = 0/* synthesis noprune */;

// LINEAR FEATURES S0_17
reg signed [linear_word_length - 1:0] u_i_SM1_17 = 0/* synthesis noprune */;
reg signed [linear_word_length - 1:0] u_im1_SM1_17 = 0/* synthesis noprune */;
reg signed [linear_word_length - 1:0] V1_i_SM1_17 = 0/* synthesis noprune */;
reg signed [linear_word_length - 1:0] V2_i_SM1_17 = 0/* synthesis noprune */;
reg signed [linear_word_length - 1:0] V1_im1_SM1_17 = 0/* synthesis noprune */;
reg signed [linear_word_length - 1:0] V2_im1_SM1_17 = 0/* synthesis noprune */;
reg signed [feature_word_length - 1:0] V1mV2_i_S0_17 = 0/* synthesis noprune */;
reg signed [feature_word_length - 1:0] V1mV2_im1_S0_17 = 0/* synthesis noprune */;

// QUADRATIC FEATURES
reg signed [quadratic_full_word_length - 1:0] V1mV2_i_squared_S1_24 = 0/* synthesis noprune */;
reg signed [quadratic_round_word_length - 1:0] V1mV2_i_squared_S1_16 = 0/* synthesis noprune */; // 1 int bit only covers case of most negative times most negative
reg signed [cubic_full_word_length - 1:0] V1mV2_i_cubed_S0_28 = 0/* synthesis noprune */; // very unsure about this representation

// CUBIC FEATURES
reg signed [cubic_round_word_length - 1:0] V1mV2_i_cubed_S0_17 = 0/* synthesis noprune */;
reg signed [cubic_round_word_length - 1:0] V1mV2_im1_cubed_S0_17 = 0/* synthesis noprune */;

// NEW FEATURES
reg signed [24-1:0] V1_im1_squared_SM1_24 = 0/* synthesis noprune */;
reg signed [18-1:0] V1_im1_squared_SM1_18 = 0/* synthesis noprune */;
reg signed [24-1:0] V1_i_squared_SM1_24 = 0/* synthesis noprune */;
reg signed [18-1:0] V1_i_squared_SM1_18 = 0/* synthesis noprune */;

reg signed [31-1:0] Nonlin1_S0_30 = 0/* synthesis noprune */;
reg signed [31-1:0] Nonlin2_S0_30 = 0/* synthesis noprune */;
reg signed [feature_word_length-1:0] Nonlin1_S0_17 = 0/* synthesis noprune */;
reg signed [feature_word_length-1:0] Nonlin2_S0_17 = 0/* synthesis noprune */;

// DESIRED
reg signed [desired_word_length - 1:0] desired_V1_ip1_SM1_16 = 0/* synthesis noprune */;
reg signed [desired_word_length - 1:0] desired_V2_ip1_SM1_16 = 0/* synthesis noprune */;
reg signed [desired_word_length - 1:0] desired_V1_i_SM1_16 = 0/* synthesis noprune */;
reg signed [desired_word_length - 1:0] desired_V2_i_SM1_16 = 0/* synthesis noprune */;
reg signed [desired_padded_word_length - 1:0] desired_V1_ip1_SM1_33 = 0/* synthesis noprune */;
reg signed [desired_padded_word_length - 1:0] desired_V2_ip1_SM1_33 = 0/* synthesis noprune */;
reg signed [desired_padded_word_length - 1:0] desired_V1_i_SM1_33 = 0/* synthesis noprune */;
reg signed [desired_padded_word_length - 1:0] desired_V2_i_SM1_33 = 0/* synthesis noprune */;

// ERROR
reg signed [error_word_length - 1:0] error_V1_i_S0_17 = 0/* synthesis noprune */;
reg signed [error_word_length - 1:0] error_V2_i_S0_17 = 0/* synthesis noprune */;

//Weights (trained on 04/02/2024 07:15 PM)
//ridge parameter: 2.1544346900318823e-05
//weight fractional bits: 15
wire signed [weight_word_length - 1:0] W_u_i_inv = -83757/* synthesis keep */; //-10.2242431640625
wire signed [weight_word_length - 1:0] W_u_i = -3205/* synthesis keep */; //-0.097808837890625
wire signed [weight_word_length - 1:0] W_u_im1 = 2929/* synthesis keep */; //0.089385986328125
wire signed [weight_word_length - 1:0] W_V1_i = 58816/* synthesis keep */; //1.794921875
wire signed [weight_word_length - 1:0] W_V2_i = 1963/* synthesis keep */; //0.059906005859375
wire signed [weight_word_length - 1:0] W_V1_im1 = -24707/* synthesis keep */; //-0.753997802734375
wire signed [weight_word_length - 1:0] W_V2_im1 = -1382/* synthesis keep */; //-0.04217529296875
wire signed [weight_word_length - 1:0] W_V1mV2_i = -31290/* synthesis keep */; //-0.95489501953125
wire signed [weight_word_length - 1:0] W_V1mV2_im1 = 5914/* synthesis keep */; //0.18048095703125
wire signed [weight_word_length - 1:0] W_V1mV2_i_cubed = -14211/* synthesis keep */; //-0.433685302734375
wire signed [weight_word_length - 1:0] W_V1mV2_im1_cubed = 15206/* synthesis keep */; //0.46405029296875

// K
//wire signed [weight_word_length - 1:0] K00 = -98304/* synthesis keep */; //-3.0
wire signed [weight_word_length - 1:0] K01 = 0/* synthesis keep */; //0.0
reg signed [Kerror_word_length  - 1:0] Kerror = 0/* synthesis noprune */;

// Prediction
reg signed [unperturbed_prediction_word_length - 1:0] unperturbed_prediction = 0/* synthesis noprune */;
reg signed [prediction_word_length - 1:0] prediction = 0/* synthesis noprune */;
reg signed [prediction_trunc_word_length - 1:0] prediction_SM1_16/* synthesis noprune */;

// Control Perturbation
reg signed [u_full_word_length - 1:0] u_i_full = 0/* synthesis noprune */;
reg signed [u_round_word_length - 1:0] u_i_Sx_16 = 0/* synthesis noprune */;


reg write_flag = 0/* synthesis noprune */;

//******************************************MODULE INSTANTIATION******************************************//

// DUAL ADC
	// MAX ADC CLOCK: 80MHz
	// MAX IP CLOCK: ~95MHz
dual_adc my_dual_adc (
	.clock_clk                (clk_80MHz),      //          clock.clk (IP CLOCK)
	.reset_sink_reset_n       (KEY[0]),       //     reset_sink.reset_n
	.adc_pll_clock_clk        (clk_80MHz),      //  adc_pll_clock.clk (ADC CLOCK)
	.adc_pll_locked_export    (locked),  // adc_pll_locked.export
	.command_valid            (cmd_valid),    //        command.valid
	.command_channel          (cmd_channel),  //               .channel
	.command_startofpacket    (cmd_sop),      //               .startofpacket
	.command_endofpacket      (cmd_eop),      //               .endofpacket
	.command_ready            (cmd_rdy),      //               .ready
	.command_2_valid          (cmd_valid),  //      command_2.valid
	.command_2_channel        (cmd_channel),//               .channel
	.command_2_startofpacket  (cmd_2_sop),    //               .startofpacket
	.command_2_endofpacket    (cmd_2_eop),    //               .endofpacket
	.command_2_ready          (cmd_2_rdy),    //               .ready
	.response_valid           (rspn_valid),   //       response.valid
	.response_channel         (rspn_ch),      //               .channel
	.response_data            (rspn_data),    //               .data
	.response_startofpacket   (),             //               .startofpacket
	.response_endofpacket     (),             //               .endofpacket
	.response_2_valid         (rspn_2_valid), //     response_2.valid
	.response_2_channel       (rspn_2_ch),    //               .channel
	.response_2_data          (rspn_2_data),  //               .data
	.response_2_startofpacket (),             //               .startofpacket
	.response_2_endofpacket   ()              //               .endofpacket
);
	
	
// PLL
pll pll1(
	.inclk0(MAX10_CLK1_50), // 50 MHz input clock
	.c0(clk_80MHz), // 80 MHz output clock with 0 phase shift and 50% duty cycle
	.c1(clk_30MHz), // 30 MHz output clock with -50.62 phase shift and 50% duty cycle
	.locked(locked)
	);
	
// DAC (FAST VERSION)
dac_fast mydac(
    .CLK( clk_30MHz ), // 30 MHz output clock with -50.62 phase shift and 50% duty cycle
    .RESET( RESET ), // active low, raises SYNC to interrupt write sequence
    .DAC_DATA( DAC_DATA ), // 16-bit data from user to convert to analog signal
    .SYNC( SYNC ), // frame synchronization signal for the input data
    .SCLK( SCLK ), // Serial clock input
    .DIN( DIN ), // Serial data input
    .CNT2( CNT ), // counter for SCLK
	 .DAC_FLAG( DAC_FLAG ) // Signal indicating that the 16 data bits are being written
    );
	 
// RAM

// READ

// Desired V1
desired_V1 mydesired_V1 (
	desired_address,
	clk_80MHz,
	datain_V1,
	write_enable_V1,
	desired_V1_data);
	
// u train test
u_train_test my_u_train_test (
	u_train_test_address,
	clk_80MHz,
	datain_u_train_test,
	write_enable_u_train_test,
	u_train_test_data);
	
// WRITE
// ADC_1_OUT
adc_1_ram myadc_1_ram (
	write_address,
	clk_80MHz,
	datain_adc_1,
	write_enable_out,
	dataout_adc_1);
	
// ADC_2_OUT
adc_2_ram myadc_2_ram (
	write_address,
	clk_80MHz,
	datain_adc_2,
	write_enable_out,
	dataout_adc_2);

// DESIRED_V1_OUT
desired_V1_out mydesired_V1_out (
	write_address,
	clk_80MHz,
	datain_V1_out,
	write_enable_out,
	dataout_V1_out);

// CONTROL_PERTURBATION
control_perturbation mycontrol_perturbation (
	write_address,
	clk_80MHz,
	datain_control_perturbation,
	write_enable_out,
	dataout_control_perturbation);

// PREDICTION
prediction myprediction (
	write_address,
	clk_80MHz,
	datain_prediction,
	write_enable_out,
	dataout_prediction);

// K00
K00 myK00 (
	K00_address,
	clk_80MHz,
	datain_K00,
	write_enable_K00,
	K00);

// collect vs. control switch
mode mymode (
	mode_address,
	clk_80MHz,
	datain_mode,
	write_enable_mode,
	mode);

// control on SWITCH
switch myswitch (
	switch_address,
	clk_80MHz,
	datain_switch,
	write_enable_switch,
	switch);

// delay control
// delay == 0 -> no delay
// delay == sample_divider -> maximum_delay
delay mydelay (
	delay_address,
	clk_80MHz,
	datain_delay,
	write_enable_delay,
	delay); // 5-bits -> 0-31, should not exceed sample_divider
	
// feedback linearization control
fl myfl (
	fl_address,
	clk_80MHz,
	datain_fl,
	write_enable_fl,
	fl);
		
//******************************************STRUCTURAL CODING******************************************//

// WIP: Should test if I even need this! Probably causes some issues
// ADC LOCK
always@(posedge clk_80MHz)	
	if ( locked && lock_dly_cnt != 80000 )
		lock_dly_cnt <= lock_dly_cnt + 1;

wire lock_rdy = (lock_dly_cnt == 80000)? 1 : 0;	

assign cmd_channel = 0;

// ADC COMMAND VALID
always@(posedge clk_80MHz)
begin
	cmd_valid <= lock_rdy;
end

//******************************************MAIN CODE******************************************//

// if the virtual switch is off, don't do control and send zero current out of the device
assign control_switch = switch;

// if the mode is 0 for applying the training and testing perturbation, 
// set the total wait to only 1 cycle, if the mode is set to 1 for the control algorithm,
// set the total wait to 1500 points
assign total_wait = mode ? train_test_wait : control_wait;

// setting memory inputs
assign datain_adc_1 = adc_1_data;
assign datain_adc_2 = adc_2_data;
assign datain_V1_out = desired_V1_i_SM1_16;
assign datain_control_perturbation = DAC_DATA;
assign datain_prediction = prediction_SM1_16;

// Turn off 7 segment displays
assign HEX0[6:0] = 7'b1111111;
assign HEX1[6:0] = 7'b1111111;

// ADC buffer used to implement added delay
localparam adc_buffer_length = 32;
reg [input_word_length-1:0] adc_1_buffer [adc_buffer_length-1:0];/* synthesis noprune */
reg [input_word_length-1:0] adc_2_buffer [adc_buffer_length-1:0];/* synthesis noprune */

// to decrease sample rate from 1MHZ ADC
localparam SAMPLE_DIVIDER = 5;//10 for 100KHz 5 for 200KHz
reg [4-1:0] sample_counter = 1/* synthesis noprune */;

integer i;

always @ (posedge clk_80MHz)
begin
	if (rspn_valid)
	begin
		// store incoming ADC data in buffer when rspn valid is high
		for (i=0; i <= adc_buffer_length-2; i = i+1) begin
			adc_1_buffer[adc_buffer_length-1-i] = adc_1_buffer[adc_buffer_length-2-i];
			adc_2_buffer[adc_buffer_length-1-i] = adc_2_buffer[adc_buffer_length-2-i];
		end
		adc_1_buffer[0] = rspn_data;
		adc_2_buffer[0] = rspn_2_data;

		// set the current ADC data to delayed value. If delay == 0, then there is
		// no added delay
		adc_1_data = adc_1_buffer[delay];
		adc_2_data = adc_2_buffer[delay];
		
		// if the sample counter is less than sample divider, increment sample counter
		if (sample_counter < SAMPLE_DIVIDER)
		begin
			sample_counter = sample_counter + 1;
		end
		// else, reset sample counter and clock counter.
		// clock counter does not reset when it reaches its 
		// maximum value, only when the sample counter does. This
		// ensures that the control algorithm only operates each time
		// sample divider resets.
		else
		begin
			sample_counter = 1;
			clk_80_counter = 0;
		end
		
	end
	
	// if the clock counter is less than the counter max, increment the counter
	else if (clk_80_counter < clk_80_counter_max)
		clk_80_counter = clk_80_counter + 1;
		
	case(clk_80_counter)
	0:
		begin
			// V1i, V1im1
			V1_im1_SM1_12 = V1_i_SM1_12;
			V1_im1_SM1_17 = V1_i_SM1_17;
			V1_i_SM1_12 = adc_1_data - 2048;
			V1_i_SM1_17 = $signed({V1_i_SM1_12, input_padding});
			
			// V2i, V2im1
			V2_im1_SM1_12 = V2_i_SM1_12;
			V2_im1_SM1_17 = V2_i_SM1_17;
			V2_i_SM1_12 = adc_2_data - 2048;
			V2_i_SM1_17 = $signed({V2_i_SM1_12, input_padding});
			
			// uim1
			u_im1_SM1_17 = u_i_SM1_17; // not used at moment. Will be needed in full prediction
			
			// desired
			desired_V1_i_SM1_16 = desired_V1_ip1_SM1_16;
			desired_V2_i_SM1_16 = desired_V2_ip1_SM1_16;
			desired_V1_ip1_SM1_16 = desired_V1_data;
			desired_V2_ip1_SM1_16 = 0;
			desired_V1_ip1_SM1_33 = $signed({desired_V1_ip1_SM1_16,desired_padding});
			desired_V2_ip1_SM1_33 = $signed({desired_V2_ip1_SM1_16,desired_padding});
		
			// error
			error_V1_i_S0_17 = $signed({V1_i_SM1_12,input_padding}) - $signed({desired_V1_i_SM1_16,desired_error_padding});
			error_V2_i_S0_17 = $signed({V2_i_SM1_12,input_padding}) - $signed({desired_V1_i_SM1_16,desired_error_padding});	
			
			// V1im1-V2im1
			V1mV2_im1_S0_12 = V1mV2_i_S0_12;
			V1mV2_im1_S0_17 = V1mV2_i_S0_17;
			
			// V1i-V2i
			V1mV2_i_S0_12 = V1_i_SM1_12 - V2_i_SM1_12;
			V1mV2_i_S0_17 = $signed({V1mV2_i_S0_12, input_padding});
			
			// (V1i-V2i)^2
			V1mV2_i_squared_S1_24 = V1mV2_i_S0_12*V1mV2_i_S0_12;
			V1mV2_i_squared_S1_16 = `round(V1mV2_i_squared_S1_24,quadratic_start,quadratic_round_word_length);
			
			// V1im1^2
			V1_im1_squared_SM1_24 = V1_i_squared_SM1_24;
			V1_im1_squared_SM1_18 = V1_i_squared_SM1_18;
			
			// V1i^2
			V1_i_squared_SM1_24 = V1_i_SM1_12 * V1_i_SM1_12;
			V1_i_squared_SM1_18 = `round(V1_i_squared_SM1_24,6,18);
			
		end
	1:
		begin		
			// (V1im1-V2im1)^3
			V1mV2_im1_cubed_S0_17 = V1mV2_i_cubed_S0_17;
			
			// (V1i-V2i)^3
			V1mV2_i_cubed_S0_28 = V1mV2_i_S0_12*V1mV2_i_squared_S1_16;
			V1mV2_i_cubed_S0_17 = `round(V1mV2_i_cubed_S0_28,cubic_start,cubic_round_word_length);

			// (V1i-V2i)*V1im1^2
			Nonlin1_S0_30 = V1mV2_i_S0_12*V1_im1_squared_SM1_18;
			Nonlin1_S0_17 = `round(Nonlin1_S0_30,13,18);
			
			// (V1im1-V2im1)*V1i^2
			Nonlin2_S0_30 = V1mV2_im1_S0_12*V1_i_squared_SM1_18;
			Nonlin2_S0_17 = `round(Nonlin2_S0_30,13,18);
		end	
	2:
		begin
			// if the feedback linearization is on, all nonlinear terms are used (NG-RC)
			if (fl)
				// Fi+1 = WF.OF in control algorithm
				unperturbed_prediction = u_im1_SM1_17*W_u_im1 + V1_i_SM1_17*W_V1_i + V2_i_SM1_17*W_V2_i + V1_im1_SM1_17*W_V1_im1 + V2_im1_SM1_17*W_V2_im1 + Nonlin1_S0_17*W_V1mV2_i + Nonlin2_S0_17*W_V1mV2_im1 + V1mV2_i_cubed_S0_17*W_V1mV2_i_cubed + V1mV2_im1_cubed_S0_17*W_V1mV2_im1_cubed;
			// if the feedback linearization is off, only linear terms are used
			else
				// Fi+1 = V1i in control algorithm
				unperturbed_prediction = $signed({V1_i_SM1_17,fl_padding});//WIP
			// K*ei in control algorithm
			Kerror = K00*error_V1_i_S0_17 + K01*error_V2_i_S0_17;
		end
	3:
		begin
			// ui in control algorithm, in full precision
			u_i_full = W_u_i_inv*(desired_V1_ip1_SM1_33 - unperturbed_prediction + Kerror);
		end		
	4:
		begin
			// ui rounded to 16 fractional bits
			u_i_Sx_16 = `round(u_i_full,u_start,u_round_word_length); // depends on W_u and W_u_inv
			// ui thresholded to 16 bits
			u_i_SM1_16 = (u_i_Sx_16 > V_FP_MAX) ? V_FP_MAX: // if it is above the maximum, set it to the maximum
                      (u_i_Sx_16 < V_FP_MIN) ? V_FP_MIN: // if it is below the minimum, set it to the minimum			
							`trunc(u_i_Sx_16,0,output_word_length); // if it is in the middle, truncate it to 16 bits
			
			// if the mode is 1, apply the training and testing perturbation
			if (mode)
				DAC_DATA = control_on ? u_train_test_data : DAC_ZERO_CURRENT;
			// if the mode is 0, apply the control perturbation
			else
				// If the control is on, invert the first bit of the ui output to make it unsigned,
				// else set the output to the DAC value that gives zero current
				DAC_DATA = control_on ? {~u_i_SM1_16[15],u_i_SM1_16[14:0]} : DAC_ZERO_CURRENT;//DAC_HALF;
			
			// If the physical switch 0 is on, overwrite the DAC value to be at half
			if (SW[0])
				DAC_DATA = DAC_HALF;
		end
	6:
		begin
			// If mode is 0, padding ui to 18 bits so it can be delayed to uim1 to be used in features,
			// or in full prediction. If the mode is 1, invert the first bit of the DAC value and pad 
			// to 18 bits.
			u_i_SM1_17 = mode ? $signed({~DAC_DATA[15],DAC_DATA[14:0],1'b0}): $signed({u_i_SM1_16,1'b0});
			// prediction in full precision, W.Ototal or WF.Of + Wu.ui in control algorithm 
			prediction = unperturbed_prediction + W_u_i*u_i_SM1_17;
			// round prediction to have 16 fractional bits
			prediction_SM1_16 = `round(prediction,pred_start,prediction_trunc_word_length);
		end
	
	7:
		begin
			// Control Wait Circuit
			// After the virtual switch is turned on, wait 
			// total_wait steps until control algorithm is activated
			if (control_switch)
			begin
				if (total_wait_ctr < total_wait - 1)
					total_wait_ctr = total_wait_ctr + 1;
				else
					control_on = 1;
			end
			else
			begin
					control_on = 0;
					total_wait_ctr = 0;
			end
			
			// Increment Desired Read Circuit
			if (control_on && (desired_address < desired_depth - 1))
			begin
				desired_address = desired_address + 1;
			end
			else
				desired_address = 0;	
				
			// Increment U Train Test Read Circuit
			if (control_on && (u_train_test_address < train_test_depth - 1))
			begin
				u_train_test_address = u_train_test_address + 1;
			end
			else
				u_train_test_address = 0;	
			
			// memory write circuit 1
			// if write flag is high, increment write address
			if (write_flag)
			begin
				if (write_address < write_depth - 1)
				begin
					write_address = write_address + 1;
				end
			end
			// if write flag is low, reset write address
			else
			begin
				write_address = 0;
			end				
		end
	
	9:	
		// memory write circuit 2
		begin
			// if the control switch is on and the write address is less than
			// the maximum, enable write and set write flag to high
			if (control_switch && (write_address < write_depth - 1))
			begin
				write_enable_out = 1;
				write_flag = 1;
			end
			// else, disable write
			else
			begin
				write_enable_out = 0;
			end
			
			// if the control switch is off, disable write flag
			if (!control_switch)
			begin
				write_flag = 0;
			end
		end
	
	interrupt_write:
		begin
			RESET = 0;
		end
		
	start_write:
	begin
		RESET = 1;
	end	
	
	endcase
end
	
endmodule
