//////////////////////////////////////////////////////////////////////////////////
// Created on Wednesday March 27 17:10 2024
//
// Description:
// Used to control the double scroll circuit using the control algorithm with
// nonlinear terms, and m=1 step ahead predictions. This code is used to estimate
// the resource utilization and the power consumption of the device during control.
// The desired signal is fixed to the origin task.
// 
//
// Notes:
//
// @author: Robert Kent
//////////////////////////////////////////////////////////////////////////////////

module dual_adc_to_dac(

	//////////// CLOCK //////////
	input 		          		MAX10_CLK1_50,

	//////////// KEY //////////
	input 		          		FPGA_RESET_n,
	input 		     [4:0]		KEY,

	//////////// SW //////////
	input 		     [9:0]		SW,

	//////////// LEDR //////////
	output		     [9:0]		LEDR,

	//////////// HEX //////////
	output		     [6:0]		HEX0,
	output		     [6:0]		HEX1,

	//////////// DAC //////////
	inout 		          		DIN,
	output		          		SCLK,
	output		          		SYNC
);

//******************************************LOCAL PARAMETERS******************************************//

//Weights (trained on 04/02/2024 07:15 PM)
// Fixed-Point Parameters
// Linear Features
localparam input_word_length = 12;
localparam difference_word_length = 13;
localparam output_word_length = 16;
localparam linear_word_length = 17;
localparam feature_word_length = 18;
reg [5 - 1:0] input_padding = 0;

// Quadratic Features
localparam quadratic_full_word_length = 26;
localparam quadratic_round_word_length = 18;
localparam quadratic_start = 8;

// Cubic Features
localparam cubic_full_word_length = 29;
localparam cubic_round_word_length = 18;
localparam cubic_start = 11;

// Desired
localparam desired_word_length = 16;
localparam desired_padded_word_length = 32;
reg [16 - 1:0] desired_padding = 0;
reg [1 - 1:0] desired_error_padding = 0;

// Error
localparam error_word_length = 18;

// Weights
localparam weight_word_length = 18;
localparam Kerror_word_length = 36;

// Prediction
localparam unperturbed_prediction_word_length = 38;
localparam prediction_word_length = 38;
localparam pred_start = 16;
localparam prediction_trunc_word_length = 16;

// Control Perturbation
localparam u_full_word_length = 55;
localparam u_round_word_length = 26;
localparam u_start = 29;

// Actuator Limits
localparam V_FP_MAX = 32767;
localparam V_FP_MIN = -32768;

// DAC
localparam DAC_ZERO_CURRENT = 33190;

// CONTROL WAIT
wire [1 - 1:0] total_wait = 1/* synthesis keep */;
//******************************************MACROS******************************************//

// Truncate 'val' from 'start' to 'end' with rounding
`define round(val,start,end) $signed(val[start - 1] ? (val[start +: end] + 1'b1) : (val[start +: end]))
//`define round(val,start,end) $signed(val[start +: end])
`define trunc(val,start,end) $signed(val[start +: end])

//******************************************REG/WIRE DECLARATIONS******************************************//

// CLOCKS
wire clk_80MHz/* synthesis keep */;
wire clk_30MHz/* synthesis keep */;
wire locked/* synthesis keep */;
wire DAC_FLAG/* synthesis keep */;

// COUNTERS/STATE MACHINES
reg [6:0] clk_80_counter = 0/* synthesis noprune */;
localparam clk_80_counter_max = 79;
localparam start_write = 63;//64;//will have to be adjusted if DAC clock changes
localparam interrupt_write = 52;//53;//will have to be adjusted if DAC clock changes

// CONTROL WAIT
reg [1 - 1:0] total_wait_ctr = 0/* synthesis noprune */;
reg control_on = 0/* synthesis noprune */;

// SWITCH
wire control_switch/* synthesis keep */;


// ADC1
reg  		   cmd_valid/* synthesis noprune */;
wire [4:0]  cmd_channel /* synthesis keep */;
reg		   cmd_sop/* synthesis noprune */;
wire		   cmd_eop/* synthesis keep */;
wire		   cmd_rdy/* synthesis keep */;

wire        rspn_valid/* synthesis keep */;
wire [4:0]  rspn_ch/* synthesis keep */;
wire [11:0] rspn_data/* synthesis keep */;

reg  [16:0] lock_dly_cnt/* synthesis noprune */;
reg  [11:0] adc_1_data/* synthesis noprune */;
reg [11:0] adc_1_out = 0/* synthesis noprune */;

// ADC2
reg		   cmd_2_sop/* synthesis noprune */;
wire		   cmd_2_eop/* synthesis keep */;
wire		   cmd_2_rdy/* synthesis keep */;

wire        rspn_2_valid/* synthesis keep */;
wire [4:0]  rspn_2_ch/* synthesis keep */;
wire [11:0] rspn_2_data/* synthesis keep */;

reg  [11:0] adc_2_data/* synthesis noprune */;
reg [11:0] adc_2_out = 0/* synthesis noprune */;

// DAC
reg RESET = 0 /* synthesis noprune */; // setting reset high (although it might need to start low then go high)
reg [3:0] reset_counter = 0 /* synthesis noprune */;
reg [output_word_length - 1:0] DAC_DATA = DAC_ZERO_CURRENT /* synthesis noprune */; // start at halfway point, which is no perturbation
wire [7:0] CNT /* synthesis noprune */;

// BRAM READ
// V1 Desired
reg signed [desired_word_length - 1:0] desired_V1_data = 0/* synthesis noprune */;
reg signed [desired_word_length - 1:0] desired_V1_data_delayed/* synthesis noprune */;

// LINEAR FEATURES
reg signed [output_word_length - 1:0] u_i_SM1_16 = 0/* synthesis noprune */;
reg signed [output_word_length - 1:0] u_im1_SM1_16 = 0/* synthesis noprune */;
reg signed [input_word_length - 1:0] V1_i_SM1_12 = 0/* synthesis noprune */;
reg signed [input_word_length - 1:0] V2_i_SM1_12 = 0/* synthesis noprune */;
reg signed [input_word_length - 1:0] V1_im1_SM1_12 = 0/* synthesis noprune */;
reg signed [input_word_length - 1:0] V2_im1_SM1_12 = 0/* synthesis noprune */;
reg signed [difference_word_length - 1:0] V1mV2_i_S0_12 = 0/* synthesis noprune */;
reg signed [difference_word_length - 1:0] V1mV2_im1_S0_12 = 0/* synthesis noprune */;

// LINEAR FEATURES S0_17
reg signed [linear_word_length - 1:0] u_i_SM1_17 = 0/* synthesis noprune */;
reg signed [linear_word_length - 1:0] u_im1_SM1_17 = 0/* synthesis noprune */;
reg signed [linear_word_length - 1:0] V1_i_SM1_17 = 0/* synthesis noprune */;
reg signed [linear_word_length - 1:0] V2_i_SM1_17 = 0/* synthesis noprune */;
reg signed [linear_word_length - 1:0] V1_im1_SM1_17 = 0/* synthesis noprune */;
reg signed [linear_word_length - 1:0] V2_im1_SM1_17 = 0/* synthesis noprune */;
reg signed [feature_word_length - 1:0] V1mV2_i_S0_17 = 0/* synthesis noprune */;
reg signed [feature_word_length - 1:0] V1mV2_im1_S0_17 = 0/* synthesis noprune */;

// QUADRATIC FEATURES
reg signed [quadratic_full_word_length - 1:0] V1mV2_i_squared_S1_24 = 0/* synthesis noprune */;
reg signed [quadratic_round_word_length - 1:0] V1mV2_i_squared_S1_16 = 0/* synthesis noprune */; // 1 int bit only covers case of most negative times most negative
reg signed [cubic_full_word_length - 1:0] V1mV2_i_cubed_S0_28 = 0/* synthesis noprune */; // very unsure about this representation

// CUBIC FEATURES
reg signed [cubic_round_word_length - 1:0] V1mV2_i_cubed_S0_17 = 0/* synthesis noprune */;
reg signed [cubic_round_word_length - 1:0] V1mV2_im1_cubed_S0_17 = 0/* synthesis noprune */;

// NEW FEATURES
reg signed [24-1:0] V1_im1_squared_SM1_24 = 0/* synthesis noprune */;
reg signed [18-1:0] V1_im1_squared_SM1_18 = 0/* synthesis noprune */;
reg signed [24-1:0] V1_i_squared_SM1_24 = 0/* synthesis noprune */;
reg signed [18-1:0] V1_i_squared_SM1_18 = 0/* synthesis noprune */;

reg signed [31-1:0] Nonlin1_S0_30 = 0/* synthesis noprune */;
reg signed [31-1:0] Nonlin2_S0_30 = 0/* synthesis noprune */;
reg signed [feature_word_length-1:0] Nonlin1_S0_17 = 0/* synthesis noprune */;
reg signed [feature_word_length-1:0] Nonlin2_S0_17 = 0/* synthesis noprune */;

// DESIRED
reg signed [desired_word_length - 1:0] desired_V1_ip1_SM1_16 = 0/* synthesis noprune */;
reg signed [desired_word_length - 1:0] desired_V1_i_SM1_16 = 0/* synthesis noprune */;
reg signed [desired_padded_word_length - 1:0] desired_V1_ip1_SM1_33 = 0/* synthesis noprune */;


// ERROR
reg signed [error_word_length - 1:0] error_V1_i_S0_17 = 0/* synthesis noprune */;

//Weights (trained on 04/02/2024 07:15 PM)
//ridge parameter: 2.1544346900318823e-05
//weight fractional bits: 15
wire signed [weight_word_length - 1:0] W_u_i_inv = -83757/* synthesis keep */; //-10.2242431640625
wire signed [weight_word_length - 1:0] W_u_i = -3205/* synthesis keep */; //-0.097808837890625
wire signed [weight_word_length - 1:0] W_u_im1 = 2929/* synthesis keep */; //0.089385986328125
wire signed [weight_word_length - 1:0] W_V1_i = 58816/* synthesis keep */; //1.794921875
wire signed [weight_word_length - 1:0] W_V2_i = 1963/* synthesis keep */; //0.059906005859375
wire signed [weight_word_length - 1:0] W_V1_im1 = -24707/* synthesis keep */; //-0.753997802734375
wire signed [weight_word_length - 1:0] W_V2_im1 = -1382/* synthesis keep */; //-0.04217529296875
wire signed [weight_word_length - 1:0] W_V1mV2_i = -31290/* synthesis keep */; //-0.95489501953125
wire signed [weight_word_length - 1:0] W_V1mV2_im1 = 5914/* synthesis keep */; //0.18048095703125
wire signed [weight_word_length - 1:0] W_V1mV2_i_cubed = -14211/* synthesis keep */; //-0.433685302734375
wire signed [weight_word_length - 1:0] W_V1mV2_im1_cubed = 15206/* synthesis keep */; //0.46405029296875

// K
wire signed [weight_word_length - 1:0] K00 = 26214/* synthesis keep */; //0.8
reg signed [Kerror_word_length  - 1:0] Kerror = 0/* synthesis noprune */;

// Prediction
reg signed [unperturbed_prediction_word_length - 1:0] unperturbed_prediction = 0/* synthesis noprune */;

// Control Perturbation
reg signed [u_full_word_length - 1:0] u_i_full = 0/* synthesis noprune */;
reg signed [u_round_word_length - 1:0] u_i_Sx_16 = 0/* synthesis noprune */;


reg write_flag = 0/* synthesis noprune */;

//******************************************MODULE INSTANTIATION******************************************//

// DUAL ADC
	// MAX ADC CLOCK: 80MHz
	// MAX IP CLOCK: ~95MHz
dual_adc my_dual_adc (
	.clock_clk                (clk_80MHz),      //          clock.clk (IP CLOCK)
	.reset_sink_reset_n       (KEY[0]),       //     reset_sink.reset_n
	.adc_pll_clock_clk        (clk_80MHz),      //  adc_pll_clock.clk (ADC CLOCK)
	.adc_pll_locked_export    (locked),  // adc_pll_locked.export
	.command_valid            (cmd_valid),    //        command.valid
	.command_channel          (cmd_channel),  //               .channel
	.command_startofpacket    (cmd_sop),      //               .startofpacket
	.command_endofpacket      (cmd_eop),      //               .endofpacket
	.command_ready            (cmd_rdy),      //               .ready
	.command_2_valid          (cmd_valid),  //      command_2.valid
	.command_2_channel        (cmd_channel),//               .channel
	.command_2_startofpacket  (cmd_2_sop),    //               .startofpacket
	.command_2_endofpacket    (cmd_2_eop),    //               .endofpacket
	.command_2_ready          (cmd_2_rdy),    //               .ready
	.response_valid           (rspn_valid),   //       response.valid
	.response_channel         (rspn_ch),      //               .channel
	.response_data            (rspn_data),    //               .data
	.response_startofpacket   (),             //               .startofpacket
	.response_endofpacket     (),             //               .endofpacket
	.response_2_valid         (rspn_2_valid), //     response_2.valid
	.response_2_channel       (rspn_2_ch),    //               .channel
	.response_2_data          (rspn_2_data),  //               .data
	.response_2_startofpacket (),             //               .startofpacket
	.response_2_endofpacket   ()              //               .endofpacket
);
	
	
// PLL
pll pll1(
	.inclk0(MAX10_CLK1_50), // 50 MHz input clock
	.c0(clk_80MHz), // 80 MHz output clock with 0 phase shift and 50% duty cycle
	.c1(clk_30MHz), // 30 MHz output clock with -50.62 phase shift and 50% duty cycle
	.locked(locked)
	);
	
// DAC (FAST VERSION)
dac_fast mydac(
    .CLK( clk_30MHz ), // 30 MHz output clock with -50.62 phase shift and 50% duty cycle
    .RESET( RESET ), // active low, raises SYNC to interrupt write sequence
    .DAC_DATA( DAC_DATA ), // 16-bit data from user to convert to analog signal
    .SYNC( SYNC ), // frame synchronization signal for the input data
    .SCLK( SCLK ), // Serial clock input
    .DIN( DIN ), // Serial data input
    .CNT2( CNT ), // counter for SCLK
	 .DAC_FLAG( DAC_FLAG ) // Signal indicating that the 16 data bits are being written
    );
	 
//******************************************STRUCTURAL CODING******************************************//

// ADC LOCK
always@(posedge clk_80MHz)	
	if ( locked && lock_dly_cnt != 80000 )
		lock_dly_cnt <= lock_dly_cnt + 1;

wire lock_rdy = (lock_dly_cnt == 80000)? 1 : 0;	

assign cmd_channel = 0;

// ADC COMMAND VALID
always@(posedge clk_80MHz)
begin
	cmd_valid <= lock_rdy;
end

//******************************************MAIN CODE******************************************//

// if the physical switch 0 is off, don't do control and send zero current out of the device
assign control_switch = SW[0];

// Turn off 7 segment displays
assign HEX0[6:0] = 7'b1111111;
assign HEX1[6:0] = 7'b1111111;

// to decrease sample rate from 1MHZ ADC
localparam SAMPLE_DIVIDER = 5;//10 for 100KHz 5 for 200KHz
reg [5-1:0] sample_counter = 1/* synthesis noprune */;

always @ (posedge clk_80MHz)
begin
	if (rspn_valid)
	begin
		// save incoming ADC data with rspn valid is high
		adc_1_data = rspn_data;
		adc_2_data = rspn_2_data;

		// if the sample counter is less than sample divider, increment sample counter
		if (sample_counter < SAMPLE_DIVIDER)
		begin
			sample_counter = sample_counter + 1;
		end
		// else, reset sample counter and clock counter.
		// clock counter does not reset when it reaches its 
		// maximum value, only when the sample counter does. This
		// ensures that the control algorithm only operates each time
		// sample divider resets.
		else
		begin
			sample_counter = 1;
			clk_80_counter = 0;
		end
		
	end
	
	// if the clock counter is less than the counter max, increment the counter
	else if (clk_80_counter < clk_80_counter_max)
		clk_80_counter = clk_80_counter + 1;
		
	case(clk_80_counter)
	0:
		begin
			// V1i, V1im1
			V1_im1_SM1_12 = V1_i_SM1_12;
			V1_im1_SM1_17 = V1_i_SM1_17;
			V1_i_SM1_12 = adc_1_data - 2048;
			V1_i_SM1_17 = $signed({V1_i_SM1_12, input_padding});
			
			// V2i, V2im1
			V2_im1_SM1_12 = V2_i_SM1_12;
			V2_im1_SM1_17 = V2_i_SM1_17;
			V2_i_SM1_12 = adc_2_data - 2048;
			V2_i_SM1_17 = $signed({V2_i_SM1_12, input_padding});
			
			// uim1
			u_im1_SM1_17 = u_i_SM1_17;
			
			// desired
			desired_V1_i_SM1_16 = desired_V1_ip1_SM1_16;
			desired_V1_ip1_SM1_16 = desired_V1_data;
			desired_V1_ip1_SM1_33 = $signed({desired_V1_ip1_SM1_16,desired_padding});
		
			// error
			error_V1_i_S0_17 = $signed({V1_i_SM1_12,input_padding}) - $signed({desired_V1_i_SM1_16,desired_error_padding});
			
			// V1im1-V2im1
			V1mV2_im1_S0_12 = V1mV2_i_S0_12;
			V1mV2_im1_S0_17 = V1mV2_i_S0_17;
			
			// V1i-V2i
			V1mV2_i_S0_12 = V1_i_SM1_12 - V2_i_SM1_12;
			V1mV2_i_S0_17 = $signed({V1mV2_i_S0_12, input_padding});
			
			// (V1i-V2i)^2
			V1mV2_i_squared_S1_24 = V1mV2_i_S0_12*V1mV2_i_S0_12;
			V1mV2_i_squared_S1_16 = `round(V1mV2_i_squared_S1_24,quadratic_start,quadratic_round_word_length);
			
			// V1im1^2
			V1_im1_squared_SM1_24 = V1_i_squared_SM1_24;
			V1_im1_squared_SM1_18 = V1_i_squared_SM1_18;
			
			// V1i^2
			V1_i_squared_SM1_24 = V1_i_SM1_12 * V1_i_SM1_12;
			V1_i_squared_SM1_18 = `round(V1_i_squared_SM1_24,6,18);
			
		end
	1:
		begin		
			// (V1im1-V2im1)^3
			V1mV2_im1_cubed_S0_17 = V1mV2_i_cubed_S0_17;
			
			// (V1i-V2i)^3
			V1mV2_i_cubed_S0_28 = V1mV2_i_S0_12*V1mV2_i_squared_S1_16;
			V1mV2_i_cubed_S0_17 = `round(V1mV2_i_cubed_S0_28,cubic_start,cubic_round_word_length);

			// (V1i-V2i)*V1im1^2
			Nonlin1_S0_30 = V1mV2_i_S0_12*V1_im1_squared_SM1_18;
			Nonlin1_S0_17 = `round(Nonlin1_S0_30,13,18);
			
			// (V1im1-V2im1)*V1i^2
			Nonlin2_S0_30 = V1mV2_im1_S0_12*V1_i_squared_SM1_18;
			Nonlin2_S0_17 = `round(Nonlin2_S0_30,13,18);
		end	
	2:
		begin
			// Fi+1 = WF.OF in control algorithm
			unperturbed_prediction = u_im1_SM1_17*W_u_im1 + V1_i_SM1_17*W_V1_i + V2_i_SM1_17*W_V2_i + V1_im1_SM1_17*W_V1_im1 + V2_im1_SM1_17*W_V2_im1 + Nonlin1_S0_17*W_V1mV2_i + Nonlin2_S0_17*W_V1mV2_im1 + V1mV2_i_cubed_S0_17*W_V1mV2_i_cubed + V1mV2_im1_cubed_S0_17*W_V1mV2_im1_cubed;
			// K*ei in control algorithm
			Kerror = K00*error_V1_i_S0_17;
		end
	3:
		begin
			// ui in control algorithm, in full precision
			u_i_full = W_u_i_inv*(desired_V1_ip1_SM1_33 - unperturbed_prediction + Kerror);
		end		
	4:
		begin
			// ui rounded to 16 fractional bits
			u_i_Sx_16 = `round(u_i_full,u_start,u_round_word_length); // depends on W_u and W_u_inv
			// ui thresholded to 16 bits
			u_i_SM1_16 = (u_i_Sx_16 > V_FP_MAX) ? V_FP_MAX: // if it is above the maximum, set it to the maximum
                      (u_i_Sx_16 < V_FP_MIN) ? V_FP_MIN: // if it is below the minimum, set it to the minimum			
							`trunc(u_i_Sx_16,0,output_word_length); // if it is in the middle, truncate it to 16 bits
			// If the control is on, invert the first bit of the ui output to make it unsigned,
			// else set the output to the DAC value that gives zero current
			DAC_DATA = control_on ? {~u_i_SM1_16[15],u_i_SM1_16[14:0]} : DAC_ZERO_CURRENT;
		end
	6:
		begin
			// padding ui to 18 bits so it can be delayed to uim1, and then used in features
			u_i_SM1_17 = $signed({u_i_SM1_16,1'b0});
		end
	
	7:
		begin
			// Control Wait Circuit
			// After the physical switch SW[0] is turned on, wait 
			// total_wait steps until control algorithm is activated
			if (control_switch)
			begin
				if (total_wait_ctr < total_wait - 1)
					total_wait_ctr = total_wait_ctr + 1;
				else
					control_on = 1;
			end
			else
			begin
					control_on = 0;
					total_wait_ctr = 0;
			end		
		end
	
	interrupt_write: // interrupt the DAC write sequence
		begin
			RESET = 0;
		end
		
	start_write: // initiate DAC write sequence
	begin
		RESET = 1;
	end	
	
	endcase
end
	
endmodule

