`timescale 1ns / 1ps
//////////////////////////////////////////////////////////////////////////////////
// Created on Wednesday March 27 17:10 2024
//
// Description:
// Used to drive the digital to analog converter (DAC). This code has been 
// optimized to run at a higher clock speed and have lower latency on the output 
// compared to the example DAC code provided by Quartus in
// CD > Demonstrations > dac_sma > V > DAC16.v. More information about the
// regarding descriptions of the signals and the required timing can be found in
// the DAC8551 manual.
//
// Notes:
//
// @author: Robert Kent
//////////////////////////////////////////////////////////////////////////////////


module dac_fast(
    input wire CLK, // 30 MHz clock with -50.62 phase shift and 50% duty cycle
    input wire RESET, // active low, raises SYNC to interrupt write sequence
    input wire [15:0] DAC_DATA, // 16-bit data from user to convert to analog signal
    output reg SYNC, // frame synchronization signal for the input data
    output wire SCLK, // Serial clock input
    output reg DIN, // Serial data input
    output wire [7:0] CNT2, // counter for SCLK
	 output reg DAC_FLAG // high when DIN is the first or Last data bit
    );


//=======================================================
//  REG/WIRE declarations
//======================================================= 

// Power-Down Mode
wire PD1 = 0/* synthesis keep */;
wire PD0 = 0/* synthesis keep */;
reg [15:0] DAC_DATA_LOCKED = 0/* synthesis noprune */;

// State Counter
reg [4:0] CNT/* synthesis noprune */;

//=======================================================
//  Structural coding
//=======================================================

// Creating copy of CLK as output
assign SCLK = CLK;

// Creating copy of CNT as output
assign CNT2 = CNT;

// Main Code
always @(negedge RESET or posedge CLK)
begin
	if (!RESET)
	begin
		SYNC = 1; // High for CNT == 25 and CNT == 0
		DIN = 0;
		CNT = 0;
	end

	else 
	begin
	case (CNT)
	5'd0:
		// no DIN bits are sent during CNT == 0
		// Unused (bit 23)
	   begin
	       SYNC = 0;
	       DIN = 0;
	       CNT = CNT + 1;
	   end
	5'd1:
		// Unused (bit 22)
	   begin
	       DIN = 0;
	       CNT = CNT + 1;
	   end
	5'd2:
		// Unused (bit 21)
	   begin
	       DIN = 0;
	       CNT = CNT + 1;
	   end   
	5'd3:
		// Unused (bit 20)
	   begin
	       DIN = 0;
	       CNT = CNT + 1;
	   end   
	5'd4:
		// Unused (bit 19)
	   begin
	       DIN = 0;
	       CNT = CNT + 1;
	   end   
	5'd5:
		// Unused (bit 18)
	   begin
	       DIN = 0;
	       CNT = CNT + 1;
	   end   
	5'd6:
		// Power mode PD1 (bit 17)
	   begin
	       DIN = PD1;
	       CNT = CNT + 1;
	   end   
	5'd7:
		// Power mode PD0 (bit 16)
	   begin
	       DIN = PD0;
	       CNT = CNT + 1;
	   end   
	5'd8:
		// Data bit (bit 15)
	   begin
			 DAC_DATA_LOCKED = DAC_DATA;
			 DAC_FLAG = 1;
	       DIN = DAC_DATA_LOCKED[15];
	       CNT = CNT + 1;
	   end   
	5'd9:
		// Data bit (bit 14)
	   begin
			 DAC_FLAG = 0;
	       DIN = DAC_DATA_LOCKED[14];
	       CNT = CNT + 1;
	   end   
	5'd10:
		// Data bit (bit 13)
	   begin
	       DIN = DAC_DATA_LOCKED[13];
	       CNT = CNT + 1;
	   end   
	5'd11:
		// Data bit (bit 12)
	   begin
	       DIN = DAC_DATA_LOCKED[12];
	       CNT = CNT + 1;
	   end   
	5'd12:
		// Data bit (bit 11)
	   begin
	       DIN = DAC_DATA_LOCKED[11];
	       CNT = CNT + 1;
	   end   
	5'd13:
		// Data bit (bit 10)
	   begin
	       DIN = DAC_DATA_LOCKED[10];
	       CNT = CNT + 1;
	   end   
	5'd14:
		// Data bit (bit 9)
	   begin
	       DIN = DAC_DATA_LOCKED[9];
	       CNT = CNT + 1;
	   end   
	5'd15:
		// Data bit (bit 8)
	   begin
	       DIN = DAC_DATA_LOCKED[8];
	       CNT = CNT + 1;
	   end   
	5'd16:
		// Data bit (bit 7)
	   begin
	       DIN = DAC_DATA_LOCKED[7];
	       CNT = CNT + 1;
	   end   
	5'd17:
		// Data bit (bit 6)
	   begin
	       DIN = DAC_DATA_LOCKED[6];
	       CNT = CNT + 1;
	   end   
	5'd18:
		// Data bit (bit 5)
	   begin
	       DIN = DAC_DATA_LOCKED[5];
	       CNT = CNT + 1;
	   end   
	5'd19:
		// Data bit (bit 4)
	   begin
	       DIN = DAC_DATA_LOCKED[4];
	       CNT = CNT + 1;
	   end   
	5'd20:
		// Data bit (bit 3)
	   begin
	       DIN = DAC_DATA_LOCKED[3];
	       CNT = CNT + 1;
	   end   
	5'd21:
		// Data bit (bit 2)
	   begin
	       DIN = DAC_DATA_LOCKED[2];
	       CNT = CNT + 1;
	   end   
	5'd22:
		// Data bit (bit 1)
	   begin
	       DIN = DAC_DATA_LOCKED[1];
	       CNT = CNT + 1;
	   end   
	5'd23:
		// Data bit (bit 0)
	   begin
			 DAC_FLAG = 1;
	       DIN = DAC_DATA_LOCKED[0];
	       CNT = CNT + 1;
	   end   
	5'd24:
		// SYNC high for two cycles (>100ns)
	   begin
			 DAC_FLAG = 0;
	       SYNC = 1;
	       CNT = CNT + 1;
	   end   
	5'd25:
	   begin
	       CNT = 0;
	   end 
    endcase
    end
end
 
endmodule