# -*- coding: utf-8 -*-
"""
Created on Wednesday March 27 17:10 2024

Description:
Uses data stored in excel files to create all figures in the main text and 
supplementary material in "Controlling Chaos Using Edge Computing Hardware" 
by Kent et al. 2024.

Notes:

@author: Robert Kent
"""

# Preamble
import pandas as pd
import numpy as np
import matplotlib
import matplotlib.pyplot as plt
from matplotlib.ticker import MaxNLocator
from helper import V_to_I

# setting dpi for plots
matplotlib.rcParams['figure.dpi'] = 300

# defining relevant physical quantities

# ADC
V_ADC_max = 6.25#V
V_ADC_min = -6.25#V

# Input Gain
V1_gain = 6
V2_gain = 8.5

# Voltage to Current Gain
gain = 2.5

# Voltage to Current range
I_min = V_to_I(2.5,gain)
I_max = V_to_I(0,gain)
#%%

# FIGURE 2: Controlling a dynamical system using the NG-RC for one-step-ahead prediction

# Store contents of excel file in a dataframe
figure_2_df = pd.read_excel('figure_data/figure_2.xlsx')

# Control Parameters
plot_pts = 4000 # total number of points to plot
transient_start = 1500 # when control begins
transient_end = 2000 # when transient behavior has ended
dt = 1/(200e3) # sampling time
time_scaling = 1e3 # scale factor to put time in milliseconds
time_label = "(ms)" # time units
current_scaling = 1e6 # scale factor to put current into microamps
scale_label = r'$(\mu A)$' # current units

# Plot Parameters
fontsize = 20
ticksize = 20
nbinsx_Top = 6
nbinsy_Top = 8
nbinsx_Bot = 5
nbinsy_Bot = 5
box_linewidth = 1.5
tick_width = 1
tick_length= 7
top_color_1 = 'darkviolet'
top_color_2 = 'silver'
top_color_3 = 'gold'
transparency_1 = 1.0
transparency_2 = 1.0
plot_V1_max = V_ADC_max/V1_gain 
plot_V1_min = V_ADC_min/V1_gain 
plot_V2_max = 0.72
plot_V2_min = -0.72
desired_max = np.max(figure_2_df['two USS desired V1'])
desired_min = np.min(figure_2_df['two USS desired V1'])

# creating figure
fig = plt.figure(layout='constrained', figsize=(14,12))

# dividing figure into top and bottom subfigures
subfigs = fig.subfigures(2, 1, wspace=0.05,height_ratios = [0.8,1.0])

# defining top subfigure as three subplots
axsTop = subfigs[0].subplots(1, 3)

# plot each task in the top three subplots
i = 0 # initialize counter
for task in ['origin','two USS','random']:
    axsTop[i].plot(figure_2_df[task + ' V1'][:transient_end],figure_2_df[task + ' V2'][:transient_end], c = top_color_2,alpha = transparency_1,zorder = -10)
    axsTop[i].plot(figure_2_df[task + ' V1'][transient_end:plot_pts],figure_2_df[task + ' V2'][transient_end:plot_pts], c = top_color_1,zorder = 20)
    axsTop[i].plot(figure_2_df[task + ' V1'][transient_start:transient_end],figure_2_df[task + ' V2'][transient_start:transient_end],'--', c = top_color_1,alpha = transparency_2)
    axsTop[i].scatter(figure_2_df[task + ' V1'][transient_start],figure_2_df[task + ' V2'][transient_start], c = top_color_1,zorder = 10,marker = 'x',s=70)
    axsTop[i].set_xlabel(r"$V_1$ (V)",fontsize = fontsize)
    axsTop[i].tick_params(axis='both', labelsize=ticksize)
    axsTop[i].xaxis.set_major_locator(MaxNLocator(nbins=nbinsx_Top))
    axsTop[i].yaxis.set_major_locator(MaxNLocator(nbins=nbinsy_Top))
    axsTop[i].set_xlim(plot_V1_min,plot_V1_max)
    axsTop[i].set_ylim(plot_V2_min,plot_V2_max)
    axsTop[i].patch.set_edgecolor('black')  
    axsTop[i].patch.set_linewidth(box_linewidth)  
    
    # label only left vertical axis
    if i == 0:
        axsTop[i].set_ylabel(r"$V_2$ (V)",fontsize = fontsize)

    # plot vertical desired trajectory lines for two USS task
    if i == 1:
        axsTop[i].axvline(x=desired_max,c='black',zorder=-1)
        axsTop[i].axvline(x=desired_min,c='black',zorder=-1)    

    # increment counter
    i+=1

# defining bottom subfigure as two subplots
axsBot = subfigs[1].subplots(2, 1, sharex=True)

# plotting controlled V1 and V2 in first bottom subplot
axsBot[0].plot(figure_2_df['time_array'][:plot_pts],figure_2_df['two USS V1'][:plot_pts],c='blue', label = r"$V_1$")
axsBot[0].plot(figure_2_df['time_array'][:plot_pts],figure_2_df['two USS V2'][:plot_pts],c='red',label = r"$V_2$")
axsBot[0].plot(figure_2_df['time_array'][transient_start:plot_pts],figure_2_df['two USS desired V1'][transient_start:plot_pts],'-',c='black', label = r"$V_{1,des}$",zorder = -1,lw=1,alpha = 1.0)
axsBot[0].set_ylabel("Voltage (V)",fontsize = fontsize)
axsBot[0].tick_params(axis='both', labelsize=ticksize)
axsBot[0].yaxis.set_major_locator(MaxNLocator(nbins=nbinsy_Bot))
axsBot[0].set_ylim(plot_V1_min,plot_V1_max)
axsBot[0].patch.set_edgecolor('black')  
axsBot[0].patch.set_linewidth(box_linewidth)  
axsBot[0].axvline(x=time_scaling*transient_start*dt,linestyle='dashed',c='black')

# plotting control current in the second bottom subplot
axsBot[1].plot(figure_2_df['time_array'][:plot_pts],figure_2_df['two USS current'][:plot_pts],c='darkorange')#label = r'$u_1$'
axsBot[1].set_ylabel("Control Current " + scale_label,fontsize = fontsize)
axsBot[1].tick_params(axis='both', labelsize=ticksize)
axsBot[1].yaxis.set_major_locator(MaxNLocator(nbins=nbinsy_Bot))
axsBot[1].yaxis.set_major_locator(MaxNLocator(nbins=nbinsx_Bot))
axsBot[1].set_ylim(I_min*current_scaling+0.05*I_min*current_scaling,I_max*current_scaling+0.05*I_max*current_scaling)
axsBot[1].set_xlabel('time ' + time_label,fontsize=fontsize)
axsBot[1].patch.set_edgecolor('black')  
axsBot[1].patch.set_linewidth(box_linewidth)
axsBot[1].axvline(x=time_scaling*transient_start*dt,linestyle='dashed',c='black')


# adjust ticks for for top plots
for ax in axsTop.flat:
    ax.tick_params(axis='both', labelsize=ticksize)    
    ax.tick_params(axis = 'x', which='minor', direction='out', length=tick_length, width=tick_width, colors='black')
    ax.tick_params(axis = 'x', which='major', direction='out', length=tick_length, width=tick_width, colors='black')
    ax.tick_params(axis = 'y', which='minor', direction='out', length=tick_length, width=tick_width, colors='black')
    ax.tick_params(axis = 'y', which='major', direction='out', length=tick_length, width=tick_width, colors='black')

    # add boxes around plots
    ax.patch.set_edgecolor('black')  
    ax.patch.set_linewidth(box_linewidth)

# adjust ticks for for bottom plots
for ax in axsBot.flat:
    ax.tick_params(axis='both', labelsize=ticksize)
    ax.tick_params(axis = 'x', which='minor', direction='out', length=tick_length, width=tick_width, colors='black')
    ax.tick_params(axis = 'x', which='major', direction='out', length=tick_length, width=tick_width, colors='black')
    ax.tick_params(axis = 'y', which='minor', direction='out', length=tick_length, width=tick_width, colors='black')
    ax.tick_params(axis = 'y', which='major', direction='out', length=tick_length, width=tick_width, colors='black')
    
    # add boxes around plots
    ax.patch.set_edgecolor('black')  
    ax.patch.set_linewidth(box_linewidth)

# turning off major ticks on top plots
axsTop[1].set_yticks([])
axsTop[2].set_yticks([])

# removing top and bottom titles
subfigs[0].suptitle('', fontsize='x-large')
subfigs[1].suptitle('', fontsize='x-large')

# adding labels in top left corner of each subplot
axsTop[0].text(-0.97, 0.58, 'a', fontsize=34, fontweight='bold', fontfamily='sans-serif')
axsTop[1].text(-0.97, 0.58, 'b', fontsize=34, fontweight='bold', fontfamily='sans-serif')
axsTop[2].text(-0.97, 0.58, 'c', fontsize=34, fontweight='bold', fontfamily='sans-serif')
axsBot[0].text(-0.8, 0.68, 'd', fontsize=34, fontweight='bold', fontfamily='sans-serif')
axsBot[1].text(-0.8, 76, 'e', fontsize=34, fontweight='bold', fontfamily='sans-serif')

# # add arrow for start of control
# axsBot[1].annotate('', 
#                     xy=(7.3, -12),
#                     xytext=(5.5, -45),
#                     arrowprops=dict(facecolor='black', width=0.9,headwidth=8,headlength=8))
# # add start control text
# axsBot[1].text(2.15, -54, 'start control', fontsize=26, fontfamily='sans-serif')


# saving figure
# plt.savefig("figure_2.eps", format="eps",dpi=300)
# plt.savefig("figure_2.png", format="png",dpi=300)

# plotting figure
plt.show()

#%%

# Figure 3

# Store contents of excel file in a dataframe
figure_3_df = pd.read_excel('figure_data/figure_3.xlsx')

# Plot parameters
markersize = 8
nbinsx = 5
nbinsy = 6
fontsize = 20
ticksize = 20
box_linewidth = 2
capsize = 6
tick_width = 1
tick_length= 7
zorders = [10,1,-1,8]
task_labels = ['origin task','two USS task','random task','two USS task (no transient)']
colors = ['red','green','purple','blue']
markers = ['o','^','d','s']

# creating figure
fig, axs = plt.subplots(1, 2, figsize=(16,6))

# plot the control RMSE and current RMS for each task
i = 0 # initialize counter
for task in ['origin','two USS','two USS no transitions','random']: 
    axs[0].errorbar(figure_3_df['K'],figure_3_df[task + ' RMSE'],yerr = figure_3_df[task + ' RMSE uncertainty'],color = colors[i],capsize = capsize,marker = markers[i],markersize = markersize,alpha = 0.9,zorder = zorders[i])
    axs[1].errorbar(figure_3_df['K'],figure_3_df[task + ' RMS'],yerr = figure_3_df[task + ' RMS uncertainty'],color = colors[i],capsize = capsize,marker = markers[i],markersize = markersize,alpha = 0.9,zorder = zorders[i])
    i+=1 # increment counter

# setting x and y limits
axs[0].set_xlim(-0.5,1)
axs[1].set_xlim(-0.5,1)
axs[0].set_ylim(1e0,1e2)
axs[1].set_ylim(2e0,1.3e2)

# setting y scale to log
axs[0].set_yscale('log')
axs[1].set_yscale('log')

# labeling axes
axs[0].set_xlabel("K",fontsize = fontsize)
axs[1].set_xlabel("K",fontsize = fontsize)
axs[0].set_ylabel("Control RMSE (mV)",fontsize = fontsize)
axs[1].set_ylabel(r"Current RMS ($\mu$A)",fontsize = fontsize)

# adjusting ticks
for ax in axs.flat:
    ax.tick_params(axis='both', labelsize=ticksize)
    ax.xaxis.set_major_locator(MaxNLocator(nbins=nbinsx))    
    ax.tick_params(axis = 'x', which='minor', direction='out', length=tick_length, width=tick_width, colors='black')
    ax.tick_params(axis = 'x', which='major', direction='out', length=tick_length, width=tick_width, colors='black')
    ax.tick_params(axis = 'y', which='minor', direction='out', length=tick_length, width=tick_width, colors='black')
    ax.tick_params(axis = 'y', which='major', direction='out', length=tick_length, width=tick_width, colors='black')

    # add box around plots
    ax.patch.set_edgecolor('black')  
    ax.patch.set_linewidth(box_linewidth)

# adjust spacing of subplots
plt.subplots_adjust(wspace=0.2, hspace=0.1)

# adding labels in top left corner of each subplot
axs[0].text(-0.47, 68, 'a', fontsize=32, fontweight='bold', fontfamily='sans-serif')
axs[1].text(-0.47, 90, 'b', fontsize=32, fontweight='bold', fontfamily='sans-serif')

# save figure
# plt.savefig("figure_3.eps", format="eps",dpi=300)
# plt.savefig("figure_3.png", format="png",dpi=300)

# plot figure
plt.show()

#%%

# Figure 5

# Store contents of excel file in a dataframe
figure_5_df = pd.read_excel('figure_data/figure_5.xlsx')

# set variables from dataframe
latex_expressions = figure_5_df['features'] # labels for points in latex form
x_positions = np.arange(len(latex_expressions)) + 1 # x positions of points
information_criteria = figure_5_df['information criteria'] # information criteria for each point

# Plot parameters
ticksize = 30
fontsize = 28
nbins_x = 9
nbins_y = 7
box_linewidth = 2

# Label parameters
top_padding = 6500
bottom_padding = 1500
label_shift = 1800
x_position_label = x_positions + np.array([0.45,0.8,1.6,1.8,1.15,0.5,1.6,0.45,0.6,0.42,0.6,0.4,0.7])
y_position_label = information_criteria  + np.array([-400,-700,-300,-300,-300,-50,-100,-400,-200,-800,-400,-600,-150])

# create figure
fig, ax = plt.subplots(1,1, figsize = (14,8))

# plot information criteria for features used in model
ax.scatter(x_positions[:-4], information_criteria[:-4], marker='o', color='red', s=200)
ax.plot(x_positions[:-3],information_criteria[:-3],color = 'red',lw = 5)

# plot information criteria for features excluded from model
ax.scatter(x_positions[9:], information_criteria[9:], marker='x', color='black', s=200, zorder = 100,linewidths=5)
ax.plot(x_positions[9:],information_criteria[9:],color = 'black', zorder = 100,lw=5)

# Annotate each point with the LaTeX labels
for i, (x, y) in enumerate(zip(x_position_label, y_position_label)):
    ax.annotate(latex_expressions[i], (x, y), fontsize=fontsize, fontweight='bold',
                 ha='center', va='bottom', color='black', rotation=45)
    ax.set_ylim(np.min(information_criteria) - bottom_padding, np.max(information_criteria) + top_padding)

# setting x limit
ax.set_xlim(0,14.7)

# adjusting ticks
ax.set_xticks(np.arange(0, 14, 2))
ax.tick_params(axis='both', labelsize=ticksize)
ax.yaxis.set_major_locator(MaxNLocator(nbins=nbins_y))

# adding box around plot
ax.patch.set_edgecolor('black')  
ax.patch.set_linewidth(box_linewidth) 

# setting labels
ax.set_xlabel('number of terms',fontsize=fontsize)
ax.set_ylabel("Information Criteria",fontsize = fontsize)

# set tight layount
plt.tight_layout()

# save figure
# plt.savefig("figure_5.eps", format="eps",dpi = 300)
# plt.savefig("figure_5.png", format="png",dpi=300)

# show plot
plt.show()

#%%

# Supplementary Figure 1

# Store contents of excel file in a dataframe
sfigure_1_df = pd.read_excel('figure_data/supplementary_figure_1.xlsx')

# control parameters
plot_pts = 4000 # total number of points to plot
transient_start = 1500 # when control begins
transient_end = 2000 # when transient behavior has ended
dt = 1/(200e3) # sampling time
time_scaling = 1e3 # scale factor to put time in milliseconds
time_label = "(ms)" # time units
current_scaling = 1e6 # scale factor to put current into microamps
scale_label = r'$(\mu A)$' # current units

# plot parameters
fontsize = 18
ticksize = 18
box_linewidth = 1
nbinsx_Left = 5
nbinsy_Left = 8
nbinsx_Right = 5
nbinsy_Right = 5
tick_width = 1
tick_length= 7

# create figure
fig = plt.figure(layout='constrained', figsize=(12,8))

# divide figure into left and right parts
subfigs = fig.subfigures(1, 2, wspace=0.05,width_ratios=[2.5, 1.5])

# defining right subfigure as two subplots
axsRight = subfigs[1].subplots(2, 1, sharey=True)

# plot each task in the right two subplots
i = 0 # initialize counter
for task in ['origin','random']:
    axsRight[i].plot(sfigure_1_df[task + ' V1'][:transient_end],sfigure_1_df[task + ' V2'][:transient_end], c = top_color_2,alpha = transparency_1,zorder = -10)
    axsRight[i].plot(sfigure_1_df[task + ' V1'][transient_end:plot_pts],sfigure_1_df[task + ' V2'][transient_end:plot_pts], c = top_color_1,zorder = 20)
    axsRight[i].plot(sfigure_1_df[task + ' V1'][transient_start:transient_end],sfigure_1_df[task + ' V2'][transient_start:transient_end],'--', c = top_color_1,alpha = transparency_2)
    axsRight[i].scatter(sfigure_1_df[task + ' V1'][transient_start],sfigure_1_df[task + ' V2'][transient_start], c = top_color_1,zorder = 10,marker = 'x',s=70)
        
    # set x and y limits
    axsRight[i].set_xlim(V_ADC_min/V1_gain ,V_ADC_max/V1_gain )
    axsRight[i].set_ylim(V_ADC_min/V2_gain ,V_ADC_max/V2_gain )
    
    # add box to plot
    axsRight[i].patch.set_edgecolor('black')  
    axsRight[i].patch.set_linewidth(box_linewidth)
    
    # adjust ticks
    axsRight[i].tick_params(axis='both', labelsize=ticksize)
    axsRight[i].xaxis.set_major_locator(MaxNLocator(nbins=nbinsx_Right))
    axsRight[i].yaxis.set_major_locator(MaxNLocator(nbins=nbinsy_Right))
    
    # add y label
    axsRight[i].set_ylabel(r"$V_2$ (V)",fontsize = fontsize)

    # no x tick labels on right top plot
    if i == 0:
        axsRight[i].set_xticklabels([])
    
    # add x label to only right bottom plot
    if i == 1:
        axsRight[i].set_xlabel(r"$V_1$ (V)",fontsize = fontsize)

    i+=1 # increment counter

# defining left subfigure as four subplots
axsLeft = subfigs[0].subplots(4, 1, sharex=True,gridspec_kw={'hspace': 0.07})

# plot the voltages and control current for each task in the left four subplots
i = 0 # initialize counter
for task in ['origin','random']:
    # voltage plots
    axsLeft[i].plot(sfigure_1_df['time array'][:plot_pts],sfigure_1_df[task + ' V1'][:plot_pts],c='blue')
    axsLeft[i].plot(sfigure_1_df['time array'][:plot_pts],sfigure_1_df[task + ' V2'][:plot_pts],c='red')
    axsLeft[i].plot(sfigure_1_df['time array'][transient_start:plot_pts],sfigure_1_df[task + ' desired V1'][transient_start:plot_pts],'-',c='black', label = r"$V_{1,des}$",zorder = -1,lw=1.0,alpha = 1.0)
    
    # set y limit
    axsLeft[i].set_ylim(V_ADC_min/V1_gain ,V_ADC_max/V1_gain )

    # adjust ticks
    axsLeft[i].tick_params(axis='both', labelsize=ticksize)
    axsLeft[i].yaxis.set_major_locator(MaxNLocator(nbins=nbinsy_Right))
    
    # add y label
    axsLeft[i].set_ylabel("Voltage (V)",fontsize = fontsize)
    
    # add dashed vertical line to signify control start
    axsLeft[i].axvline(x=time_scaling*transient_start*dt,linestyle='dashed',c='black')
    
    # add box around plot
    axsLeft[i].patch.set_edgecolor('black')  
    axsLeft[i].patch.set_linewidth(box_linewidth)
    
    # current plots
    i+=1 # increment counter
    axsLeft[i].plot(sfigure_1_df['time array'][:plot_pts],sfigure_1_df[task + ' current'][:plot_pts],c='darkorange')
    
    # set y limit
    axsLeft[i].set_ylim(I_min*current_scaling,I_max*current_scaling)

    # adjust ticks
    axsLeft[i].tick_params(axis='both', labelsize=ticksize)
    axsLeft[i].yaxis.set_major_locator(MaxNLocator(nbins=nbinsy_Right))
    
    # add y label
    axsLeft[i].set_ylabel(r'$u_1$ ' + scale_label, fontsize = fontsize)
    
    # add dashed vertical line to signify control start
    axsLeft[i].axvline(x=time_scaling*transient_start*dt,linestyle='dashed',c='black')
    
    # add box around plot
    axsLeft[i].patch.set_edgecolor('black')  
    axsLeft[i].patch.set_linewidth(box_linewidth)
    
    # add x label only to bottom plot
    if i == 3:
        axsLeft[i].set_xlabel('time ' + time_label,fontsize=fontsize)
    
    i+=1 # increment counter

# adjust ticks
for ax in axsLeft.flat:
    ax.tick_params(axis='both', labelsize=ticksize)
    ax.patch.set_edgecolor('black')  
    ax.patch.set_linewidth(box_linewidth)
    
    ax.tick_params(axis = 'x', which='minor', direction='out', length=tick_length, width=tick_width, colors='black')
    ax.tick_params(axis = 'x', which='major', direction='out', length=tick_length, width=tick_width, colors='black')
    ax.tick_params(axis = 'y', which='minor', direction='out', length=tick_length, width=tick_width, colors='black')
    ax.tick_params(axis = 'y', which='major', direction='out', length=tick_length, width=tick_width, colors='black')

for ax in axsRight.flat:
    ax.tick_params(axis='both', labelsize=ticksize)
    ax.patch.set_edgecolor('black')  
    ax.patch.set_linewidth(box_linewidth)
    
    ax.tick_params(axis = 'x', which='minor', direction='out', length=tick_length, width=tick_width, colors='black')
    ax.tick_params(axis = 'x', which='major', direction='out', length=tick_length, width=tick_width, colors='black')
    ax.tick_params(axis = 'y', which='minor', direction='out', length=tick_length, width=tick_width, colors='black')
    ax.tick_params(axis = 'y', which='major', direction='out', length=tick_length, width=tick_width, colors='black')

axsLeft[1].set_yticks([-100,-50,0,50,100])
axsLeft[3].set_yticks([-100,-50,0,50,100])

subfigs[0].suptitle('', fontsize='x-large')

# remove plot title
fig.suptitle('', fontsize='xx-large')

# adding labels in top left corner of each subplot
axsLeft[0].text(-0.84, 0.65, 'a', fontsize=26, fontweight='bold', fontfamily='sans-serif')
axsLeft[1].text(-0.84, 60, 'b', fontsize=26, fontweight='bold', fontfamily='sans-serif')
axsLeft[2].text(-0.84, 0.65, 'c', fontsize=26, fontweight='bold', fontfamily='sans-serif')
axsLeft[3].text(-0.84, 60, 'd', fontsize=26, fontweight='bold', fontfamily='sans-serif')
axsRight[0].text(-0.985, 0.59, 'e', fontsize=26, fontweight='bold', fontfamily='sans-serif')
axsRight[1].text(-0.985, 0.59, 'f', fontsize=26, fontweight='bold', fontfamily='sans-serif')

# # add arrow for start of control
# axsLeft[3].annotate('', 
#                     xy=(7.3, -15),
#                     xytext=(5.3, -45),
#                     arrowprops=dict(facecolor='black', width=0.9,headwidth=8,headlength=8))
# # add start control text
# axsLeft[3].text(-0.25, -57, 'start control', fontsize=21, fontfamily='sans-serif')

# save figure
# plt.savefig("supplementary_figure_1.eps", format="eps",dpi=300)
# plt.savefig("supplementary_figure_1.png", format="png",dpi=300)

plt.show()

#%%

# Supplementary Figure 2

# Store contents of excel file in a dataframe
sfigure_2_df = pd.read_excel('figure_data/supplementary_figure_2.xlsx')

# defining variables from dataframe
delay_real_list = sfigure_2_df['delays us'].dropna().tolist()
delay0 = delay_real_list[0]

# Plot parameters
markersize = 5
nbinsx = 5
nbinsy = 6
fontsize = 20
ticksize = 20
box_linewidth = 1
zorders = [10,1,-1,8]
task_labels = ['origin task','two USS task','two USS task (no transient)','random task']
colors = ['red','green','purple','blue']
markers = ['o','^','d','s']

matplotlib.rcParams['hatch.linewidth'] = 1.5
hatch_patterns = ['//','\\\\','None','|',]

# create figure
fig, axs = plt.subplots(3, 3, figsize=(16,12))

# plot each method and task
i = 0
for method in ['linear','one step','two step']:
    j = 0
    for task in ['origin','two USS','two USS no transition','random']:
        
        # plot control RMSE
        axs[0,i].errorbar(sfigure_2_df[method + ' K'],sfigure_2_df[method + ' ' + task + ' RMSE'],yerr = sfigure_2_df[method + ' ' + task + ' RMSE uncertainty'],color = colors[j],capsize = 3,marker = markers[j],markersize = markersize,alpha = 0.9,zorder = zorders[j])
        
        # plot current RMS
        axs[1,i].errorbar(sfigure_2_df[method + ' K'],sfigure_2_df[method + ' ' + task + ' RMS'],yerr = sfigure_2_df[method + ' ' + task + ' RMS uncertainty'],color = colors[j],capsize = 3,marker = markers[j],markersize = markersize,alpha = 0.9,zorder = zorders[j])
        
        # plot domain of control
        if task != 'two USS no transition':
            # axs[2,i].fill(sfigure_2_df[method + ' ' + task + ' y values'], sfigure_2_df[method + ' ' + task + ' x values'], color=colors[j], alpha=0.3)
            axs[2,i].fill(sfigure_2_df[method + ' ' + task + ' y values'], sfigure_2_df[method + ' ' + task + ' x values'], color='None',edgecolor = colors[j], hatch=hatch_patterns[j])
            axs[2,i].plot(sfigure_2_df[method + ' ' + task + ' y values'], sfigure_2_df[method + ' ' + task + ' x values'], color = colors[j],marker = markers[j],markersize = markersize,alpha = 1.0)

        j+=1 # increment task counter
        
    # plot horizontal dashed line representing the minimum latency
    axs[2,i].axhline(delay0,linestyle = 'dashed',color = 'black',lw = 3,zorder = -10)    
    i+=1 # increment method counter

# setting x and y limits
axs[0,0].set_xlim(-2.0,1.0)
axs[1,0].set_xlim(-2.0,1.0)
axs[2,0].set_xlim(-2.0,1.0)
axs[0,1].set_xlim(-0.5,1.0)
axs[1,1].set_xlim(-0.5,1.0)
axs[2,1].set_xlim(-0.5,1.0)
axs[0,2].set_xlim(-1.0,1.0)
axs[1,2].set_xlim(-1.0,1.0)
axs[2,2].set_xlim(-1.0,1.0)
axs[0,0].set_ylim(1e0,2e2)
axs[0,1].set_ylim(1e0,2e2)
axs[0,2].set_ylim(1e0,2e2)
axs[1,0].set_ylim(2e0,1.6e2)
axs[1,1].set_ylim(2e0,1.6e2)
axs[1,2].set_ylim(2e0,1.6e2)
axs[2,0].set_ylim(0,delay_real_list[-1]+0.5)
axs[2,1].set_ylim(0,delay_real_list[-1]+0.5)
axs[2,2].set_ylim(0,delay_real_list[-1]+0.5)

# setting y scale to log
axs[0,0].set_yscale('log')
axs[0,1].set_yscale('log')
axs[0,2].set_yscale('log')
axs[1,0].set_yscale('log')
axs[1,1].set_yscale('log')
axs[1,2].set_yscale('log')

# set axes labels
axs[2,0].set_xlabel("K",fontsize = fontsize)
axs[2,1].set_xlabel("K",fontsize = fontsize)
axs[2,2].set_xlabel("K",fontsize = fontsize)
axs[0,0].set_ylabel("Control RMSE (mV)",fontsize = fontsize)
axs[1,0].set_ylabel(r"Current RMS ($\mu$A)",fontsize = fontsize)
axs[2,0].set_ylabel(r"Delay (us)",fontsize = fontsize)

# set ticks
axs[0,0].set_xticks([])
axs[0,1].set_xticks([])
axs[0,2].set_xticks([])
axs[1,0].set_xticks([])
axs[1,1].set_xticks([])
axs[1,2].set_xticks([])
axs[2,0].set_xticks([-2.0,-1.0,0.0,1.0])
axs[2,1].set_xticks([-0.5,0.0,0.5,1.0])
axs[2,2].set_xticks([-1.0,-0.5,0.0,0.5,1.0])
axs[0,1].set_yticks([])
axs[0,2].set_yticks([])
axs[1,1].set_yticks([])
axs[1,2].set_yticks([])
axs[2,1].set_yticks([])
axs[2,2].set_yticks([])

# adjust ticks
axs[2,0].yaxis.set_major_locator(MaxNLocator(nbins=7))

# set title
axs[0,0].set_title('linear',fontsize=fontsize)
axs[0,1].set_title('one-step ahead',fontsize=fontsize)
axs[0,2].set_title('two-step ahead',fontsize=fontsize)

# adjust ticks
for ax in axs.flat:
    ax.tick_params(axis='both', labelsize=ticksize)
    ax.tick_params(axis = 'x', which='minor', direction='out', length=tick_length, width=tick_width, colors='black')
    ax.tick_params(axis = 'x', which='major', direction='out', length=tick_length, width=tick_width, colors='black')
    ax.tick_params(axis = 'y', which='minor', direction='out', length=tick_length, width=tick_width, colors='black')
    ax.tick_params(axis = 'y', which='major', direction='out', length=tick_length, width=tick_width, colors='black')

    # add box around plots
    ax.patch.set_edgecolor('black')  
    ax.patch.set_linewidth(box_linewidth)

axs[0,1].minorticks_off()
axs[0,2].minorticks_off()
axs[1,1].minorticks_off()
axs[1,2].minorticks_off()

# adjust subplot spacing
plt.subplots_adjust(wspace=0.2, hspace=0.15)

# adding labels in top left corner of each subplot
axs[0,0].text(-1.93, 107, 'a', fontsize=28, fontweight='bold', fontfamily='sans-serif')
axs[0,1].text(-0.46, 107, 'b', fontsize=28, fontweight='bold', fontfamily='sans-serif')
axs[0,2].text(-0.94, 107, 'c', fontsize=28, fontweight='bold', fontfamily='sans-serif')
axs[1,0].text(-1.93, 91, 'd', fontsize=28, fontweight='bold', fontfamily='sans-serif')
axs[1,1].text(-0.46, 91, 'e', fontsize=28, fontweight='bold', fontfamily='sans-serif')
axs[1,2].text(-0.94, 91, 'f', fontsize=28, fontweight='bold', fontfamily='sans-serif')
axs[2,0].text(-1.93, 11.9, 'g', fontsize=28, fontweight='bold', fontfamily='sans-serif')
axs[2,1].text(-0.46, 11.9, 'h', fontsize=28, fontweight='bold', fontfamily='sans-serif')
axs[2,2].text(-0.94, 11.9, 'i', fontsize=28, fontweight='bold', fontfamily='sans-serif')

# save figure
# plt.savefig("supplementary_figure_2.eps", format="eps",dpi=300)
# plt.savefig("supplementary_figure_2.png", format="png",dpi=300)

# show figure
plt.show()
#%%

# Supplementary Figure 3

# Store contents of excel file in a dataframe
sfigure_3_df = pd.read_excel('figure_data/supplementary_figure_3.xlsx')

# plot parameters
markersize = 8
nbinsx = 5
nbinsy = 6
fontsize = 20
ticksize = 20
box_linewidth = 2
capsize = 6
tick_width = 1
tick_length= 7
zorders = [10,1,8,-1]
task_labels = ['origin task','two USS task','two USS task (no transient)','random task']
task_list_2 = ['task_1','task_2','task_2_USS','task_4']
colors = ['red','green','purple','blue']
markers = ['o','^','d','s']

# create figure
fig, axs = plt.subplots(1, 3, figsize=(16,5))

i = 0 # initialize method counter
for method in ['linear','one step','two step']:
    j = 0 # initialize task counter
    axs[i].errorbar(sfigure_3_df[method + ' K'],sfigure_3_df[method + ' origin'],yerr = sfigure_3_df[method + ' origin uncertainty'],color = colors[j],capsize = capsize,marker = markers[j],markersize = markersize,alpha = 0.9,zorder = zorders[j])
    j+=1 # increment task counter
    axs[i].errorbar(sfigure_3_df[method + ' K'],sfigure_3_df[method + ' two USS'],yerr = sfigure_3_df[method + ' two USS uncertainty'],color = colors[j],capsize = capsize,marker = markers[j],markersize = markersize,alpha = 0.9,zorder = zorders[j])
    j+=1 # increment task counter
    axs[i].errorbar(sfigure_3_df[method + ' K'],sfigure_3_df[method + ' two USS no transition'],yerr = sfigure_3_df[method + ' two USS no transition uncertainty'],color = colors[j],capsize = capsize,marker = markers[j],markersize = markersize,alpha = 0.9,zorder = zorders[j])
    j+=1 # increment task counter
    axs[i].errorbar(sfigure_3_df[method + ' K'],sfigure_3_df[method + ' random'],yerr = sfigure_3_df[method + ' random uncertainty'],color = colors[j],capsize = capsize,marker = markers[j],markersize = markersize,alpha = 0.9,zorder = zorders[j])
    i+=1 # increment method counter

# set x and y limits
axs[0].set_xlim(-2.0,1.0)
axs[1].set_xlim(-0.5,1.0)
axs[2].set_xlim(-1.0,1.0)
axs[0].set_ylim(3e0,1.9e2)
axs[1].set_ylim(3e0,1.9e2)
axs[2].set_ylim(3e0,1.9e2)

# set y scale to log
axs[0].set_yscale('log')
axs[1].set_yscale('log')
axs[2].set_yscale('log')

# set x and y labels
axs[0].set_xlabel("K",fontsize = fontsize)
axs[1].set_xlabel("K",fontsize = fontsize)
axs[2].set_xlabel("K",fontsize = fontsize)
axs[0].set_ylabel(r"Max abs. current ($\mu$A)",fontsize = fontsize)

# set x ticks
axs[0].set_xticks([-2.0,-1.0,0.0,1.0])
axs[1].set_xticks([-0.5,0.0,0.5,1.0])
axs[2].set_xticks([-1.0,-0.5,0.0,0.5,1.0])

# set titles
axs[0].set_title("linear",fontsize = fontsize)
axs[1].set_title("one-step ahead",fontsize = fontsize)
axs[2].set_title("two-step ahead",fontsize = fontsize)


# adjust ticks
for ax in axs.flat:
    ax.tick_params(axis='both', labelsize=ticksize)
    ax.tick_params(axis = 'x', which='minor', direction='out', length=tick_length, width=tick_width, colors='black')
    ax.tick_params(axis = 'x', which='major', direction='out', length=tick_length, width=tick_width, colors='black')
    ax.tick_params(axis = 'y', which='minor', direction='out', length=tick_length, width=tick_width, colors='black')
    ax.tick_params(axis = 'y', which='major', direction='out', length=tick_length, width=tick_width, colors='black')

    # add boxes around plots
    ax.patch.set_edgecolor('black')  
    ax.patch.set_linewidth(box_linewidth)

# adjust subplot spacing
plt.subplots_adjust(wspace=0.2, hspace=0.1)

# adding labels in top left corner of each subplot
axs[0].text(-1.93, 128, 'a', fontsize=28, fontweight='bold', fontfamily='sans-serif')
axs[1].text(-0.46, 128, 'b', fontsize=28, fontweight='bold', fontfamily='sans-serif')
axs[2].text(-0.94, 128, 'c', fontsize=28, fontweight='bold', fontfamily='sans-serif')


# save figure
# plt.savefig("supplementary_figure_3.eps", format="eps",dpi=300)
# plt.savefig("supplementary_figure_3.png", format="png",dpi=300)

# show figure
plt.show()
#%%

# Supplementary Figure 4

# Store contents of excel file in a dataframe
sfigure_4_df = pd.read_excel('figure_data/supplementary_figure_4.xlsx')

# create figure
fig, ax = plt.subplots(1, 1, figsize=(12, 8))

# Create a list of y-values for the horizontal bar graph
y = list(range(1, len(sfigure_4_df['one step ahead weights']) + 1))

# plot the weights using horizontal bar graph with error bars
ax.barh(y, sfigure_4_df['one step ahead weights'], xerr=sfigure_4_df['one step ahead weights uncertainty'], capsize=20,ecolor = 'black', color='blue')
ax.barh(y, sfigure_4_df['two step ahead weights'], xerr=sfigure_4_df['two step ahead weights uncertainty'], capsize=20, ecolor = 'black', color='red',zorder = -1)

# set custom y labels
ax.set_yticks(y, sfigure_4_df['labels'])

# set x limit
ax.set_xlim(-2.6, 2.6)

# set x and y labels
ax.set_xlabel("Weights", fontsize=fontsize)
ax.set_ylabel("Features", fontsize=fontsize)

# adjust ticks
ax.tick_params(axis='both', labelsize=ticksize)
ax.xaxis.set_major_locator(MaxNLocator(nbins=nbinsx))

# add box around plot
ax.patch.set_edgecolor('black')  
ax.patch.set_linewidth(box_linewidth)

# set tight layout
fig.tight_layout()

# save figure
# plt.savefig("supplementary_figure_4.eps", format="eps",dpi=300)
# plt.savefig("supplementary_figure_4.png", format="png",dpi=300)

# show figure
plt.show()
#%%

# Supplementary Figure 5

# Store contents of excel file in a dataframe
sfigure_5_df = pd.read_excel('figure_data/supplementary_figure_5.xlsx')

# plot parameters
markersize = 8
nbinsx = 5
nbinsy = 6
fontsize = 20
ticksize = 20
box_linewidth = 2
capsize = 6
zorders = [10,1,-1,8]
task_labels = ['origin task','two USS task','random task','two USS task (no transient)']
colors = ['red','green','purple','blue']
markers = ['o','^','d','s']

# create figure
fig, axs = plt.subplots(1, 2, figsize=(16,6))

# plot each method
i = 0 # initialize method counter
for method in ['one step','two step']:
    # plot each task
    j = 0 # initialize task counter
    for task in ['origin','two USS','two USS no transition','random']:
        axs[i].errorbar(sfigure_5_df[method + ' K'],sfigure_5_df[method + ' ' + task + ' RMSE'],yerr = sfigure_5_df[method + ' ' + task + ' RMSE uncertainty'],color = colors[j],capsize = capsize,marker = markers[j],markersize = markersize,alpha = 0.9,zorder = zorders[j])
        j+=1 # increment task counter
    i+=1 # increment metod counter

# set x and y limits
axs[0].set_xlim(-0.52,1.03)
axs[1].set_xlim(-0.52,1.03)
axs[0].set_ylim(4e-1,1.7e2)
axs[1].set_ylim(4e-1,1.7e2)

# set y scale to log
axs[0].set_yscale('log')
axs[1].set_yscale('log')

# set x and y label
axs[0].set_xlabel("K",fontsize = fontsize)
axs[1].set_xlabel("K",fontsize = fontsize)
axs[0].set_ylabel("Prediction RMSE (mV)",fontsize = fontsize)

# adjust ticks
for ax in axs.flat:
    ax.tick_params(axis='both', labelsize=ticksize)
    ax.tick_params(axis = 'x', which='minor', direction='out', length=tick_length, width=tick_width, colors='black')
    ax.tick_params(axis = 'x', which='major', direction='out', length=tick_length, width=tick_width, colors='black')
    ax.tick_params(axis = 'y', which='minor', direction='out', length=tick_length, width=tick_width, colors='black')
    ax.tick_params(axis = 'y', which='major', direction='out', length=tick_length, width=tick_width, colors='black')
    
    # add box around plot
    ax.patch.set_edgecolor('black')  
    ax.patch.set_linewidth(box_linewidth)

axs[0].set_xticks([-0.5,0.0,0.5,1.0])
axs[1].set_xticks([-0.5,0.0,0.5,1.0])
axs[1].yaxis.set_ticks([])

# adjust subplot spacing
plt.subplots_adjust(wspace=0.1, hspace=0.1)

# adding labels in top left corner of each subplot
axs[0].text(-0.49, 103, 'a', fontsize=32, fontweight='bold', fontfamily='sans-serif')
axs[1].text(-0.49, 103, 'b', fontsize=32, fontweight='bold', fontfamily='sans-serif')

# save figure
# plt.savefig("supplementary_figure_5.eps", format="eps",dpi=300)
# plt.savefig("supplementary_figure_5.png", format="png",dpi=300)

# show plot
plt.show()
#%%

# Supplementary Figure 6

# Store contents of excel file in a dataframe
sfigure_6_df = pd.read_excel('figure_data/supplementary_figure_6.xlsx')

# plot parameters
ticksize = 24
fontsize = 24
nbins_x = 6
nbins_y = 7
box_linewidth = 2

# create figure
fig,ax = plt.subplots(1,1,figsize = (10,7))

# plot delay vs. sample rate
ax.plot(sfigure_6_df['sample rate (kHz)'],sfigure_6_df['delay (us)'],c='red',lw = 4)

# set x and y labels
ax.set_xlabel(r'Sample Rate (KHz)',fontsize = fontsize)
ax.set_ylabel(r'Delay $(\mu s)$',fontsize = fontsize)

# adjust ticks
ax.tick_params(axis='both', labelsize=ticksize)
ax.xaxis.set_major_locator(MaxNLocator(nbins=nbins_x))
ax.yaxis.set_major_locator(MaxNLocator(nbins=nbins_y))

# add box around plot
ax.patch.set_edgecolor('black')  
ax.patch.set_linewidth(box_linewidth)  

# save figure
# plt.savefig("supplementary_figure_6.eps", format="eps",dpi = 300)
# plt.savefig("supplementary_figure_6.png", format="png",dpi=300)

# show figure
plt.show()
