# -*- coding: utf-8 -*-
"""
Created on Wednesday March 27 17:10 2024

Description:
Contains helper functions for main.py

Notes:

@author: Robert Kent
"""
# preamble
import numpy as np
from datetime import datetime
import mif
import quartustcl
from scipy.optimize import fsolve
from fxpmath import Fxp
#%%

class FPGAController:
    """
    Read and write to memory of a connected Max 10 FPGA device to use the
    control algorithm.
    """
    def __init__(self,
                 desired_word_length=16,
                 desired_depth=4200,
                 W_n_frac=0,
                 weight_word_length=18):
        """
        Connect to FPGA and begin a memory edit, and initialize variables

        Parameters:
            desired_word_length (int): Number of bits used to represent desired
            signal
            
            desired_depth (int): Number of memory addresses required to store
            desired signal
            
            W_n_frac (int): Number of fractional bits in the fixed point
            representation of the weights
            
            weight_word_length (int): Number of bits used to represent the
            weights
            
        """
        # Instantiate a QuartusTcl object to start a shell.
        self.quartus = quartustcl.QuartusTcl()
        
        # Connect to first detected hardware
        hwnames = self.quartus.parse(self.quartus.get_hardware_names())
        self.hwname = hwnames[0]
        
        # Connect to first device in hardware
        devnames = self.quartus.parse(self.quartus.get_device_names(hardware_name=self.hwname))
        self.devname = devnames[0]
        
        # Get editable memory instances
        memories_raw = self.quartus.get_editable_mem_instances(hardware_name=self.hwname,
                                                               device_name=self.devname)
        memories = self.quartus.parse(memories_raw, levels=2)
        
        # convert memory instances to numpy array
        self.memories_np = np.asarray(memories)
        
        # begin memory edit
        self.quartus.begin_memory_edit(hardware_name=self.hwname, device_name=self.devname)
        
        # setting parameters
        self.desired_depth = desired_depth
        self.desired_word_length = desired_word_length
        self.W_n_frac = W_n_frac
        self.weight_word_length = weight_word_length
        
        
    def write_train_test(self, 
                         int16_train_test_arr, 
                         output_word_length):
        """
        Write perturbation for training and testing to FPGA memory
        
        Parameters:
            int16_train_test_arr (numpy.ndarray): Training and testing
            perturbation stored in array with 16-bit unsigned integer values
            to match DAC output.
            
            output_word_length (int): Number of bits used to represent the
            perturbation at each timestep, typically 16.
        """
        
        # write memory initialization file 
        with open(r'train_test_new.mif', 'w') as fp:
            fp.write("DEPTH=" + str(len(int16_train_test_arr)) + ";\n")
            fp.write("WIDTH=" + str(output_word_length) + ";\n")
            fp.write("ADDRESS_RADIX=DECIMAL;\n")
            fp.write("DATA_RADIX=DECIMAL;\n")
            fp.write("CONTENT\n")
            fp.write("BEGIN\n")
            for i in range(len(int16_train_test_arr)):
                # write each item on a new line
                string = str(i) + ":" + str(int(int16_train_test_arr[i])) + ";\n"
                fp.write(string)
            fp.write("END")
        
        # write to memory, 'UTT' corresponds to 'u train test'
        found_memid = np.argwhere(self.memories_np=='UTT')[0][0]
        self.quartus.update_content_to_memory_from_file(instance_index=found_memid,
                                                        mem_file_path='train_test_new.mif',
                                                        mem_file_type='mif')
        
    def control_switch(self, 
                       state):
        """
        Start or stop the control sequence by writing to FPGA memory
        
        Parameters:
            state (int): Whether to start the control sequence (1) or stop
            the control sequence (0)
        """
        
        # write memory initialization file 
        with open(r'switch_new.mif', 'w') as fp:
            fp.write("DEPTH=" + str(1) + ";\n")
            fp.write("WIDTH=" + str(1) + ";\n")
            fp.write("ADDRESS_RADIX=DECIMAL;\n")
            fp.write("DATA_RADIX=DECIMAL;\n")
            fp.write("CONTENT\n")
            fp.write("BEGIN\n")
            for i in range(1):
                # write each item on a new line
                string = str(i) + ":" + str(state) + ";\n"
                fp.write(string)
            fp.write("END")
            
        # write to memory, 'SW' corresponds to 'switch'
        found_memid = np.argwhere(self.memories_np=='SW')[0][0]
        self.quartus.update_content_to_memory_from_file(instance_index=found_memid,
                                                        mem_file_path='switch_new.mif',
                                                        mem_file_type='mif')

    def set_mode(self, 
                 state):
        """
        Set FPGA to training and testing mode or closed loop feedback mode by
        writing to the FPGA memory
        
        Parameters:
            state (int): Whether to set the mode to training and testing (1) or
            closed loop feedback (0)
        """
        with open(r'op_mode_new.mif', 'w') as fp:
            fp.write("DEPTH=" + str(1) + ";\n")
            fp.write("WIDTH=" + str(1) + ";\n")
            fp.write("ADDRESS_RADIX=DECIMAL;\n")
            fp.write("DATA_RADIX=DECIMAL;\n")
            fp.write("CONTENT\n")
            fp.write("BEGIN\n")
            for i in range(1):
                # write each item on a new line
                string = str(i) + ":" + str(state) + ";\n"
                fp.write(string)
            fp.write("END")
        
        # write to memory
        found_memid = np.argwhere(self.memories_np=='MODE')[0][0]
        self.quartus.update_content_to_memory_from_file(instance_index=found_memid,
                                                        mem_file_path='op_mode_new.mif',
                                                        mem_file_type='mif')

    def write_delay(self,state):
        """
        Set the added delay in applying the control perturbations by writing
        to the FPGA memory.
        
        Parameters:
            state (int): Determines the delay amount in microseconds to add, a
            0 means no delay, and 31 means the maximum delay of 31us
        """
        with open(r'delay_new.mif', 'w') as fp:
            fp.write("DEPTH=" + str(1) + ";\n")
            fp.write("WIDTH=" + str(5) + ";\n")
            fp.write("ADDRESS_RADIX=DECIMAL;\n")
            fp.write("DATA_RADIX=DECIMAL;\n")
            fp.write("CONTENT\n")
            fp.write("BEGIN\n")
            for i in range(1):
                # write each item on a new line
                string = str(i) + ":" + str(state) + ";\n"
                fp.write(string)
            fp.write("END")
        
        # write to memory, 'DELA' corresponds to 'delay'
        found_memid = np.argwhere(self.memories_np=='DELA')[0][0]
        self.quartus.update_content_to_memory_from_file(instance_index=found_memid,
                                                        mem_file_path='delay_new.mif',
                                                        mem_file_type='mif')
        
    def write_feedback_linearization(self,
                                     state):
        """
        Determine whether the control algorithm is strictly linear, or includes
        nonlinear terms (NG-RC) by writing to FPGA memory
        
        Parameters:
            state (int): Whether to include nonlinear terms (1) or exclude 
            the nonlinear terms (0)
        """
        with open(r'feed_new.mif', 'w') as fp:
            fp.write("DEPTH=" + str(1) + ";\n")
            fp.write("WIDTH=" + str(1) + ";\n")
            fp.write("ADDRESS_RADIX=DECIMAL;\n")
            fp.write("DATA_RADIX=DECIMAL;\n")
            fp.write("CONTENT\n")
            fp.write("BEGIN\n")
            for i in range(1):
                # write each item on a new line
                string = str(i) + ":" + str(state) + ";\n"
                fp.write(string)
            fp.write("END")
        
        # write to memory, 'FEED' corresponds to 'feedback linearization'
        found_memid = np.argwhere(self.memories_np=='FEED')[0][0]
        self.quartus.update_content_to_memory_from_file(instance_index=found_memid,
                                                        mem_file_path='feed_new.mif',
                                                        mem_file_type='mif')
        
    def load_data(self, 
                  ID, 
                  signed = False):
        """
        Load data stored in FPGA memory with specified ID
        
        Parameters:
            ID (str): The ID string corresponding to a specific block memory on
            the FPGA device. The ID is chosen when instantiating the RAM in the
            FPGA code
            
            signed (bool): Whether the data stored in the memory is signed (True)
            or unsigned (False)
        
        Returns:
            return_val (numpy.ndarray): Array containing signed or unsigned 
            decimal values loaded from memory with specified ID
        """
        
        # Search through memories to find specific ID
        found_memid = np.argwhere(self.memories_np==ID)[0][0]
        self.quartus.save_content_from_memory_to_file(
            instance_index=found_memid,
            mem_file_path= ID + '.mif',
            mem_file_type='mif',
        )    
        
        # load data stored in RAM with specific ID        
        with open(ID + '.mif') as f:
            data = mif.load(f)
            
        # convert binary array to signed or unsigned decimal
        return_val = binary_array_to_decimal(data,signed = signed)
        
        return return_val
    
    def write_desired(self,
                      x_desired):
        """
        Write the desired signal to FPGA memory
        
        Parameters:
            x_desired (numpy.ndarray): Array containing desired signal with 
            16-bit signed digital values.
        """
        with open(r'desired_V1_new.mif', 'w') as fp:
            fp.write("DEPTH=" + str(self.desired_depth) + ";\n")
            fp.write("WIDTH=" + str(self.desired_word_length) + ";\n")
            fp.write("ADDRESS_RADIX=DECIMAL;\n")
            fp.write("DATA_RADIX=DECIMAL;\n")
            fp.write("CONTENT\n")
            fp.write("BEGIN\n")
            for i in range(self.desired_depth):
                # write each item on a new line
                string = str(i) + ":" + str(int(x_desired[i])) + ";\n"
                fp.write(string)
            fp.write("END")
            
        # write to memory, 'DV1R' corresponds to 'desired V1 read'
        found_memid = np.argwhere(self.memories_np=='DV1R')[0][0]
        self.quartus.update_content_to_memory_from_file(instance_index=found_memid,
                                                        mem_file_path='desired_V1_new.mif',
                                                        mem_file_type='mif')
        
    def write_K(self,
                K00):
        """
        Set the control gain by writing to FPGA memory
        
        Parameters:
            K00 (float): The control gain
            
        Returns:
            K00_int (int): The control gain in integer form
        """
        # converting the control gain to integer form
        K00_int = np.rint(K00*2**self.W_n_frac)
        with open(r'K00_new.mif', 'w') as fp:
            fp.write("DEPTH=" + str(1) + ";\n")
            fp.write("WIDTH=" + str(self.weight_word_length) + ";\n")
            fp.write("ADDRESS_RADIX=DECIMAL;\n")
            fp.write("DATA_RADIX=DECIMAL;\n")
            fp.write("CONTENT\n")
            fp.write("BEGIN\n")
            for i in range(1):
                # write each item on a new line
                string = str(i) + ":" + str(int(K00_int)) + ";\n"
                fp.write(string)
            fp.write("END")
    
        # write to memory
        found_memid = np.argwhere(self.memories_np=='K00')[0][0]
        self.quartus.update_content_to_memory_from_file(instance_index=found_memid,
                                                        mem_file_path='K00_new.mif',
                                                        mem_file_type='mif')
        return K00_int

class fixed_point_helper:
    """
    Used to access physical quantities pertaining to the experimental setup, as
    well as the fixed-point forms that are commmonly used in the code. There are
    functions to help convert between digital quantities, fixed point quantities,
    real quantities with gain, and real quantities. There is also a function to 
    compute the nonzero unstable steady states of the double scroll system.
    """
    def __init__(self):
        
        """
        Initialize physical quantities and all relevant fixed point forms used
        in the main code.
        """
        
        # defining physical quantities
        self.V_ADC_max = 6.25#V
        self.V_ADC_min = -6.25#V
        self.V1_gain = 6
        self.V2_gain = 8.5
        self.V_DAC_max = 2.5#V
        self.V_DAC_min = 0#V
        
        self.V1_real_min = self.V_ADC_min/self.V1_gain
        self.V1_real_max = self.V_ADC_max/self.V1_gain
        
        self.V2_real_min = self.V_ADC_min/self.V2_gain
        self.V2_real_max = self.V_ADC_max/self.V2_gain
        
        # defining ADC and DAC word lengths
        self.adc_word_length = 12
        self.dac_word_length = 16
        
        # Voltage to Current Gain
        self.gain = 2.5
        
        # getting the maximum and minimum current
        self.I_min = V_to_I(2.5,self.gain)
        self.I_max = V_to_I(0,self.gain)
        
        # setting fixed-point arithmetic parameters

        # linear features
        self.input_word_length = self.adc_word_length # V1_n_U12_0, V2_n_U12_0, V1_n_SM1_12, V2_n_SM1_12, V1_nm1_SM1_12, V2_nm1_SM1_12
        self.input_fractional_bits = 12
        self.U12_0_TYPE = Fxp(val=None,signed=False,n_word=self.input_word_length,n_frac=0)
        self.S11_0_TYPE = Fxp(val=None,signed=True,n_word=self.input_word_length,n_frac=0)
        self.SM1_12_TYPE = Fxp(val=None,signed=True,n_word=self.input_word_length,n_frac=self.input_fractional_bits)
        self.difference_word_length = self.input_word_length + 1 # V1mV2_n_S0_12, V1mV2_nm1_S0_12
        self.S0_12_TYPE = Fxp(val=None,signed=True,n_word=self.input_word_length+1,n_frac=self.input_fractional_bits)
        self.output_word_length = self.dac_word_length # u_n_U16_0, u_n_SM1_16, u_nm1_SM1_16
        self.output_fractional_bits = 16
        self.U16_0_TYPE = Fxp(val=None,signed=False,n_word=self.output_word_length,n_frac=0)
        self.S15_0_TYPE = Fxp(val=None,signed=True,n_word=self.output_word_length,n_frac=0)
        self.SM1_16_TYPE = Fxp(val=None,signed=True,n_word=self.output_word_length,n_frac=self.output_fractional_bits)

        # padded linear features (0s added to right hand side to match fixed point form of other features)
        self.linear_word_length = 17 # u_n_SM1_17, u_nm1_SM1_17, V1_n_SM1_17, V2_n_SM1_17, V1_nm1_SM1_17, V2_nm1_SM1_17, 
        self.input_padding = self.linear_word_length - self.input_word_length
        self.output_padding = self.linear_word_length - self.output_word_length
        self.SM1_17_TYPE = Fxp(val=None,signed=True,n_word=self.linear_word_length,n_frac=self.input_fractional_bits + self.input_padding) # output requires output_padding
        self.feature_word_length = 18 # V1mV2_n_S0_17, V1mV2_nm1_S0_17
        self.S0_17_TYPE = Fxp(val=None,signed=True,n_word=self.feature_word_length,n_frac=self.input_fractional_bits + self.input_padding)

        # quadratic features
        self.quadratic_full_word_length = 2*self.difference_word_length # = V1mV2_n_S0_12*V1mV2_n_S0_12 = V1mV2_n_squared_S1_24, V1mV2_nm1_squared_S1_24
        self.S1_24_TYPE = Fxp(val=None,signed=True,n_word=self.quadratic_full_word_length,n_frac=2*self.input_fractional_bits)
        self.S1_16_TYPE = Fxp(val=None,signed=True,n_word=self.feature_word_length,n_int = self.S1_24_TYPE.n_int)

        # cubic features
        self.cubic_min_value = (self.S0_12_TYPE.lower)**3
        self.S0_28_TYPE = Fxp(val=self.cubic_min_value,signed=True,n_frac = self.S1_16_TYPE.n_frac + self.S0_12_TYPE.n_frac) # this is more complicated because cubic features do not follow normal multiplication scaling rules (we can get away with less bits)
        self.S0_28_TYPE.set_val(0)

        # other nonlinear features
        self.SM1_24_TYPE = Fxp(val=None,signed=True,n_word = 2*self.SM1_12_TYPE.n_word,n_frac = self.SM1_12_TYPE.n_frac + self.SM1_12_TYPE.n_frac)
        self.SM1_18_TYPE = Fxp(val=None,signed=True,n_word = self.feature_word_length,n_frac = self.feature_word_length)
        self.S0_30_TYPE = Fxp(val=None,signed=True,n_word = self.SM1_18_TYPE.n_word + self.S0_12_TYPE.n_word, n_frac = self.SM1_17_TYPE.n_frac + self.S0_12_TYPE.n_frac)

        # desired trajectory
        self.desired_word_length = 16 # desired_V1_np1_SM1_16, desired_V2_np1_SM1_16, desired_V1_n_SM1_16, desired_V2_n_SM1_16
        self.desired_fractional_bits = 16

        # padded error
        self.error_word_length = 18 #error_V1_n_S0_17, error_V2_n_S0_17

        # weights (fixed point form of weights is determined dynamically during learning)
        self.weight_word_length = 18

        # limits of output signal in fixed point form
        self.DAC_FP_max = self.SM1_16_TYPE.upper
        self.DAC_FP_min = self.SM1_16_TYPE.lower
        
        # limits of output signal in signed integer form
        self.DAC_FP_max_int = self.S15_0_TYPE.upper
        self.DAC_FP_min_int = self.S15_0_TYPE.lower
        
        # specifying that there is no offset in the double scroll digital values
        # WIP: CHECK THIS VS ORIGINAL CODE
        self.V1_digital_mean = (self.S11_0_TYPE.upper-self.S11_0_TYPE.lower)/2
        self.V2_digital_mean = (self.S11_0_TYPE.upper-self.S11_0_TYPE.lower)/2
    
        # calculating the mean for the real values with gain
        self.V1_gain_offset = digital_to_gain(self.V1_digital_mean,self.U12_0_TYPE.upper,self.V_ADC_min,self.V_ADC_max)
        self.V2_gain_offset = digital_to_gain(self.V2_digital_mean,self.U12_0_TYPE.upper,self.V_ADC_min,self.V_ADC_max)
        
        # calculating the mean for the real values
        self.V1_real_offset = self.V1_gain_offset/self.V1_gain
        self.V2_real_offset =self. V2_gain_offset/self.V2_gain
        
        # calculating the mean for the fixed point values WIP: CHECK THIS VS ORIGINAL CODE
        self.V1_FP_mean = (self.V1_digital_mean + self.S11_0_TYPE.lower)/2**self.input_word_length
        self.V2_FP_mean = (self.V2_digital_mean + self.S11_0_TYPE.lower)/2**self.input_word_length
    
        
    def digital_to_fixed_point(self,
                               ADC_digital,
                               DAC_digital):
        
        """
        Convert digital ADC and DAC values into their fixed point form
        
        Parameters:
            ADC_digital (numpy.ndarray): Array containing time series of
            digital values of (V1,V2)
            
            DAC_digital (numpy.ndarray): Array containing time series of
            digital values of u
            
        Returns:
            doublescroll_soln_FP (numpy.ndarray): Array containing time series
            of fixed point values of (V1,V2)
            
            DAC_FP (numpy.ndarray): Array containing time series of fixed point
            values of u
        """
        
        # =============================================================================
        # ADC Data
        # =============================================================================
        
        # Store ADC values as 12-bit unsigned integers
        ADC_DATA_U12_0 = Fxp().like(self.U12_0_TYPE)
        ADC_DATA_U12_0.overflow = 'wrap'
        ADC_DATA_U12_0.set_val(ADC_digital)
        
        # Convert 12-bit unsigned integers to signed integers by subtracting
        # the mid point.
        ADC_DATA_S11_0 = Fxp().like(self.S11_0_TYPE)
        ADC_DATA_S11_0.overflow = 'wrap'
        ADC_DATA_S11_0.set_val(ADC_DATA_U12_0 + self.S11_0_TYPE.lower)
        
        # Convert the signed integers to fixed point form        
        ADC_DATA_SM1_12 = Fxp().like(self.SM1_12_TYPE)
        ADC_DATA_SM1_12.set_val(ADC_DATA_S11_0.raw(),raw=True)
        
        # =============================================================================
        # DAC Data
        # =============================================================================
    
        # Store DAC values as 16-bit unsigned integers
        DAC_U16_0 = Fxp().like(self.U16_0_TYPE)
        DAC_U16_0.overflow = 'wrap'
        DAC_U16_0.set_val(DAC_digital)
        
        # Convert 16-bit unsigned integers to signed integers by subtracting the
        # mid point        
        DAC_S15_0 = Fxp().like(self.S15_0_TYPE)
        DAC_S15_0.overflow = 'wrap'
        DAC_S15_0.set_val(DAC_U16_0 + int(self.S15_0_TYPE.lower))
        
        # Convert the signed integers to fixed point form        
        DAC_SM1_16 = Fxp().like(self.SM1_16_TYPE)
        DAC_SM1_16.set_val(DAC_S15_0.raw(),raw=True)
        
        # Converting fixed point to float for output
        doublescroll_soln_FP = ADC_DATA_SM1_12.astype(float)
        DAC_FP = DAC_SM1_16.astype(float)
        
        return doublescroll_soln_FP, DAC_FP

    def gain_to_real(self,
                     doublescroll_soln_gain):
        
        """
        Convert real values with gain applied to real values without gain
        
        Parameters:
            doublescroll_soln_gain (numpy.ndarray): Array containing time series 
            of real values of (V1,V2) with gain applied
            
        Returns:
            doublescroll_soln_real (numpy.ndarray): Array containing time series
            of real values of (V1,V2) without gain applied
            
        """
        # initialize array to zeros
        doublescroll_soln_real = np.zeros((2,np.shape(doublescroll_soln_gain)[1]))
        
        # Remove the gain and any offset from V1
        doublescroll_soln_real[0,:] = doublescroll_soln_gain[0,:]/self.V1_gain - self.V1_real_offset
        
        # Remove the gain and offset from V2
        doublescroll_soln_real[1,:] = doublescroll_soln_gain[1,:]/self.V2_gain - self.V2_real_offset
        
        return doublescroll_soln_real
    
    def print_verilog(self,
                      desired_padded_word_length,
                      desired_padding,
                      desired_error_padding,
                      K_FP,
                      F_TYPE,
                      Y_TYPE,
                      U_FULL_PRECISION_TYPE,
                      u_i_SX_16,
                      ridge_param,
                      W_n_frac,
                      W_u_inv_FP,
                      W_out_FP,
                      W_u_inv,
                      train_test_depth,
                      desired_depth,
                      write_depth,
                      control_wait):
        """
        Print FPGA code to paste into Verilog code, which consists of fixed
        point parameters and weights. Does not print any of structural code
        to implement the control algorithm.
    
        
        Parameters:
            desired_padded_word_length (int): Word length of the padded desired
            signal. Is tuned to match the number of fractional bits in Fipm and Kei
            in the control algorithm
            
            desired_padding (int): Number of zeros to pad on the right hand side
            of the 16-bit desired signal to reach the desired_padded_word_length
            
            desired_error_padding (int): Number of zeros to pad the desired signal 
            Ydipm to match the fixed point form of V1i for the error calculation 
            ei = V1i - Ydipm
            
            K_FP (Fxp object): The control gain in fixed point form
            
            F_TYPE (Fxp object): The unperturbed prediction (Fi+m) in fixed 
            point form
            
            Y_TYPE (Fxp object): The perturbed prediction (Yi+m) in fixed 
            point form
            
            U_FULL_PRECISION_TYPE (Fxp object): The fixed point form for ui
            after computing the control algorithm, before rounding and truncation
            to 16 bits
            
            u_i_SX_16 (Fxp object): The fixed point form for ui after rounding
            to 16 fractional bits
            
            ridge_param (float): The ridge parameter used in ridge regression
            
            W_n_frac (int): The number of fractional bits used by the fixed
            point form of the weights
            
            W_u_inv_FP (Fxp object): The fixed point form for Wu^-1
            
            W_out_FP (Fxp object): The fixed point form of the weights W
            
            W_u_inv (float): Wu^-1
            
            train_test_depth (int): The number of memory addresses used
            to store the training and testing perturbation
            
            desired_depth (int): The number of memory addresses used to store
            the desired signal
            
            write_depth (int): The number of memory addresses used to save
            time series data after control
            
            control_wait (int): The number of time steps to wait until the
            control perturbation is applied after the capture sequence has begun
        """
        
        # get the current date and time to mark when the weights were learned
        current_time = datetime.now()
        formatted_time = current_time.strftime("%m/%d/%Y %I:%M %p")
        
        # printing FPGA code
        print("//******************************************LOCAL PARAMETERS******************************************//\n")
        print(f"//Weights (trained on {formatted_time})")
        
        print("// Fixed-Point Parameters")
        
        print("// Linear Features")
        print("localparam input_word_length = " + str(self.input_word_length) + ";")
        print("localparam difference_word_length = " + str(self.difference_word_length) + ";")
        print("localparam output_word_length = " + str(self.output_word_length) + ";")
        print("localparam linear_word_length = " + str(self.linear_word_length) + ";")
        print("localparam feature_word_length = " + str(self.feature_word_length) + ";")
        print("reg [" + str(self.input_padding) + " - 1:0] input_padding = 0;")
        
        print("\n// Quadratic Features")
        print("localparam quadratic_full_word_length = " + str(self.quadratic_full_word_length) + ";")
        print("localparam quadratic_round_word_length = " + str(self.S1_16_TYPE.n_word) + ";")
        print("localparam quadratic_start = " + str(self.S1_24_TYPE.n_frac - self.S1_16_TYPE.n_frac) + ";")
        
        print("\n// Cubic Features")
        print("localparam cubic_full_word_length = " + str(self.S0_28_TYPE.n_word) + ";")
        print("localparam cubic_round_word_length = " + str(self.S0_17_TYPE.n_word) + ";")
        print("localparam cubic_start = " + str(self.S0_28_TYPE.n_frac - self.S0_17_TYPE.n_frac) + ";")
        
        print("\n// Desired")
        print("localparam desired_word_length = " + str(self.desired_word_length) + ";")
        print("localparam desired_padded_word_length = " + str(desired_padded_word_length) + ";")
        print("reg [" + str(desired_padding) + " - 1:0] desired_padding = 0;")
        print("reg [" + str(desired_error_padding) + " - 1:0] desired_error_padding = 0;")
        
        print("\n// Error")
        print("localparam error_word_length = " + str(self.error_word_length) + ";")
        
        print("\n// Weights")
        print("localparam weight_word_length = " + str(self.weight_word_length) + ";")
        print("localparam Kerror_word_length = " + str(K_FP.n_word + self.S0_17_TYPE.n_word) + ";") # WRONG
        print("localparam weight_fractional_bits = " + str(K_FP.n_frac) + ";")
        
        print("\n// Prediction")
        print("localparam unperturbed_prediction_word_length = " + str(F_TYPE.n_word) + ";")
        print("localparam prediction_word_length = " + str(Y_TYPE.n_word) + ";")
        print("localparam pred_start = " + str(Y_TYPE.n_frac - self.SM1_16_TYPE.n_frac) + ";")
        print("localparam prediction_trunc_word_length = " + str(self.SM1_16_TYPE.n_word) + ";")
        
        print("\n// Control Perturbation")
        print("localparam u_full_word_length = " + str(U_FULL_PRECISION_TYPE.n_word) + ";")
        print("localparam u_round_word_length = " + str(u_i_SX_16.n_word) + ";")
        print("localparam u_start = " + str(U_FULL_PRECISION_TYPE.n_frac - u_i_SX_16.n_frac) + ";")
        
        print("\n// Actuator Limits")
        print("localparam V_FP_MAX = " + str(int(self.DAC_FP_max_int)) + ";")
        print("localparam V_FP_MIN = " + str(int(self.DAC_FP_min_int)) + ";")
        
        print("\n// DAC")
        print("localparam DAC_MIN = " + str(0) + ";")
        print("localparam DAC_HALF = " + str(int(-self.S15_0_TYPE.lower)) + ";")
        print("localparam DAC_ZERO_CURRENT = " + str(33190) + ";") # measured zero current value
        print("localparam DAC_MAX = " + str(int(self.U16_0_TYPE.upper)) + ";")
        
        print("\n// RAM")
        print("localparam train_test_depth = " + str(train_test_depth) + ";")
        print("localparam train_test_depth_length = " + str(int(np.ceil(np.log2(train_test_depth)))) + ";")
        print("localparam desired_depth = " + str(desired_depth) + ";")
        print("localparam desired_depth_length = " + str(int(np.ceil(np.log2(desired_depth)))) + ";")
        print("localparam write_depth = " + str(write_depth) + ";")
        print("localparam write_depth_length = " + str(int(np.ceil(np.log2(write_depth)))) + ";")
        
        print("\n// CONTROL WAIT")
        print("localparam control_wait = " + str(control_wait) + ";")
        print("localparam train_test_wait = 1;")
        print("localparam total_wait_ctr_length = 12;")
        print("wire [total_wait_ctr_length - 1:0] total_wait/* synthesis keep */;")
        
        print("\n//******************************************MACROS******************************************//\n")
        
        print("// Truncate 'val' from 'start' to 'end' with rounding")
        print("`define round(val,start,end) $signed(val[start - 1] ? (val[start +: end] + 1'b1) : (val[start +: end]))")
        print("//`define round(val,start,end) $signed(val[start +: end])")
        print("`define trunc(val,start,end) $signed(val[start +: end])")
        
        print("\n//******************************************REG/WIRE DECLARATIONS******************************************//\n")
        
        print("\n// KEEP OTHER DECLARATIONS CONSTANT IN CODE\n")
            
        print(f"\n//Weights (trained on {formatted_time})")
        print("//ridge parameter: " + str(ridge_param))
        print("//weight fractional bits: " + str(W_n_frac))
        print("wire signed [weight_word_length - 1:0] W_u_i_inv = " + str(W_u_inv_FP[0].raw()) + "/* synthesis keep */; //" + str(W_u_inv_FP[0]))
        print("wire signed [weight_word_length - 1:0] W_u_i = " + str(W_out_FP[0].raw()) + "/* synthesis keep */; //" + str(W_out_FP[0]))
        print("wire signed [weight_word_length - 1:0] W_u_im1 = " + str(W_out_FP[1].raw()) + "/* synthesis keep */; //" + str(W_out_FP[1]))
        print("wire signed [weight_word_length - 1:0] W_V1_i = " + str(W_out_FP[2].raw()) + "/* synthesis keep */; //" + str(W_out_FP[2]))
        print("wire signed [weight_word_length - 1:0] W_V2_i = " + str(W_out_FP[3].raw()) + "/* synthesis keep */; //" + str(W_out_FP[3]))
        print("wire signed [weight_word_length - 1:0] W_V1_im1 = " + str(W_out_FP[4].raw()) + "/* synthesis keep */; //" + str(W_out_FP[4]))
        print("wire signed [weight_word_length - 1:0] W_V2_im1 = " + str(W_out_FP[5].raw()) + "/* synthesis keep */; //" + str(W_out_FP[5]))
        print("wire signed [weight_word_length - 1:0] W_V1mV2_i = " + str(W_out_FP[6].raw()) + "/* synthesis keep */; //" + str(W_out_FP[6]))
        print("wire signed [weight_word_length - 1:0] W_V1mV2_im1 = " + str(W_out_FP[7].raw()) + "/* synthesis keep */; //" + str(W_out_FP[7]))
        print("wire signed [weight_word_length - 1:0] W_V1mV2_i_cubed = " + str(W_out_FP[8].raw()) + "/* synthesis keep */; //" + str(W_out_FP[8]))
        print("wire signed [weight_word_length - 1:0] W_V1mV2_im1_cubed = " + str(W_out_FP[9].raw()) + "/* synthesis keep */; //" + str(W_out_FP[9]))
        
        print("\n// K")
        print("//wire signed [weight_word_length - 1:0] K00 = " + str(K_FP[0,0].raw()) + "/* synthesis keep */; //" + str(K_FP[0,0]))
        print("wire signed [weight_word_length - 1:0] K01 = " + str(K_FP[0,1].raw()) + "/* synthesis keep */; //" + str(K_FP[0,1]))
        print("reg signed [Kerror_word_length  - 1:0] Kerror = 0/* synthesis noprune */;")
        
        print("\n// Prediction")
        print("reg signed [unperturbed_prediction_word_length - 1:0] unperturbed_prediction = 0/* synthesis noprune */;")
        print("reg signed [prediction_word_length - 1:0] prediction = 0/* synthesis noprune */;")
        print("reg signed [prediction_trunc_word_length - 1:0] prediction_SM1_16/* synthesis noprune */;")
        
        print("\n// Control Perturbation")
        print("reg signed [u_full_word_length - 1:0] u_i_full = 0/* synthesis noprune */;")
        print("reg signed [u_round_word_length - 1:0] u_i_Sx_16 = 0/* synthesis noprune */;")
        
        
        print("\n//******************************************MODULE INSTANTIATION******************************************//\n")
        
        print("\n//******************************************STRUCTURAL CODING******************************************//\n")
        
        print("\n//******************************************MAIN CODE******************************************//\n")

    def compute_steady_states(self):
        """
        Compute nonzero unstable steady states for the double scroll system
        using the differential equations for the system. Computes real and 
        fixed point quantities, but only returns fixed point quantities
        
        Returns:
            V1ss_plus_FP (float): The positive unstable steady state for V1 in
            fixed point form
            
            V1ss_minus_FP (float): The negative unstable steady state for V1 in
            fixed point form
        """
        
        # defining double scroll parameters

        # Normalizations (Chang 1998)
        vd = 0.58 # voltage normalization
        ic = 0.25*10**(-3) # current normalization 
        tc = 2.35*10**(-5) # time normalization
        rc = 2.35*10**(3) # resistance normalization 

        rp = 1.511e4 # measured with multimeter
        rn0 = 2.61*1e3
        rn = 1./(1/rn0 - 1/rp)
        rm = 455
        rd = 7.86*1e3
        alpha = 11.6
        ir = 5.63*1e-9

        # defining steady state system
        def steadystate_real(V1ss):
            dVss = V1ss*(1.-rm/rn)
            gss = (dVss/rd)+2.*ir*np.sinh(alpha*dVss/vd)
            return ((V1ss/rn)-gss)

        # Solving for V1ss numerically, then using V1ss to find V2ss and Iss
        V1ss = fsolve(steadystate_real, 1.)
        V1ss = V1ss[0]
        V2ss = rm*V1ss/rn
        Iss = V1ss/rn
        yss = np.array([V1ss,V2ss,Iss])

        # convert fixed points to right form

        # add offset to fixed-points (WIP)
        V1_origin = self.V1_real_offset
        V1ss_minus = -V1ss + self.V1_real_offset
        V1ss_plus = V1ss + self.V1_real_offset
        V2ss_minus = -V2ss + self.V2_real_offset
        V2ss_plus = V2ss + self.V2_real_offset

        # Converting to gain
        V1ss_minus_gain = V1ss_minus*self.V1_gain
        V1ss_plus_gain = V1ss_plus*self.V1_gain
        V2ss_minus_gain = V2ss_minus*self.V2_gain
        V2ss_plus_gain = V2ss_plus*self.V2_gain

        # Convert to digital
        V1ss_minus_digital = gain_to_digital(V1ss_minus_gain,self.U12_0_TYPE.upper,self.V_ADC_min,self.V_ADC_max)
        V1ss_plus_digital = gain_to_digital(V1ss_plus_gain,self.U12_0_TYPE.upper,self.V_ADC_min,self.V_ADC_max)
        V2ss_minus_digital = gain_to_digital(V2ss_minus_gain,self.U12_0_TYPE.upper,self.V_ADC_min,self.V_ADC_max)
        V2ss_plus_digital = gain_to_digital(V2ss_plus_gain,self.U12_0_TYPE.upper,self.V_ADC_min,self.V_ADC_max)

        # Convert to S(-1).12
        V1ss_minus_FP = (V1ss_minus_digital + self.S11_0_TYPE.lower)/2**self.input_word_length
        V1ss_plus_FP = (V1ss_plus_digital + self.S11_0_TYPE.lower)/2**self.input_word_length
        V2ss_minus_FP = (V2ss_minus_digital + self.S11_0_TYPE.lower)/2**self.input_word_length
        V2ss_plus_FP = (V2ss_plus_digital + self.S11_0_TYPE.lower)/2**self.input_word_length
        
        # print results
        
        # nonzero steady state in real units
        print("\nTheoretical unstable steady states in real units:")
        print('V1 steady state (+): {:.3f} V'.format(V1ss))
        print('V2 steady state (+): {:.3f} V'.format(V2ss))
        print('I steady state (+): {:.4f} A'.format(Iss))
        
        # nonzero steady state in fixed-point form
        print("\nTheoretical unstable steady states in fixed-point form:")
        print('V1 steady state (+): {:.3f}'.format(V1ss_plus_FP))
        print('V1 steady state (-): {:.3f}'.format(V1ss_minus_FP))

        return V1ss_plus_FP,V1ss_minus_FP
    
    
# =============================================================================
# Desired Functions
# =============================================================================

def smooth_square_desired(num_pts,
                          V1_max,
                          V1_min, 
                          omega, 
                          b):
    """
    Create square wave with smooth transitions to use as a desired signal
    
    Parameters:
        num_pts (int): Length of the desired signal
        
        V1_max (float): Maximum value of the square wave
        
        V1_min (float): Minimum value of the square wave
        
        omega (float): Frequency of the square wave
        
        b (float): Smoothing parameter, which controls the rate of the 
        transition between the maximum and minimum. A larger b value means
        the transition is more smooth.
        
    Returns:
        x_desired (numpy.ndarray): Contains the desired signal
    """
    
    # The amplitude of the signal
    A = (np.abs(V1_max) + np.abs(V1_min))/2.
    
    # The vertical shift to apply to the signal
    shift = (V1_max + V1_min) / 2.
    
    # initializing desired signal to zeros
    x_desired = np.zeros(num_pts)
    
    # loop over each time step
    for i in range(num_pts):
        # compute the smooth square wave at each time step, and store it
        x_desired[i] = A*np.tanh(np.sin(omega*i)/b) + shift
    return x_desired

# =============================================================================
# Other helper functions
# =============================================================================

# converts 0-2.5V to 0 to 2^16-1 integer
def V_to_int16(V):
    """
    Converts a voltage from 0-2.5V to an integer from 0 to 2^16-1. This is used
    for converting the voltage output by the DAC to a digital value
    
    Parameters:
        V (float): Voltage from 0-2.5V
        
    Returns:
        int: digital value of DAC
    """
    return np.rint(V*26214.)

# convert voltage input to control circuit, to the current output of control circuit
def V_to_I(Vin, gain): # Vin must be 0V to 2.5V, gain must be 0 to 10
    """
    Converts a voltage from 0-2.5V into a current with a specified gain. The
    conversion is done using a custom built voltage to current converter.
    
    Parameters:
        Vin (float): Voltage from 0-2.5V
        
        gain (float): Gain of voltage to current conversion, can be 0-10
        
    Returns:
        Iout (float): output of voltage to current converter (A)
    """
    # voltage across 1Kohm shunt resistor
    Vout = gain*(-0.03224*Vin + 0.04082)
    # equivalent resistance of 1Kohm shunt resistor in parallel with 1Mohm resistor in oscilloscope
    Req = 999.001 
    # dividing by the resistance to obtain the current
    Iout = Vout / Req
    return Iout

    
def find_weight_fractional_bits(concatenated_weights,
                                weight_word_length):
    """
    Given the learned weights and control gain and the number of bits used
    to represent these quantities, find the optimal number of integer bits and
    fractional bits in fixed point form. The optimal number is defined by the
    minimum number of integer bits needed to represent all the weights and
    the control gain. This maximizes the precision of the weights.
    
    Parameters:
        concatenated_weights (numpy.ndarray): Array containing the weights and 
        control gain (W_out, K)
        
        weight_word_length (int): Number of bits used to represent the weights
        
    Returns:
        weight_integer_bits (int): Optimal number of integer bits to store the
        weights and control gain in fixed point form
            
        weight_fractional_bits (int): Optimal number of fractional bits to 
        store the weights and control gain in fixed point form
    """
    
    # finding the maximum and minimum of the weights and control gain
    max_weight = np.amax(concatenated_weights)
    min_weight = np.amin(concatenated_weights)
    
    # increase the number of integer bits until the fixed point representation
    # can store the maximum and minimum weights
    for num_integer in range(weight_word_length):
        max_FP = 0
        for i in range(weight_word_length): # loop through all the bits
            if i == 0:
                # compute the most negative fixed point number
                min_FP = -2**(num_integer)
            else:
                # compute the maximum possible fixed point number
                max_FP = max_FP + 2**(num_integer - i)
        # check if the minimum and maximum values can be stored
        if (min_FP < min_weight) and (max_FP > max_weight):
            # if the minimum and maximum values can be stored, break the loop
            weight_integer_bits = num_integer
            weight_fractional_bits = weight_word_length - weight_integer_bits - 1
            break
        # if there is no number of integer bits that can store the minimum
        # and maximum value given weight_word_length, print an error message
        elif (num_integer == weight_word_length - 1):
            print("Weight word length " + str(weight_word_length) + " may not be sufficient to store all weights!")
            raise SystemExit(0)
    return weight_integer_bits, weight_fractional_bits
    

def smooth_envelope(num_pts,
                    envelope_smoothing):
    """
    Create an envelope that starts at 0, smoothly approaches 1, and then
    smoothly approaches 0 again
    
    Parameters:
        num_pts (int): Number of points in the envelope
        
        envelope_smoothing (float): Smoothing parameter which determines 
        the transition speed. A higher value gives a smoother transition.
        
    Returns:
        envelope (numpy.ndarray): Array containing values in envelope
    """
    # defining the maximum and minimum values of the envelope
    envelope_max = 1.0
    envelope_min = -1.0
    
    # computing the frequency that makes the envelope not repeat
    omega = 2*np.pi*(1/num_pts)/2.
    
    # computing the amplitude of the envelope
    A = (np.abs(envelope_max) + np.abs(envelope_min))/2.
    
    # computing the vertical shift of the envelope
    shift = (envelope_max + envelope_min) / 2.
    
    # initializing the envelope to all zeros
    envelope = np.zeros(num_pts)
    
    # looping through each time step
    for i in range(num_pts):
        envelope[i] = A*np.tanh(np.sin(omega*i)/envelope_smoothing) + shift
        
    # force the last value to be zero, so the envelope repeats exactly
    envelope[-1] = 0
    
    return envelope

def digital_to_gain(unsigned_digital_values,
                    digital_max,
                    V_min,
                    V_max):
    """
    Convert digital values to real values with gain
    
    Parameters:
        unsigned_digital_values (numpy.ndarray): Array of digital values
        
        digital_max (int): Maximum digital value that can be represented for
        a given number of bits, e.g. 4095 for 12-bits
        
        V_min (float): The minimum voltage that the digital values represent,
        e.g. 0V for the DAC and -6.25V for the ADC
        
        V_max (float): The maximum voltage that the digital values represent,
        e.g. 2.5V for the DAC and +6.25V for the ADC
        
    Returns:
        gain_values (numpy.ndarray): Array of real values with gain applied
        that were converted from digital values
    """
    # normalize digital values to 1 (assumes unsigned)
    normalized_values = (unsigned_digital_values / digital_max)
    
    # scaling to correct voltage range
    scaled_values = normalized_values * (np.abs(V_min) + np.abs(V_max))
    
    # subtracting the minimum voltage
    gain_values = scaled_values + V_min
    
    return gain_values

def gain_to_digital(gain_values,
                    digital_max,
                    V_min,
                    V_max):
    """
    Convert real values with gain to digital values
    
    Parameters:
        gain_values (numpy.ndarray): Array of real values with gain
        
        digital_max (int): Maximum digital value that can be represented for
        a given number of bits, e.g. 4095 for 12-bits
        
        V_min (float): The minimum voltage that the digital values represent,
        e.g. 0V for the DAC and -6.25V for the ADC
        
        V_max (float): The maximum voltage that the digital values represent,
        e.g. 2.5V for the DAC and +6.25V for the ADC
        
    Returns:
        gain_values (numpy.ndarray): Array of real values with gain applied
        that were converted from digital values
    """
    # performing the reverse steps of the "digital_to_gain" function.
    scaled_values = gain_values - V_min
    normalized_values = scaled_values / (np.abs(V_min) + np.abs(V_max))
    unsigned_digital_values = normalized_values*digital_max
    return unsigned_digital_values

# note: this does not return an integer, it keeps all the precision
def fixed_point_to_digital(data,
                           n_frac,
                           n_word):
    """
    Convert signed fixed point values into unsigned digital scale. Note that
    this function does not convert to an integer, but it has the correct scale.
    This is to keep extra precision that may be in fixed point value.
    
    Parameters:
        data (numpy.ndarray): Array containing signed fixed point values
        
        n_frac (int): Number of fractional bits in the fixed point number
        
        n_word (int): Number of integer bits in the fixed point number
        
    Returns:
        unsigned_digital_values (numpy.ndarray): Array of digital values that 
        have been converted from fixed point
    """
    
    # return to integer scale, but don't round to nearest integer
    signed_digital_values = data*(2**n_frac)
    
    # convert from signed digital to unsigned digital values
    unsigned_digital_values = signed_digital_values + 2**(n_word - 1)
    
    return unsigned_digital_values

def binary_array_to_decimal(data,
                            signed = False):
    """
    Convert binary array that comes from reading the FPGA memory contents
    directly into signed or unsigned decimal values
    
    Parameters:
        data (numpy.ndarray): Binary array of unpacked bits in little-endian
        order
        
        signed (bool): Whether data is signed (True) or unsigned (False)
        
    Returns:
        data_signed (numpy.ndarray): Array of signed decimal values
    """
    
    # initialize output array as zeros
    data_signed = np.zeros(len(data))
    
    # loop through each element in data
    for i in range(len(data)):
        num = 0
        # loop through each bit in data[i]
        for j in range(len(data[i])):
            # treat the most significant bit differently whether the data 
            # is signed or unsigned
            if j == len(data[i]) - 1:
                if signed:
                    # if signed, subtract contribution from most significant bit
                    num = num - data[i][j]*2**j
                else:
                    # if unsigned, add contribution from most significant bit
                    num = num + data[i][j]*2**j
            else:
                # for all the other bits, add the contribution
                num = num + data[i][j]*2**j
        # store the resulting integer in the array for each i
        data_signed[i] = num
        
    return data_signed