# -*- coding: utf-8 -*-
"""
Created on Wednesday March 27 17:10 2024

Description:
Control double scroll chaotic circuit using next generation reservoir computing algorithm 
using a Max 10 10M50DAF484C6G FPGA on a Terasic Max 10 Plus development board. This code can be
used to reproduce all results in "Controlling Chaos Using Edge Computing Hardware" by Kent et al. 2024.

Notes: 
    -'im1' or 'ip1' in the code means i-1 and i+1 respectively

@author: Robert Kent
"""

# =============================================================================
# Preamble
# =============================================================================

# preamble
import numpy as np
import matplotlib.pyplot as plt
from matplotlib.ticker import MaxNLocator
import pandas as pd
import time
import pickle
from fxpmath import Fxp
import scipy.signal as signal
from helper import *
from sysidentpy.model_structure_selection import FROLS
from sysidentpy.basis_function._basis_function import Polynomial
from sysidentpy.utils.display_results import results

# Initialize fixed point class and physical quantities
FP = fixed_point_helper()

# Scaling of physical units
current_scaling = 1e6 # scale factor to put A into uA    
voltage_scaling = 1e3 # convert V to mV
time_scaling = 1e3 # convert s to ms

# specifying RAM sizes on FPGA
train_test_depth = 4000 # length of perturbation for training and testing
desired_depth = 4200 # length of desired signal during control
write_depth = 16000 # length of data to save and send back to the PC
control_wait = 1500 # number of points to wait until control begins
transient_end = 2000 # number of points to wait until controlled system is stabilized

# if you are connected to a Max 10 FPGA, set to True
fpga_connected = False

# Silence all deprecation warnings
import warnings
warnings.filterwarnings("ignore", category=FutureWarning)

#%%

# =============================================================================
# Generate perturbation for training and testing
# =============================================================================

# specifying the length of the perturbation to be stored in RAM on FPGA
train_test_start = 0
train_test_end = train_test_depth
train_test_pts = train_test_end - train_test_start
train_test_depth_length = int(np.ceil(np.log2(train_test_depth)))

# perturbation parameters
max_uniform = 1.5 # maximum value for uniform random noise
sample_f = 200e3 # sampling rate of ADC, DAC, and control algorithm
dt = 1/(sample_f) # sampling time
critical_f = 900 # critical frequency for butterworth digital low-pass filter
order = 10 # order of butterworth digital low-pass filter
envelope_smoothing = 0.1 # envelope smoothing parameter so perturbation can smoothly repeat
voltage_offset = 1.25 # output voltage range is 0-2.5V, so set offset to middle
skip_pts = 0 # number of points to cut off at the beginning of the perturbation

# write description for perturbation based on parameters
description = f"butter_lowpass_order{order}_criticalf{critical_f:.1e}_samplef{sample_f:.1e}".replace("+", "").replace(".","p")

# generate uniform random noise
uniform_noise = np.random.uniform(-max_uniform, max_uniform, train_test_pts + skip_pts)

# apply butterworth digital low-pass filter to uniform random noise
low_pass_filter = signal.butter(order, critical_f, btype='lowpass', fs=sample_f, output='sos')
filtered_noise = signal.sosfilt(low_pass_filter, uniform_noise)

# generate smooth envelope so that perturbation can repeat smoothly
envelope = smooth_envelope(train_test_pts, 
                           envelope_smoothing)
envelope_noise = filtered_noise[skip_pts:]*envelope + voltage_offset

# ensure that maximum and minimum values of the perturbation falls within DAC limits
v_train_test = np.clip(envelope_noise,FP.V_DAC_min,FP.V_DAC_max)

# convert perturbation into 16-bit unsigned integers that can be used by the DAC
int16_train_test_arr = np.zeros(train_test_pts) # array of 0-int16
for i in range(train_test_pts):
    int16_train_test_arr[i] = V_to_int16(v_train_test[i]) # int16 voltage at step i 0-2^16 - 1

# plot perturbation
fontsize = 20
fig, ax = plt.subplots(1, 1, figsize=(16,6))    
ax.plot(int16_train_test_arr, color = 'orange',linewidth=5)

# set x and y limits
ax.set_xlim(0,train_test_depth)
ax.set_ylim(0,FP.U16_0_TYPE.upper)

# set x and y labels
ax.set_xlabel('Memory address',fontsize=fontsize)
ax.set_ylabel('DAC Value',fontsize=fontsize)

# set title
ax.set_title('Perturbation for training and testing',fontsize=fontsize)

# adjust ticks
ax.tick_params(axis='both', labelsize=fontsize)    

# show plot
plt.show()
#%%

# =============================================================================
# Connect to FPGA and apply perturbation
# =============================================================================

if fpga_connected:
    
    # connect to FPGA
    fpga = FPGAController()
    
    # write perturbation to FPGA memory    
    fpga.write_train_test(int16_train_test_arr, FP.output_word_length)
    
    # wait 1s
    time.sleep(1)
    
    # additional delay in applying perturbation
    fpga.write_delay(0) # 0 means no delay added, 31 is maximum delay (31us)
    
    # set the controller to perturbation mode (closed loop control is 0)
    fpga.set_mode(1)
    
    # apply the perturbation
    fpga.control_switch(1)
    
    # wait 1s
    time.sleep(1)
    
    # read digital data stored in on chip memory
    V1_digital = fpga.load_data('ADC1', signed = False) # V1
    V2_digital = fpga.load_data('ADC2', signed = False) # V2
    DAC_digital = fpga.load_data('CTRL', signed = False) # control perturbation
    doublescroll_soln_digital = np.array([V1_digital,V2_digital]) # (V1,V2)
    
    # convert the digital values to real units with gain applied
    doublescroll_soln_gain = digital_to_gain(doublescroll_soln_digital,
                                             FP.U12_0_TYPE.upper,
                                             FP.V_ADC_min,
                                             FP.V_ADC_max)
    
    # convert real units with gain to just real units
    doublescroll_soln_real = FP.gain_to_real(doublescroll_soln_gain)
    DAC_voltage = digital_to_gain(DAC_digital,
                                  FP.U16_0_TYPE.upper,
                                  FP.V_DAC_min,
                                  FP.V_DAC_max)
    DAC_current = V_to_I(DAC_voltage,FP.gain)
    
    # convert digital values to fixed-point values
    doublescroll_soln_FP, DAC_FP = FP.digital_to_fixed_point(doublescroll_soln_digital,
                                                              DAC_digital)
    
    # stop applying the perturbation (when this is done is arbitrary)
    fpga.control_switch(0)
    
    # save perturbation data in a dictionary
    perturbation_dict = {'traintest_soln_FP': doublescroll_soln_FP,
                         'traintest_soln_gain': doublescroll_soln_gain,
                         'traintest_soln_real': doublescroll_soln_real,
                         'traintest_soln_digital': doublescroll_soln_digital,
                         'traintest_u_FP': DAC_FP,
                         'traintest_u_digital' : DAC_digital,
                         'traintest_u_voltage' : DAC_voltage,
                         'traintest_u_current' : DAC_current,
                         'u_description' : description}

else:
    # If there is no FPGA connected, load example perturbation data
    with open('example_data.pkl', 'rb') as fp:
        example_data_dict = pickle.load(fp)
    perturbation_dict = example_data_dict['perturbation_dict']    

#%%

# =============================================================================
# Plot perturbed double scroll system
# =============================================================================

# select maximum number of points to use in plotting
plot_pts = 4000

# create array of timesteps in real units (ms)
time_array = np.array([i for i in range(plot_pts)])*dt*time_scaling

# Plot Parameters
fontsize = 20
ticksize = 20
nbinsx_attractor = 6
nbinsy_attractor = 8
nbinsx_time_series = 9
nbinsy_time_series = 5
box_linewidth = 1.5
tick_width = 1
tick_length= 7
attractor_color = 'darkviolet'
transient_color = 'silver'

# creating figure
fig, axs = plt.subplots(2, 1, figsize=(14, 9))

# plot V1 and V2
axs[0].plot(time_array,perturbation_dict['traintest_soln_real'][0][:plot_pts],c='blue', label = r"$V_1$")
axs[0].plot(time_array,perturbation_dict['traintest_soln_real'][1][:plot_pts],c='red',label = r"$V_2$")

# plotting control current
axs[1].plot(time_array,perturbation_dict['traintest_u_current'][:plot_pts]*current_scaling,c='darkorange',label = "u")

# set y limits
axs[0].set_ylim(FP.V1_real_min,FP.V1_real_max)
axs[1].set_ylim(FP.I_min*current_scaling+0.05*FP.I_min*current_scaling,
                FP.I_max*current_scaling+0.05*FP.I_max*current_scaling)

# set x and y labels
axs[0].set_ylabel("Voltage (V)",fontsize = fontsize)
axs[1].set_ylabel(r'Control Current $(\mu A)$',fontsize = fontsize)
axs[1].set_xlabel('time (ms)',fontsize=fontsize)


# adjust ticks for for bottom plots
for ax in axs.flat:
    ax.tick_params(axis='both', labelsize=ticksize)
    ax.xaxis.set_major_locator(MaxNLocator(nbins=nbinsx_time_series))
    ax.yaxis.set_major_locator(MaxNLocator(nbins=nbinsy_time_series))
    ax.tick_params(axis = 'x', which='minor', direction='out', length=tick_length, width=tick_width, colors='black')
    ax.tick_params(axis = 'x', which='major', direction='out', length=tick_length, width=tick_width, colors='black')
    ax.tick_params(axis = 'y', which='minor', direction='out', length=tick_length, width=tick_width, colors='black')
    ax.tick_params(axis = 'y', which='major', direction='out', length=tick_length, width=tick_width, colors='black')
    
    # add boxes around plots
    ax.patch.set_edgecolor('black')  
    ax.patch.set_linewidth(box_linewidth)

# remove x ticks for top plot
axs[0].set_xticks([])

# setting title
fig.suptitle("Perturbed data for training and testing",fontsize=fontsize)

# adding a legend
fig.legend(bbox_to_anchor=(1.02,0.8),fontsize=fontsize)

# adjust subplot spacing
plt.subplots_adjust(hspace=0.1)

# plotting figure
plt.show()


# plot attractor
fig, ax = plt.subplots(1, 1, figsize=(8, 8))
ax.plot(perturbation_dict['traintest_soln_real'][0][:plot_pts],perturbation_dict['traintest_soln_real'][1][:plot_pts], c = attractor_color,zorder = 20)

# setting x and y limits
ax.set_xlim(FP.V1_real_min,FP.V1_real_max)
ax.set_ylim(FP.V2_real_min,FP.V2_real_max)

# setting x and y labels
ax.set_xlabel(r"$V_1$ (V)",fontsize = fontsize)
ax.set_ylabel(r"$V_2$ (V)",fontsize = fontsize)

# adjust ticks
ax.tick_params(axis='both', labelsize=ticksize)
ax.xaxis.set_major_locator(MaxNLocator(nbins=nbinsx_attractor))
ax.yaxis.set_major_locator(MaxNLocator(nbins=nbinsy_attractor))

# draw box around plot
ax.patch.set_edgecolor('black')  
ax.patch.set_linewidth(box_linewidth)

# setting title
ax.set_title("perturbed attractor for training and testing",fontsize=fontsize)

# plotting figure
plt.show()

#%%
# =============================================================================
# Training and Testing the NG-RC Model
# =============================================================================

def training_and_testing(ridge_param,
                         train_pts,
                         test_pts,
                         perturbed_data,
                         perturbed_data_real,
                         skip_pts=1,
                         m=1,
                         u_delay=0,
                         show_plots = False,
                         print_fpga_code = False):
    """
    Makes features out of perturbed data in fixed point form, then learns weights
    using ridge regression, then uses learned weights to make predictions on
    training and testing data and evaluate the RMSE, then prints FPGA code.
    
    Parameters:
        ridge_param (float): Ridge parameter to be used in ridge regression
        
        train_pts (int): Number of points to use in training  
        
        test_pts (int): Number of points to use in testing
        
        perturbed_data (numpy.ndarray): Array containing time-series data of V1,V2, and u
        in fixed-point form
        
        perturbed_data_real (numpy.ndarray): Array containing time-series data of V1,V2, and u
        in real units (Volts and Amps)
        
        skip_pts (int): Number of points to skip in time-series data before training data starts
        
        m (int): Number of time-steps ahead that the model is trained to predict, e.g. if m=1 
        then features of V1i,V2i, and ui are used to predict V1i+1
        
        u_delay (int): Manual delay of ui in the features
        
        show_plots (bool): Whether to show additional plots (True) or not (False)
        
        print_fpga_code (bool): Whether to print FPGA code (True) or not (False)
        
        
    Returns:
        train_test_dict (dict): A dictionary containing the following keys:
            - 'ridge_param' : Ridge parameter to be used in ridge regression
            - 'train_rmse' : RMSE of prediction on training data
            - 'train_rmse_real' : RMSE of prediction on training data (real units)
            - 'test_rmse' : RMSE of prediction on testing data
            - 'test_rmse_real' : RMSE of prediction on testing data (real units)
            - 'W_out' : Learned weights
            - 'W_out_FP' : Learned weights as Fxp object
            - 'W_u_inv' : Learned Wu^-1
            - 'W_u_inv_FP' : Learned Wu^-1 as Fxp object
            - 'W_n_frac' : Number of fractional bits used by the weights
    """
    
    # =============================================================================
    # Defining control and NG-RC parameters
    # =============================================================================
    
    # defining control gain matrix    
    K = np.zeros((1,2))
    # The gain is chosen to be -3.0 to represent the largest magnitude it may ever become,
    # higher gains like this are used in linear control, whereas the NG-RC controllers fail
    # if the gain is |K|>1.0. The fixed-point form for K is chosen based on the maximum value
    # needed to be represented, so that is why the maximum magnitude is chosen here.
    K[0,0] = -3.0 # gain for V1
    K[0,1] = 0.0 # gain for V2 (not used in this experiment)
    
    # defining useful values for slicing the data    
    skip_train_pts = skip_pts + train_pts # all skip points and train points
    maxtime_pts = skip_pts + train_pts + test_pts # all skip,train, and test points
    
    # defining NG-RC parameters
    d = 2 # input dimension: V1,V2
    k = 2 # number of time delay taps: i,i-1
    dlin = k*d # number of linear features for X: V1i,V2i,V1im1,V2im1
    dlinu = 2 # number of linear features for u: ui,uim1
    # number of nonlinear features: 
    # (V1i-V2i)V1im1^2, (V1im1-V2im1)V1i^2, (V1i-V2i)^3,(V1im1-V2im1)^3
    dtot = dlin + dlinu + 4 
    
    # =============================================================================
    # Linear feature creation
    # =============================================================================
    
    # defining feature vectors
    O_lin = np.zeros((dlin,maxtime_pts)) # linear feature vector
    O_lin_real = np.zeros((dlin,maxtime_pts)) # linear feature vector (real units)
    O_lin_u = np.zeros((dlinu,maxtime_pts)) # linear feature vector (u)
    O_total = np.zeros((dtot,maxtime_pts)) # total feature vector

    # initializing data (fixed point)
    doublescroll_soln = perturbed_data[:2,u_delay:maxtime_pts+u_delay]
    DAC = perturbed_data[2,:maxtime_pts]

    # initializing data (real values)
    doublescroll_soln_real = perturbed_data_real[:2,u_delay:maxtime_pts+u_delay]
    DAC_current = perturbed_data_real[2,:maxtime_pts]
    
    # create delayed copy of (V1,V2)    
    doublescroll_soln_delayed = np.insert(doublescroll_soln[:,:-1],0,0,axis=1)

    # fill in the linear part of the feature vector for all times
    for delay in range(k):
        for j in range(delay,maxtime_pts):
            O_lin[d*delay:d*(delay+1),j]=doublescroll_soln[:d,j-delay]
            O_lin_real[d*delay:d*(delay+1),j]=doublescroll_soln_real[:d,j-delay]


    # fill in the u feature vector for all times
    O_lin_u[0,:] = DAC.astype(float) # ui
    O_lin_u[1,1:] = DAC[:-1].astype(float) # uim1

    # =============================================================================
    # Nonlinear feature creation using fxpmath package
    # =============================================================================

    # Computing V1i-V2i (13 bits)
    V1mV2_i_S0_12 = Fxp().like(FP.S0_12_TYPE)
    V1mV2_i_S0_12.set_val(doublescroll_soln[0,:] - doublescroll_soln[1,:])

    # Computing V1im1-V2im1 (13 bits)
    V1mV2_im1_S0_12 = Fxp().like(FP.S0_12_TYPE)
    V1mV2_im1_S0_12.set_val(doublescroll_soln_delayed[0,:] - doublescroll_soln_delayed[1,:])

    # Computing (V1i-V2i)^2 (26 bits)
    V1mV2_i_squared_S1_24 = Fxp().like(FP.S1_24_TYPE)
    V1mV2_i_squared_S1_24.set_val(V1mV2_i_S0_12*V1mV2_i_S0_12)

    # Computing (V1im1-V2im1)^2 (26 bits)
    V1mV2_im1_squared_S1_24 = Fxp().like(FP.S1_24_TYPE)
    V1mV2_im1_squared_S1_24.set_val(V1mV2_im1_S0_12*V1mV2_im1_S0_12)

    # Rounding (V1i-V2i)^2 (18 bits)
    V1mV2_i_squared_S1_16 = Fxp().like(FP.S1_16_TYPE)
    V1mV2_i_squared_S1_16.rounding = 'around'
    V1mV2_i_squared_S1_16.set_val(V1mV2_i_squared_S1_24)

    # Rounding (V1im1-V2im1)^2 (18 bits)
    V1mV2_im1_squared_S1_16 = Fxp().like(FP.S1_16_TYPE)
    V1mV2_im1_squared_S1_16.rounding = 'around'
    V1mV2_im1_squared_S1_16.set_val(V1mV2_im1_squared_S1_24)

    # Computing (V1i-V2i)^3 (29 bits: extra integer bits are not needed)
    V1mV2_i_cubed_S0_28 = Fxp().like(FP.S0_28_TYPE)
    V1mV2_i_cubed_S0_28.set_val(V1mV2_i_squared_S1_16*V1mV2_i_S0_12)
    
    # Computing (V1im1-V2im1)^3 (29 bits: extra integer bits are not needed)
    V1mV2_im1_cubed_S0_28 = Fxp().like(FP.S0_28_TYPE)
    V1mV2_im1_cubed_S0_28.set_val(V1mV2_im1_squared_S1_16*V1mV2_im1_S0_12)

    # Rounding (V1i-V2i)^3 (18 bits)
    V1mV2_i_cubed_S0_17 = Fxp().like(FP.S0_17_TYPE)
    V1mV2_i_cubed_S0_17.rounding = 'around'
    V1mV2_i_cubed_S0_17.set_val(V1mV2_i_cubed_S0_28)

    # Rounding (V1im1-V2im1)^3 (18 bits)
    V1mV2_im1_cubed_S0_17 = Fxp().like(FP.S0_17_TYPE)
    V1mV2_im1_cubed_S0_17.rounding = 'around'
    V1mV2_im1_cubed_S0_17.set_val(V1mV2_im1_cubed_S0_28)
    
    # Computing V1im1^2 (24 bits)
    V1_im1_squared_SM1_24 = Fxp().like(FP.SM1_24_TYPE)
    V1_im1_squared_SM1_24.set_val(doublescroll_soln_delayed[0,:]*doublescroll_soln_delayed[0,:])
    
    # Rounding V1im1^2 (18 bits)
    V1_im1_squared_SM1_18 = Fxp().like(FP.SM1_18_TYPE)
    V1_im1_squared_SM1_18.rounding = 'around'
    V1_im1_squared_SM1_18.set_val(V1_im1_squared_SM1_24)
    
    # Computing (V1i-V2i)*V1im1^2 (31 bits)
    NONLIN1_S0_30 = Fxp().like(FP.S0_30_TYPE)
    NONLIN1_S0_30.set_val(V1_im1_squared_SM1_18*V1mV2_i_S0_12)
    
    # Rounding (V1i-V2i)*V1im1^2 (18 bits)
    NONLIN1_S0_17 = Fxp().like(FP.S0_17_TYPE)
    NONLIN1_S0_17.rounding = 'around'
    NONLIN1_S0_17.set_val(NONLIN1_S0_30)
    
    # Computing V1i^2 (24 bits)    
    V1_i_squared_SM1_24 = Fxp().like(FP.SM1_24_TYPE)
    V1_i_squared_SM1_24.set_val(doublescroll_soln[0,:]*doublescroll_soln[0,:])
    
    # Rounding V1i^2 (18 bits)  
    V1_i_squared_SM1_18 = Fxp().like(FP.SM1_18_TYPE)
    V1_i_squared_SM1_18.rounding = 'around'
    V1_i_squared_SM1_18.set_val(V1_i_squared_SM1_24)
    
    # Computing (V1im1-V2im1)*V1i^2 (31 bits)
    NONLIN2_S0_30 = Fxp().like(FP.S0_30_TYPE)
    NONLIN2_S0_30.set_val(V1_i_squared_SM1_18*V1mV2_im1_S0_12)
    
    # Rounding (V1im1-V2im1)*V1i^2 (18 bits)
    NONLIN2_S0_17 = Fxp().like(FP.S0_17_TYPE)
    NONLIN2_S0_17.rounding = 'around'
    NONLIN2_S0_17.set_val(NONLIN2_S0_30)
    
    # Putting all features into total feature vector
    O_total[:dlinu,:] = O_lin_u # ui, uim1
    O_total[dlinu:dlinu + dlin,:] = O_lin # V1i,V2i,V1im1,V2im1
    O_total[dlinu+dlin,:] = NONLIN1_S0_17.astype(float) # (V1i-V2i)*V1im1^2
    O_total[dlinu+dlin+1,:] = NONLIN2_S0_17.astype(float) # (V1im1-V2im1)*V1i^2
    O_total[dlinu+dlin+2,:] = V1mV2_i_cubed_S0_17.astype(float) # (V1i-V2i)^3
    O_total[dlinu+dlin+3,:] = V1mV2_im1_cubed_S0_17.astype(float) # (V1im1-V2im1)^3

    # =============================================================================
    # Learning weights
    # =============================================================================

    # separating the total feature vector for training and testing
    O_train = O_total[:,skip_pts:skip_train_pts-m] # training feature vector
    O_test = O_total[:,skip_train_pts:maxtime_pts-m] # testing feature vector
    
    # creating truth labels for training and testing sets
    Y_train = O_lin[0,skip_pts+m:skip_train_pts] 
    Y_test = O_lin[0,skip_train_pts+m:maxtime_pts]
    
    # creating truth labels for training and testing sets (in real units)
    Y_train_real = O_lin_real[0,skip_pts+m:skip_train_pts]
    Y_test_real = O_lin_real[0,skip_train_pts+m:maxtime_pts]
    
    # computing output weights using ridge regression
    W_out = (Y_train) @ O_train.T @ np.linalg.pinv(O_train @ O_train.T + ridge_param*np.identity(dtot))

    # =============================================================================
    # Converting weights to fixed-point using fxpmath
    # =============================================================================

    # concatenating the learned weights and the control gain,
    # as they share the same fixed point representation
    concatenated_weights = np.concatenate([W_out, K], axis=None)
    
    # finding the optimal number of integer and fractional bits
    # for a 18-bit representation of the weights
    W_n_int, W_n_frac = find_weight_fractional_bits(concatenated_weights,
                                                    FP.weight_word_length)
    
    # discretizing the weights to 18-bit
    W_out_FP = Fxp(val=None,signed = True, n_word=FP.weight_word_length, n_frac = W_n_frac, n_int = W_n_int)
    W_out_FP.rounding = 'around'
    W_out_FP.set_val(W_out)
    
    # discretizing the control gain to 18-bit using the same representation
    # as the weights
    K_FP = Fxp().like(W_out_FP)
    K_FP.set_val(K)

    # finding Wu^-1, which is just 1/Wu in the case where only 
    # one variable has feedback. It is put in a list to be used
    # in matrix multiplication.
    W_u_inv = [1./W_out[0]]
    
    # finding the optimal number of integer and fractional bits
    # for Wu^-1.
    W_u_inv_int, W_u_inv_frac = find_weight_fractional_bits(W_u_inv, FP.weight_word_length) # can be a different word length to before
    
    # discretizing Wu^-1 to 18-bit
    W_u_inv_FP = Fxp(val=None,signed = True, n_word=FP.weight_word_length, n_frac = W_u_inv_frac, n_int = W_u_inv_int)
    W_u_inv_FP.rounding = 'around'
    W_u_inv_FP.set_val(W_u_inv)

    # convert weights and control gain to floating point
    # to be used later
    W_out = W_out_FP.astype(float)
    W_u_inv = W_u_inv_FP.astype(float)
    K = K_FP.astype(float)

    # separating learned weights into Wu and WF
    W_u = W_out[0]
    W_F = W_out[1:]
    
    # determining the maximum value of the unperturbed prediction (WF @ OF),
    # which will allow us to find the appropriate fixed point representation.
    # We do this by making each feature and weight the most negative possible value,
    # which will allow us to determine the maximum possible magnitude we need to 
    # represent using fixed point. The contributions from each feature are as follows:
    # (1) uim1 (16-bit), (4) V1i,V2i,V1im1,V2im1 (12-bit), (4) (V1i-V2i)*V1im1^2,(V1im1-V2im1)*V1i^2,(V1i-V2i)^3,(V1im1-V2im1)^3 (18-bit)
    unperturbed_prediction_max = 1*FP.SM1_16_TYPE.lower*W_out_FP.lower + 4*FP.SM1_12_TYPE.lower*W_out_FP.lower + 4*FP.S0_17_TYPE.lower*W_out_FP.lower

    # getting the optimal fixed point form for the unperturbed prediction
    F_TYPE = Fxp(unperturbed_prediction_max,n_frac = FP.S0_17_TYPE.n_frac + W_out_FP.n_frac)
    F_TYPE.set_val(0)

    # for the perturbed prediction maximum, which now includes Wu*ui, we need to
    # add the following contribution:
    # (1) ui (16-bit)
    perturbed_prediction_max = unperturbed_prediction_max + 1*FP.SM1_16_TYPE.lower*W_out_FP.lower
    
    # getting the optimal fixed point form for the perturbed prediction
    Y_TYPE = Fxp(perturbed_prediction_max,n_frac = FP.S0_17_TYPE.n_frac + W_out_FP.n_frac)
    Y_TYPE.set_val(0)
    
    # finding the maximum possible value of ui after computing the control law.
    # Wu^-1 (18-bit) * (Ydipm (16-bit) - unperturbed_prediction_max (?-bit) - K (18-bit) * ei (13-bit) )
    u_max_FP = W_u_inv_FP.lower*(FP.SM1_16_TYPE.lower - unperturbed_prediction_max - K_FP.lower*FP.S0_12_TYPE.lower)
    
    # getting the optimal fixed point form for ui
    U_FULL_PRECISION_TYPE = Fxp(u_max_FP,n_frac = F_TYPE.n_frac + W_u_inv_FP.n_frac)
    U_FULL_PRECISION_TYPE.set_val(0)
    
    # truncating the fractional bits of ui to 16-bits to match original precision.
    # on the FPGA, the integer bits are then truncated using thresholding to get
    # the 16-bit output needed by the DAC.
    u_i_SX_16 = Fxp(val=None,n_int = U_FULL_PRECISION_TYPE.n_int,n_frac = FP.SM1_16_TYPE.n_frac)

    # padding the desired signal Ydipm with zeros in the fractional bits to match the number of 
    # fractional bits in Fipm and Kei, which is needed to evaluate the control algorithm
    desired_padded_word_length = FP.SM1_17_TYPE.n_word + W_out_FP.n_frac # total word length
    desired_padding = desired_padded_word_length - FP.SM1_16_TYPE.n_word # number of zeros
    
    # padding the desired signal Ydipm to match the fixed point form of V1i
    # for the error calculation ei = V1i - Ydipm
    desired_error_padding = FP.S0_17_TYPE.n_frac - FP.SM1_16_TYPE.n_frac # number of zeros

    # =============================================================================
    # Making predictions
    # =============================================================================

    # make prediction on training data
    train_prediction = W_out_FP.astype(float) @ O_train # prediction with Wu*ui

    # compute the training RMSE
    train_rmse = np.sqrt(np.mean((Y_train-train_prediction)**2)) # prediction with Wu*ui

    # make prediction on testing data
    test_prediction = W_out.astype(float) @ O_test # prediction with Wu*ui

    # compute the testing RMSE
    test_rmse = np.sqrt(np.mean((Y_test-test_prediction)**2)) # prediction with Wu*ui

    # =============================================================================
    # Converting predictions to real units (V)
    # =============================================================================

    # converting training prediction from fixed point to digital, then digital to real with gain, then real with gain to real.
    train_prediction_digital = fixed_point_to_digital(train_prediction,FP.SM1_12_TYPE.n_frac,FP.SM1_12_TYPE.n_word)
    train_prediction_gain = digital_to_gain(train_prediction_digital, FP.U12_0_TYPE.upper,FP.V_ADC_min,FP.V_ADC_max)
    train_prediction_real = train_prediction_gain/FP.V1_gain - FP.V1_real_offset
    
    # compute the training RMSE in real units (V)    
    train_rmse_real = np.sqrt(np.mean((Y_train_real-train_prediction_real)**2))

    # converting testing prediction from fixed point to digital, then digital to real with gain, then real with gain to real.
    test_prediction_digital = fixed_point_to_digital(test_prediction,FP.SM1_12_TYPE.n_frac,FP.SM1_12_TYPE.n_word)
    test_prediction_gain = digital_to_gain(test_prediction_digital, FP.U12_0_TYPE.upper,FP.V_ADC_min,FP.V_ADC_max)
    test_prediction_real = test_prediction_gain/FP.V1_gain - FP.V1_real_offset

    # computing the testing RMSE in real units (V)
    test_rmse_real = np.sqrt(np.mean((Y_test_real-test_prediction_real)**2))

    # =============================================================================
    # Plotting training and testing predictions
    # =============================================================================
    
    if show_plots:
        
        # =============================================================================
        # Plotting training attractor (fixed point)
        # =============================================================================
        
        fig, ax = plt.subplots(1, 1, figsize=(8, 8))
        ax.plot(O_lin[0,skip_pts+m:skip_train_pts],O_lin[1,skip_pts+m:skip_train_pts], c = 'C0',zorder = 1, label = 'true')
        ax.plot(train_prediction,O_lin[1,skip_pts+m:skip_train_pts], c = 'red',zorder = 10, label = 'predicted')

        # setting x and y labels
        ax.set_xlabel(r"$V_1$ (FP)",fontsize = fontsize)
        ax.set_ylabel(r"$V_2$ (FP)",fontsize = fontsize)
        
        # add title
        ax.set_title('Prediction on training data',fontsize=fontsize)
        
        # setting x and y limits
        ax.set_xlim(FP.SM1_12_TYPE.lower,FP.SM1_12_TYPE.upper)
        ax.set_ylim(FP.SM1_12_TYPE.lower,FP.SM1_12_TYPE.upper)

        # adjust ticks
        ax.tick_params(axis='both', labelsize=ticksize)
        ax.xaxis.set_major_locator(MaxNLocator(nbins=nbinsx_attractor))
        ax.yaxis.set_major_locator(MaxNLocator(nbins=nbinsy_attractor))

        # draw box around plot
        ax.patch.set_edgecolor('black')  
        ax.patch.set_linewidth(box_linewidth)
        
        # add legend
        ax.legend(bbox_to_anchor = (1,1),fontsize=fontsize)

        # plotting figure
        plt.show()
        
        # =============================================================================
        # Plotting training time-series (fixed point)
        # =============================================================================

        # subplots
        fig, axs = plt.subplots(3, 1, figsize=(10,8))
        axs[0].plot(O_lin[0,skip_pts+m:skip_train_pts], c = 'C0', label = r"true $V_1$")
        axs[0].plot(train_prediction, c = "red", label = r"predicted $V_1$")
        axs[1].plot(O_lin[1,skip_pts+m:skip_train_pts], c = 'blue',label = r"true $V_2$")
        axs[2].plot(O_lin_u[0,skip_pts+m:skip_train_pts], label = "u", c = "darkorange")
        
        # set x and y labels
        axs[0].set_ylabel(r"$V_1$ FP",fontsize=fontsize)
        axs[1].set_ylabel(r"$V_2$ FP",fontsize=fontsize)
        axs[2].set_ylabel(r"u FP",fontsize=fontsize)
        axs[2].set_xlabel("time step",fontsize=fontsize)
        
        for ax in axs.flat:
            # adjust ticks
            ax.tick_params(axis='both', labelsize=ticksize)
            ax.xaxis.set_major_locator(MaxNLocator(nbins=nbinsx_time_series))
            ax.yaxis.set_major_locator(MaxNLocator(nbins=nbinsy_time_series))

            # draw box around plot
            ax.patch.set_edgecolor('black')  
            ax.patch.set_linewidth(box_linewidth)
            
        # remove ticks on V1 and V2 plots
        axs[0].set_xticks([])
        axs[1].set_xticks([])
        
        # adjust subplot spacing
        plt.subplots_adjust(hspace=0.1)
        
        # set title
        fig.suptitle("Prediction on training data",fontsize=fontsize)
        
        # add legend
        fig.legend(bbox_to_anchor = (1.2,0.9),fontsize = fontsize)
        
        # show plot
        plt.show()
        
        # =============================================================================
        # Plotting training error (fixed point)
        # =============================================================================

        fig, ax = plt.subplots(1, 1, figsize=(10,6))
        ax.plot(train_prediction - O_lin[0,skip_pts+m:skip_train_pts],c='C0')
        
        # set x and y labels
        ax.set_ylabel("error FP",fontsize=fontsize)
        ax.set_xlabel("time step",fontsize=fontsize)
        
        # adjust ticks
        ax.tick_params(axis='both', labelsize=ticksize)
        ax.xaxis.set_major_locator(MaxNLocator(nbins=nbinsx_time_series))
        ax.yaxis.set_major_locator(MaxNLocator(nbins=nbinsy_time_series))

        # draw box around plot
        ax.patch.set_edgecolor('black')  
        ax.patch.set_linewidth(box_linewidth)
        
        # add title
        ax.set_title("Training error",fontsize=fontsize)
        
        # show plot
        plt.show()
        
        # WIP below
        
        # =============================================================================
        # Plotting testing attractor (fixed point)
        # =============================================================================
        
        fig, ax = plt.subplots(1, 1, figsize=(8, 8))
        ax.plot(O_lin[0,skip_train_pts+m:maxtime_pts],O_lin[1,skip_train_pts+m:maxtime_pts], c = 'C0',zorder = 1, label = 'true')
        ax.plot(test_prediction,O_lin[1,skip_train_pts+m:maxtime_pts], c = 'red',zorder = 10, label = 'predicted')

        # setting x and y labels
        ax.set_xlabel(r"$V_1$ (FP)",fontsize = fontsize)
        ax.set_ylabel(r"$V_2$ (FP)",fontsize = fontsize)
        
        # add title
        ax.set_title('Prediction on testing data',fontsize=fontsize)
        
        # setting x and y limits
        ax.set_xlim(FP.SM1_12_TYPE.lower,FP.SM1_12_TYPE.upper)
        ax.set_ylim(FP.SM1_12_TYPE.lower,FP.SM1_12_TYPE.upper)

        # adjust ticks
        ax.tick_params(axis='both', labelsize=ticksize)
        ax.xaxis.set_major_locator(MaxNLocator(nbins=nbinsx_attractor))
        ax.yaxis.set_major_locator(MaxNLocator(nbins=nbinsy_attractor))

        # draw box around plot
        ax.patch.set_edgecolor('black')  
        ax.patch.set_linewidth(box_linewidth)
        
        # add legend
        ax.legend(bbox_to_anchor = (1,1),fontsize=fontsize)

        # plotting figure
        plt.show()
        
        # =============================================================================
        # Plotting testing time-series (fixed point)
        # =============================================================================

        # subplots
        fig, axs = plt.subplots(3, 1, figsize=(10,8))
        axs[0].plot(O_lin[0,skip_train_pts+m:maxtime_pts], c = 'C0', label = r"true $V_1$")
        axs[0].plot(test_prediction, c = "red", label = r"predicted $V_1$")
        axs[1].plot(O_lin[1,skip_train_pts+m:maxtime_pts], c = 'blue',label = r"true $V_2$")
        axs[2].plot(O_lin_u[0,skip_train_pts+m:maxtime_pts], label = "u", c = "darkorange")
        
        # set x and y labels
        axs[0].set_ylabel(r"$V_1$ FP",fontsize=fontsize)
        axs[1].set_ylabel(r"$V_2$ FP",fontsize=fontsize)
        axs[2].set_ylabel(r"u FP",fontsize=fontsize)
        axs[2].set_xlabel("time step",fontsize=fontsize)
        
        for ax in axs.flat:
            # adjust ticks
            ax.tick_params(axis='both', labelsize=ticksize)
            ax.xaxis.set_major_locator(MaxNLocator(nbins=nbinsx_time_series))
            ax.yaxis.set_major_locator(MaxNLocator(nbins=nbinsy_time_series))

            # draw box around plot
            ax.patch.set_edgecolor('black')  
            ax.patch.set_linewidth(box_linewidth)
            
        # remove ticks on V1 and V2 plots
        axs[0].set_xticks([])
        axs[1].set_xticks([])
        
        # adjust subplot spacing
        plt.subplots_adjust(hspace=0.1)
        
        # set title
        fig.suptitle("Prediction on testing data",fontsize=fontsize)
        
        # add legend
        fig.legend(bbox_to_anchor = (1.2,0.9),fontsize = fontsize)
        
        # show plot
        plt.show()
        
        # =============================================================================
        # Plotting testing error (fixed point)
        # =============================================================================

        fig, ax = plt.subplots(1, 1, figsize=(10,6))
        ax.plot(test_prediction - O_lin[0,skip_train_pts+m:maxtime_pts],c='C0')
        
        # set x and y labels
        ax.set_ylabel("error FP",fontsize=fontsize)
        ax.set_xlabel("time step",fontsize=fontsize)
        
        # adjust ticks
        ax.tick_params(axis='both', labelsize=ticksize)
        ax.xaxis.set_major_locator(MaxNLocator(nbins=nbinsx_time_series))
        ax.yaxis.set_major_locator(MaxNLocator(nbins=nbinsy_time_series))

        # draw box around plot
        ax.patch.set_edgecolor('black')  
        ax.patch.set_linewidth(box_linewidth)
        
        # add title
        ax.set_title("Testing error",fontsize=fontsize)
        
        # show plot
        plt.show()
    
    # keeping training and testing results in a dictionary
    train_test_dict = {'ridge_param':ridge_param,
                       'train_rmse':train_rmse,
                       'train_rmse_real':train_rmse_real,
                       'test_rmse': test_rmse,
                       'test_rmse_real':test_rmse_real,
                       'W_out':W_out,
                       'W_out_FP':W_out_FP,
                       'W_u_inv':W_u_inv,
                       'W_u_inv_FP':W_u_inv_FP,
                       'W_n_frac':W_n_frac}

    
    # =============================================================================
    # Print FPGA code to paste into Verilog file in Quartus
    # =============================================================================
    if print_fpga_code:
        print("Printing FPGA code:\n")
        FP.print_verilog(desired_padded_word_length,
                          desired_padding,
                          desired_error_padding,
                          K_FP,
                          F_TYPE,
                          Y_TYPE,
                          U_FULL_PRECISION_TYPE,
                          u_i_SX_16,
                          ridge_param,
                          W_n_frac,
                          W_u_inv_FP,
                          W_out_FP,
                          W_u_inv,
                          train_test_depth,
                          desired_depth,
                          write_depth,
                          control_wait)
        print("Done printing FPGA code")
                
    return train_test_dict

#%%

# =============================================================================
# Find the ridge parameter that gives lowest testing error
# =============================================================================

# Learning parameters
train_pts = 2000 # number of points used for training
test_pts = 2000 # number of points used for testing
m = 1 # one step ahead prediction

# use perturbed data
perturbed_data = np.array([perturbation_dict['traintest_soln_FP'][0,:],perturbation_dict['traintest_soln_FP'][1,:],perturbation_dict['traintest_u_FP']])
perturbed_data_real = np.array([perturbation_dict['traintest_soln_real'][0,:],perturbation_dict['traintest_soln_real'][1,:],perturbation_dict['traintest_u_current']])

# defining array of ridge parameters
ridge_param_array = np.logspace(-10,1,100)

# defining array for all RMSE train with real units
rmse_train_real_array = np.zeros(len(ridge_param_array))

# defining array for all RMSE test with real units
rmse_test_real_array = np.zeros(len(ridge_param_array))

# perform training and testing for each ridge parameter
for i in range(len(ridge_param_array)):
    # get ridge parameter
    ridge_param = ridge_param_array[i]
    
    # perform training and testing with selected ridge parameter
    train_test_dict_i = training_and_testing(ridge_param,
                                             train_pts,
                                             test_pts,
                                             perturbed_data,
                                             perturbed_data_real,
                                             m=m)
    
    # save training and testing RMSE in their respective arrays
    rmse_train_real_array[i] = train_test_dict_i['train_rmse_real']
    rmse_test_real_array[i] = train_test_dict_i['test_rmse_real']

# finding the lowest training and testing error
best_train_real_rmse = np.min(rmse_train_real_array)
best_test_real_rmse = np.min(rmse_test_real_array)

# finding the ridge parameter that corresponds to the lowest error
best_train_real_ridge = ridge_param_array[np.argmin(rmse_train_real_array)]
best_test_real_ridge = ridge_param_array[np.argmin(rmse_test_real_array)]

# saving results that are specific to the ridge parameter search
ridge_search_dict = {'ridge_param_array':ridge_param_array,
                     'rmse_train_real_array':rmse_train_real_array,
                     'rmse_test_real_array':rmse_test_real_array,
                     'best train real rmse': best_train_real_rmse,
                     'best train real ridge': best_train_real_ridge,
                     'best test real rmse':best_test_real_rmse,
                     'best test real ridge':best_test_real_ridge}

# specify that FPGA code should be printed
print_fpga_code = True

# save the training and testing results that correspond to the lowest testing
# error. These are the weights that will be programmed on the FPGA.
train_test_dict = training_and_testing(best_test_real_ridge,
                                       train_pts,
                                       test_pts,
                                       perturbed_data,
                                       perturbed_data_real,
                                       m=m,
                                       show_plots=True, 
                                       print_fpga_code=print_fpga_code)

# plotting the training and testing error as a function of the ridge parameter.
# The units are scaled to mV by multiplying the RMSE by 1000.
fig, ax = plt.subplots(1, 1, figsize=(10,8))
ax.plot(ridge_param_array,rmse_train_real_array*1000, label = 'Training RMSE')
ax.plot(ridge_param_array,rmse_test_real_array*1000, label = 'Testing RMSE')

# setting x scale to be log
ax.set_xscale('log')

# set x and y axes labels
ax.set_xlabel('ridge parameter',fontsize=fontsize)
ax.set_ylabel('RMSE (mV)',fontsize=fontsize)

# setting title
ax.set_title('Ridge parameter search',fontsize=fontsize)

# adjust ticks
ax.tick_params(axis='both', labelsize=ticksize)

# draw box around plot
ax.patch.set_edgecolor('black')  
ax.patch.set_linewidth(box_linewidth)

# add legend
ax.legend(bbox_to_anchor=(1.02,0.8),fontsize=fontsize)
plt.show()

# printing results of ridge parameter search
print('\nResults of ridge parameter search:')
print('min training RMSE: %.2f mV' % (best_train_real_rmse*1000))
print('min testing RMSE: %.2f mV' % (best_test_real_rmse*1000))
print('best training ridge: %.2e' % best_train_real_ridge)
print('best testing ridge: %.2e' % best_test_real_ridge)

#%%

# =============================================================================
# If the Max10 FPGA is connected, copy and paste the printed FPGA code into
# the main Verilog file in Quartus, then compile the code and program the
# FPGA using the In-System Memory Content Editor.
# =============================================================================

#%%

# =============================================================================
# Calculating nonzero unstable steady states of double scroll using 
# the differential equations for the system, then converting to fixed-point form
# =============================================================================

# computing nonzero steady states
V1ss_plus_FP,V1ss_minus_FP = FP.compute_steady_states()

# here are the measured values for the unstable steady states, which were found
# by adjusting the fixed-points until the control current is zero when the 
# system is at either of the nonzero fixed points. Both the positive and negative
# unstable steady states are within 2% of the theoretically predicted value
V1ss_plus_measured_FP = 0.274044
V1ss_minus_measured_FP = -0.2755

#%%

# =============================================================================
# Creating the desired signals
# =============================================================================

def task_picker(task):
    """
    Create desired signals for use in the control algorithm.
    
    Parameters:
        task (int): The origin task is chosen with task=1, the two USS task is 
        chosen with task=2, and the random task is chosen with task=3
        
    Returns:
        x_desired (numpy.ndarray): Contains the desired signal in digital form
    """

    # origin task
    if task == 1:
        
        # setting the desired trajectory to zero
        x_desired = np.zeros(desired_depth)
        
    # two USS task
    elif task == 2:
        
        # setting the location of the nonzero steady states in fixed point form
        V1_max = V1ss_plus_measured_FP
        V1_min = V1ss_minus_measured_FP
        
        # set how often the switching between the two USS points occurs
        omega = 0.007575
        
        # set the rate of the requested transition with the smoothing parameter
        b = 0.11 # 20-KHz
        
        # offset the start of the desired signal to ensure it starts at one
        # of the fixed-points
        offset = 595
        
        # creating desired signal
        x_desired = smooth_square_desired(desired_depth+offset,
                                          V1_max,
                                          V1_min,
                                          omega,
                                          b)
        
        # cutting out the offset at the beginning
        x_desired = x_desired[offset:desired_depth+offset]
        
    # random task
    elif task == 3:
        
        # setting the parameters for random uniform noise
        
        # setting maximum value of the random uniform noise
        max_uniform = 1.0 
        # setting the smoothing parameter which determines how smoothly the
        # random signal repeats
        envelope_smoothing = 0.3
        
        # make the uniform random noise
        uniform_noise = np.random.uniform(-max_uniform, max_uniform, desired_depth)
    
        # setting the parameters for the low-pass filter    
        critical_f = 2500 # critical frequency for butterworth digital low-pass filter
        order = 20 # order of butterworth digital low-pass filter
    
        # make the smooth envelope to apply to the noise so it can smoothly
        # repeat
        envelope = smooth_envelope(desired_depth, 
                                   envelope_smoothing)
    
        # make the butterworth low-pass filter        
        low_pass_filter = signal.butter(order, critical_f, btype='lowpass', fs=sample_f, output='sos')
        
        # apply the low pass filter to the uniform noise, and apply the
        # smoothing envelope
        x_desired = signal.sosfilt(low_pass_filter, uniform_noise)*envelope
        

    # converting the desired signal to fixed point form
    x_desired_FP = Fxp().like(FP.SM1_16_TYPE) # to match scale of linear features, so error can have scale of differences
    x_desired_FP.rounding = 'around'
    x_desired_FP.set_val(x_desired)
    
    # setting x_desired equal to the raw digital value, which is written to
    # the FPGA memory
    x_desired = x_desired_FP.raw()
    
    # Plotting the desired signal
    fig, ax = plt.subplots(1, 1, figsize=(10,8))
    ax.plot(x_desired,c='black',linewidth=2)
    
    # set x and y limits
    ax.set_xlim(0,desired_depth)
    ax.set_ylim(FP.S15_0_TYPE.lower,FP.S15_0_TYPE.upper)
    
    # set x and y labels
    ax.set_xlabel('Memory address',fontsize=fontsize)
    ax.set_ylabel('Desired Value',fontsize=fontsize)
    
    # set title
    ax.set_title('Desired Signal',fontsize=fontsize)
    
    # adjust ticks
    ax.tick_params(axis='both', labelsize=fontsize)  
    
    # draw box around plot
    ax.patch.set_edgecolor('black')  
    ax.patch.set_linewidth(box_linewidth)
    
    # show plot
    plt.show()
        
    return x_desired
    
#%%

# =============================================================================
# Connect to the FPGA 
# =============================================================================

if fpga_connected:
    # get the number of fractional bits for the weights. This is needed to
    # convert the control gain K to its integer form for writing to memory
    W_n_frac = train_test_dict['W_n_frac']
    
    # connect to the FPGA, providing additional parameters for writing
    # to memory
    fpga = FPGAController(desired_word_length=FP.desired_word_length,
                          desired_depth=desired_depth,
                          W_n_frac=W_n_frac,
                          weight_word_length=FP.weight_word_length)
#%%

# =============================================================================
# Defining control function
# =============================================================================

def do_control(K,
               delay_level,
               task,
               feedback_linearization,
               x_desired):
    """
    Start control and collect results by writing and reading memory addresses
    on the connected FPGA.
    
    Parameters:
        K (float): The control gain
        
        delay_level (int): The amount of delay in microseconds to add before
        the control perturbation is applied, i.e. the closed-loop latency. Can
        take on integers from 0 to 31
        
        task (int): Specifies the desired signal. The origin task is chosen 
        with task=1, the two USS task is chosen with task=2, and the random 
        task is chosen with task=3
        
        feedback_linearization (int): Whether the controller is strictly
        linear (False) or contains nonlinear terms (True)
        
        x_desired (numpy.ndarray): Contains the desired signal in digital form.
        
    Returns:
        control_dict (dict): A dictionary containing the following keys:
            - 'doublescroll_soln_real' : Array containing time-series of 
              (V1,V2) in Volts
            - 'DAC_current' : Control current (uA)
            - 'V1_desired_real' : Desired signal (V)
            - 'V1_prediction_real' : Prediction (V)
            - 'control_rmse' : The RMSE between the controlled system and
              the desired signal
            - 'prediction_rmse' : The RMSE between the controlled system and
              the predicted system. The prediction is evaluated by the FPGA
            - 'current_rms' : The RMS of the control current in uA
            - 'time_array' : Array of time steps in (ms)
    """
    
    # write control gain
    K_int = fpga.write_K(K)

    # set added control loop latency
    fpga.write_delay(delay_level)
    
    # determine whether the controller is linear or uses NG-RC
    fpga.write_feedback_linearization(feedback_linearization)

    # set desired state
    fpga.write_desired(x_desired) # WIP: careful, x_desired must be set by task picker before

    # set mode to close loop feedback (rather than open when perturbing the system for training/testing)
    fpga.set_mode(0)
    
    # start control
    fpga.control_switch(1)
    
    # 
    time.sleep(1)
    
    # load data from controlled system
    V1_digital = fpga.load_data('ADC1', signed = False)
    V2_digital = fpga.load_data('ADC2', signed = False)
    DAC_digital = fpga.load_data('CTRL', signed = False)
    V1_desired_digital = fpga.load_data('DV1o', signed = True)
    V1_prediction_digital = fpga.load_data('PRED',signed = True)
        
    # converting 16-bit values to 12-bit values for error calculations
    V1_desired_digital_U12 = V1_desired_digital / 2**(FP.U16_0_TYPE.n_word - FP.U12_0_TYPE.n_word) + np.abs(FP.S11_0_TYPE.lower)
    V1_prediction_digital_U12 = V1_prediction_digital / 2**(FP.U16_0_TYPE.n_word - FP.U12_0_TYPE.n_word) + np.abs(FP.S11_0_TYPE.lower)
    
    # gathering V1 and V2 in single array
    doublescroll_soln_digital = np.array([V1_digital,V2_digital])
    
    # converting digital data to fixed-point form
        
    # converting V1,V2, and control perturbation to fixed-point
    doublescroll_soln_FP, DAC_FP = FP.digital_to_fixed_point(doublescroll_soln_digital,
                                                             DAC_digital)
    
    # converting the prediction for V1 to fixed-point
    V1_prediction_FP = Fxp().like(FP.SM1_16_TYPE)
    V1_prediction_FP.set_val(V1_prediction_digital,raw = True)
    
    # converting the desired signal for V1 to fixed-point
    V1_desired_FP = Fxp().like(FP.SM1_16_TYPE)
    V1_desired_FP.set_val(V1_desired_digital,raw = True)
    
    # converting digital data to real values with gain applied
    doublescroll_soln_gain = digital_to_gain(doublescroll_soln_digital,FP.U12_0_TYPE.upper,FP.V_ADC_min,FP.V_ADC_max)
    V1_desired_gain = digital_to_gain(V1_desired_digital_U12,FP.U12_0_TYPE.upper,FP.V_ADC_min,FP.V_ADC_max)
    V1_prediction_gain = digital_to_gain(V1_prediction_digital_U12,FP.U12_0_TYPE.upper,FP.V_ADC_min,FP.V_ADC_max)
    
    # converting real values with gain applied to real values
    doublescroll_soln_real = FP.gain_to_real(doublescroll_soln_gain)
    DAC_voltage = digital_to_gain(DAC_digital,FP.U16_0_TYPE.upper,FP.V_DAC_min,FP.V_DAC_max)
    DAC_current = V_to_I(DAC_voltage,FP.gain)
    V1_desired_real = V1_desired_gain/FP.V1_gain - FP.V1_real_offset
    V1_prediction_real = V1_prediction_gain/FP.V1_gain - FP.V1_real_offset

    # calculating error signal
    error_control_FP = doublescroll_soln_FP[0,transient_end:] - V1_desired_FP[transient_end:]

    # calculate the control RMSE (mV)
    # Note that the solution is shifted forwards in time by one time step. This is because the
    # FPGA saves the desired signal at time step i+1, and the solution at time step i. 
    control_rmse = np.sqrt(np.mean((V1_desired_real[transient_end:-1] - doublescroll_soln_real[0,transient_end + 1:])**2))*voltage_scaling
    
    # calculate the prediction RMSE (mV)
    # Note that the solution is shifted forwards in time by the prediction horizon m.
    prediction_rmse = np.sqrt(np.mean((V1_prediction_real[transient_end:-m] - doublescroll_soln_real[0,transient_end + m:])**2))*voltage_scaling

    # calculate the current RMS (uA)    
    current_rms = np.sqrt(np.mean((DAC_current[transient_end:])**2))*current_scaling
    
    # make time array (ms)
    time_array = np.array([i for i in range(write_depth)])*dt*time_scaling
    
    # convert current from A to uA
    DAC_current*=current_scaling
    
    # store control data in a dictionary
    control_dict = {'doublescroll_soln_real':doublescroll_soln_real,
                    'DAC_current':DAC_current,
                    'V1_desired_real':V1_desired_real,
                    'V1_prediction_real':V1_prediction_real,
                    'control_rmse':control_rmse,
                    'prediction_rmse':prediction_rmse,
                    'current_rms':current_rms,
                    'time_array':time_array}

    # stop control
    fpga.control_switch(0)

    return control_dict

#%%

# =============================================================================
# Control the double scroll system
# =============================================================================

# select the task (1:origin,2:two USS,3:random)
task = 2

# get the desired signal that corresponds to the specified task
x_desired = task_picker(task)

# set control parameters
K = 0.8 # set the control gain
delay_level = 0 # set the added delay in the control loop in microseconds

# determine whether the controller is linear or uses NG-RC (1:NG-RC)
feedback_linearization = 1

if fpga_connected:
    # do control
    control_dict = do_control(K,
                              delay_level,
                              task,
                              feedback_linearization,
                              x_desired)
    
    
else:
    # if the FPGA is not connected, load previously collected control data
    control_dict = example_data_dict['control_dict']
#%%

# =============================================================================
# Plot control results
# =============================================================================

# Control parameters
# Plot Parameters
fontsize = 20
ticksize = 20
nbinsx_attractor = 6
nbinsy_attractor = 8
nbinsx_time_series = 9
nbinsy_time_series = 5
box_linewidth = 1.5
tick_width = 1
tick_length= 7
attractor_color = 'darkviolet'
transient_color = 'silver'
desired_max = np.max(control_dict['V1_desired_real'])
desired_min = np.min(control_dict['V1_desired_real'])

# creating figure
fig, axs = plt.subplots(2, 1, figsize=(14, 9))

# plotting controlled V1 and V2 in first bottom subplot
axs[0].plot(control_dict['time_array'][:plot_pts],control_dict['doublescroll_soln_real'][0][:plot_pts],c='blue', label = r"$V_1$")
axs[0].plot(control_dict['time_array'][:plot_pts],control_dict['doublescroll_soln_real'][1][:plot_pts],c='red',label = r"$V_2$")
axs[0].plot(control_dict['time_array'][control_wait:plot_pts],control_dict['V1_desired_real'][control_wait:plot_pts],'-',c='black', label = r"$V_{1,des}$",zorder = -1,lw=1,alpha = 1.0)
axs[0].axvline(x=time_scaling*control_wait*dt,linestyle='dashed',c='black')

# plotting control current in the second bottom subplot
axs[1].plot(control_dict['time_array'][:plot_pts],control_dict['DAC_current'][:plot_pts],c='darkorange',label = "u")#label = r'$u_1$'
axs[1].axvline(x=time_scaling*control_wait*dt,linestyle='dashed',c='black')

# set y limits
axs[0].set_ylim(FP.V1_real_min,FP.V1_real_max)
axs[1].set_ylim(FP.I_min*current_scaling+0.05*FP.I_min*current_scaling,FP.I_max*current_scaling+0.05*FP.I_max*current_scaling)

# set x and y labels
axs[0].set_ylabel("Voltage (V)",fontsize = fontsize)
axs[1].set_ylabel(r'Control Current $(\mu A)$',fontsize = fontsize)
axs[1].set_xlabel('time (ms)',fontsize=fontsize)

# adjust ticks for for bottom plots
for ax in axs.flat:
    ax.tick_params(axis='both', labelsize=ticksize)
    ax.xaxis.set_major_locator(MaxNLocator(nbins=nbinsx_time_series))
    ax.yaxis.set_major_locator(MaxNLocator(nbins=nbinsy_time_series))
    ax.tick_params(axis = 'x', which='minor', direction='out', length=tick_length, width=tick_width, colors='black')
    ax.tick_params(axis = 'x', which='major', direction='out', length=tick_length, width=tick_width, colors='black')
    ax.tick_params(axis = 'y', which='minor', direction='out', length=tick_length, width=tick_width, colors='black')
    ax.tick_params(axis = 'y', which='major', direction='out', length=tick_length, width=tick_width, colors='black')
    
    # add boxes around plots
    ax.patch.set_edgecolor('black')  
    ax.patch.set_linewidth(box_linewidth)

# remove x ticks for top plot
axs[0].set_xticks([])

# adjust subplot spacing
plt.subplots_adjust(hspace=0.1)

# add legend
fig.legend(bbox_to_anchor=(1.05,0.8),fontsize=fontsize)

# set title 
fig.suptitle('Controlled system', fontsize=fontsize)

# plotting figure
plt.show()


# plot attractor
fig, ax = plt.subplots(1, 1, figsize=(8, 8))
ax.plot(control_dict['doublescroll_soln_real'][0][:transient_end],control_dict['doublescroll_soln_real'][1][:transient_end], c = transient_color,zorder = -10,label='unperturbed system')
ax.plot(control_dict['doublescroll_soln_real'][0][transient_end:plot_pts],control_dict['doublescroll_soln_real'][1][transient_end:plot_pts], c = attractor_color,zorder = 20,label='controlled system')
ax.plot(control_dict['doublescroll_soln_real'][0][control_wait:transient_end],control_dict['doublescroll_soln_real'][1][control_wait:transient_end],'--', c = attractor_color,label='transient')
ax.scatter(control_dict['doublescroll_soln_real'][0][control_wait],control_dict['doublescroll_soln_real'][1][control_wait], c = attractor_color,zorder = 10,marker = 'x',s=70,label='control start')

# setting x and y limits
ax.set_xlim(FP.V1_real_min,FP.V1_real_max)
ax.set_ylim(FP.V2_real_min,FP.V2_real_max)

# setting x and y labels
ax.set_xlabel(r"$V_1$ (V)",fontsize = fontsize)
ax.set_ylabel(r"$V_2$ (V)",fontsize = fontsize)

# adjust ticks
ax.tick_params(axis='both', labelsize=ticksize)
ax.xaxis.set_major_locator(MaxNLocator(nbins=nbinsx_attractor))
ax.yaxis.set_major_locator(MaxNLocator(nbins=nbinsy_attractor))

# draw box around plot
ax.patch.set_edgecolor('black')  
ax.patch.set_linewidth(box_linewidth)  

# plot vertical desired trajectory lines for two USS task
if task == 2:
    ax.axvline(x=desired_max,c='black',zorder=-1,label='desired max/min')
    ax.axvline(x=desired_min,c='black',zorder=-1)    

# add legend
ax.legend(bbox_to_anchor=(1.05,0.8),fontsize=fontsize)

# set title 
ax.set_title('Controlled attractor', fontsize=fontsize)

# print key results
print("\nControl results:")
print("Control RMSE: {:.2f} mV".format(round(control_dict['control_rmse'], 2)))
print("Current RMS: {:.2f} uA".format(round(control_dict['current_rms'], 2)))

#%%

# =============================================================================
# Finding optimal features for the model using system identification
# =============================================================================

# setting number of training and testing points
train_pts = 9400
test_pts = 4700
maxtime_pts = train_pts + test_pts

# setting sysidentpy parameters
ridge_param = 1e-2
degree = 3
n_info_values = 30

# load example data
with open('example_data.pkl', 'rb') as fp:
    example_data_dict = pickle.load(fp)
sysident_dict = example_data_dict['sysident_dict']
perturbed_data = sysident_dict['perturbed_data']

# separating perturbed data
doublescroll_soln = perturbed_data[:2,:maxtime_pts] # (V1,V2)
u_train_test = perturbed_data[2,:maxtime_pts] # u

# setting dimension of the x and y features. Notice how V1i is put into y,
# rather than x. This is because of the structure of the 'NARMAX' model used below.
# This can be avoided by using the 'NFIR' model instead.
x_dim = 3 # for V2i,V1i-V2i,ui
y_dim = 1 # for V1i

# initializing features to zero
x_train = np.zeros((x_dim,train_pts))
x_test = np.zeros((x_dim,test_pts))
y_train = np.zeros((y_dim,train_pts))
y_test = np.zeros((y_dim,test_pts))

# setting training features
x_train[0,:] = doublescroll_soln[1,:train_pts] # V2i
x_train[1,:] = doublescroll_soln[0,:train_pts]-doublescroll_soln[1,:train_pts] # V1i-V2i
x_train[2,:] = u_train_test[:train_pts] # ui
y_train[0,:] = doublescroll_soln[0,:train_pts] # V1i

# setting testing features
x_test[0,:] = doublescroll_soln[1,train_pts:train_pts + test_pts] # V2i
x_test[1,:] = doublescroll_soln[0,train_pts:train_pts + test_pts]-doublescroll_soln[0,train_pts:train_pts + test_pts] # V1i-V2i
x_test[2,:] = u_train_test[train_pts:train_pts + test_pts] # ui
y_test[0,:] = doublescroll_soln[0,train_pts:train_pts + test_pts] # V1i

# transposing data
x_train = x_train.T
x_test = x_test.T
y_train = y_train.T
y_test = y_test.T

# setting the basis function to be a polynomial with degree 3
basis_function = Polynomial(degree=degree)

# making a model using sysidentpy
model = FROLS(
    order_selection=True,
    n_info_values=n_info_values, # limits number of regressors
    ylag=2, # V1i,V1im1
    xlag=[[1,2], [1,2], [1,2]], # V2i,V2im1, V1i-V2i, V1im1-V2im1, ui, uim1
    info_criteria="aic",
    estimator="ridge_regression", # Requires sysidentpy version 0.3.4 or higher
    ridge_param=ridge_param, # Requires sysidentpy version 0.3.4 or higher
    basis_function=basis_function,
    model_type = 'NARMAX',
)

# fit the model
model.fit(X=x_train, y=y_train)

# make predictions with trained model

# training data
yhat_train = model.predict(X=x_train, y=y_train,steps_ahead=1)

# testing data
yhat_test = model.predict(X=x_test, y=y_test,steps_ahead=1)

# computing training and testing error
train_rmse = np.sqrt(np.mean((yhat_train - y_train)**2))
test_rmse = np.sqrt(np.mean((yhat_test - y_test)**2))

# storing the results in a dataframe
my_results = results(
        model.final_model,
        model.theta,
        model.err,
        model.n_terms,
        err_precision=8,
        dtype="sci",
    )
r = pd.DataFrame(my_results,columns=["Regressors", "Parameters", "ERR"])
pd.set_option('display.max_colwidth', None)
pd.set_option('display.max_columns', None)

# the default for of the features is in terms of x and y, so we 
# define a dictionary of replacements to switch the variables to V1i,V2i,ui,etc.

replacements = {
    'y': 'V1',
    'x1': 'V2',
    'x2': '(V1-V2)',
    'x3': 'u',
    '(k-1)': '_{i}',
    '(k-2)': '_{i-1}',
    '(k-3)':'_{i-2}',
    '(k-4)':'_{i-3}',
    '(k-5)':'_{i-4}',
    '(k-6)':'_{i-5}',
    '(k-7)':'_{i-6}',
    '(k-8)':'_{i-7}',
    '(k-9)':'_{i-8}',
    '(k-10)':'_{i-9}'
    # Add more replacements as needed
}

# replacing feature names with new ones
new_regressor_list = []
latex_expressions = []
for string in r['Regressors']:
    for old, new in replacements.items():
        string = string.replace(old, new)
        latex_string = r'$' + string + '$'
    new_regressor_list.append(string)
    latex_expressions.append(latex_string)

r_renamed = r
r_renamed['Regressors'] = new_regressor_list

# printing results
print("\nSystem identification results")
print("train nrmse:",train_rmse)
print("test nrmse:",test_rmse)
# this gives the features, the corresponding value of the weights, and the 
# error reduction corresponding to each term
print(r_renamed)
#%%

# =============================================================================
# Plotting results of system identification
# =============================================================================

# creating custom math expressions to label the points in the plot. This is not
# generalizeable.
latex_expressions = [r'$V_{1,i}$',
                     r'$V_{1,i-1}$',
                     r'$(V_{1,i}-V_{2,i})V_{1,i-1}^2$',
                     r'$(V_{1,i-1}-V_{2,i-1})V_{1,i}^2$',
                     r'$(V_{1,i}-V_{2,i})^3$',
                     r'$u_{1,i}$',
                     r'$(V_{1,i-1}-V_{2,i-1})^3$',
                     r'$V_{2,i}$',
                     r'$V_{2,i-1}$',
                     r'$V_{2,i}^3$',
                     r'$V_{2,i-1}$',
                     r'$V_{1,i}^3$',
                     r'$u_i V_{1,i-1}^2$',
                     r'$(V_{1,i-1}-V_{2,i-1}) V_{2,i-1}$'
                     ]

# removing the last feature for readability, and it is not used in the final model
latex_expressions = latex_expressions[:-1]

# plot parameters
ticksize = 30
fontsize = 28
nbins_x = 9
nbins_y = 7
box_linewidth = 2
top_padding = 6500 # padding above maximum information criteria value
bottom_padding = 1500 # padding below minimum information criteria value
label_shift = 1800 # vertical shift of labels relative to points

# setting the x and y positions of the points in the plot
number_of_terms = np.arange(len(latex_expressions)) + 1 
information_criteria = model.info_values[:len(latex_expressions)]

x_position_label = number_of_terms + np.array([0.45,0.8,1.6,1.8,1.15,0.5,1.6,0.45,0.6,0.42,0.6,0.4,0.7])
y_position_label = information_criteria  + np.array([-400,-700,-300,-300,-300,-50,-100,-400,-200,-800,-400,-600,-150])

# Create figure
fig, ax = plt.subplots(1,1, figsize = (14,8))
# These are the terms used in the model
ax.scatter(number_of_terms[:-4], information_criteria[:-4], marker='o', color='red', s=200)
ax.plot(number_of_terms[:-3],information_criteria[:-3],color = 'red',lw = 5,label='features used in model')
# These are the terms excluded from the model
ax.scatter(number_of_terms[9:], information_criteria[9:], marker='x', color='black', s=200, zorder = 100,linewidths=5)
ax.plot(number_of_terms[9:],information_criteria[9:],color = 'black', zorder = 100,lw=5,label='features excluded from model')
# Annotate each point with the custom labels (rotate labels vertically and shift upwards)
for i, (x, y) in enumerate(zip(x_position_label, y_position_label)):
    ax.annotate(latex_expressions[i], (x, y), fontsize=fontsize, fontweight='bold',
                 ha='center', va='bottom', color='black', rotation=45)

# setting x and y limits
ax.set_xlim(0,14.7)
ax.set_ylim(np.min(information_criteria) - bottom_padding, np.max(information_criteria) + top_padding)

# setting x and y labels
ax.set_xlabel('number of terms',fontsize=fontsize)
ax.set_ylabel("Information Criteria",fontsize = fontsize)

# adjusting ticks
ax.set_xticks(np.arange(0, 14, 2))
ax.tick_params(axis='both', labelsize=ticksize)
# ax.xaxis.set_major_locator(MaxNLocator(nbins=nbins_x))
ax.yaxis.set_major_locator(MaxNLocator(nbins=nbins_y))

# adding box around plot
ax.patch.set_edgecolor('black')  
ax.patch.set_linewidth(box_linewidth) 

# add legend
fig.legend(bbox_to_anchor=(1.5,0.8),fontsize=fontsize)

# set title 
ax.set_title('Optimal feature selection', fontsize=fontsize)

# Show the plot
plt.tight_layout()
plt.show()

#%%

# sysident_dict = {'perturbed_data':perturbed_data}
# example_data_dict = {'perturbation_dict':perturbation_dict,
#                       'control_dict':control_dict,
#                       'sysident_dict':sysident_dict
#                       }

# with open('example_data_2.pkl', 'wb') as f:
#     pickle.dump(example_data_dict, f)
